@@ -68,17 +68,18 @@ use std::{
     fs::{symlink_metadata, Metadata},
     os::{
         fd::RawFd,
+        linux::fs::MetadataExt as LinuxMetadataExt,
         raw::{c_char, c_int},
         unix::{
             ffi::OsStrExt,
-            fs::{FileTypeExt, MetadataExt},
+            fs::{FileTypeExt, MetadataExt as UnixMetadataExt},
         },
     },
     path::{Path, PathBuf},
 };
 
 /// `lock_state_t_t` type represents possible states for the sandbox lock.
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types)]
 pub type lock_state_t = u8;
 
 /// The sandbox lock is off, allowing all sandbox commands.
@@ -127,7 +128,7 @@ impl fmt::Display for LockState {
 }
 
 /// `action_t` type represents possible sandboxing action values.
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types)]
 pub type action_t = u8;
 
 /// Allow system call.
@@ -211,20 +212,42 @@ impl fmt::Display for Action {
 
 const EFAULT: i32 = 14;
 const EINVAL: i32 = 22;
+const LIB_MAJOR: &'static str = env!("CARGO_PKG_VERSION_MAJOR");
+const SYD_ATIME: i64 = 505958400;
+const SYD_CTIME: i64 = -2036448000;
+const SYD_MTIME: i64 = -842745600;
 
-#[inline(always)]
 fn check_stat(stat: &Metadata) -> bool {
+    let lib_major = if let Ok(lib_major) = LIB_MAJOR.parse() {
+        lib_major
+    } else {
+        return false;
+    };
+
+    // Check file type.
     if !stat.file_type().is_char_device() {
         return false;
     }
 
+    // Check timestamps.
+    if stat.st_atime() != SYD_ATIME {
+        return false;
+    }
+    if stat.st_ctime() != SYD_CTIME {
+        return false;
+    }
+    if stat.st_mtime() != SYD_MTIME {
+        return false;
+    }
+
     let rdev = stat.rdev();
 
-    let major = (rdev >> 8) & 0xff;
-    let minor = rdev & 0xff;
+    let api_major = (rdev >> 8) & 0xff;
+    let api_minor = rdev & 0xff;
 
-    // dev/null
-    major == 1 && minor == 3
+    // Device type represents syd(2) API version.
+    // See RETURN VALUE section in syd(2) manual page.
+    api_major == lib_major && api_minor >= 1
 }
 
 fn stat<P: AsRef<Path>>(path: P) -> c_int {
@@ -331,6 +354,54 @@ pub extern "C" fn syd_lock(state: lock_state_t) -> c_int {
     stat(&format!("/dev/syd/lock:{state}"))
 }
 
+/// Checks if Filesystem sandboxing is enabled.
+///
+/// Returns true if Filesystem sandboxing is enabled, false otherwise.
+#[no_mangle]
+pub extern "C" fn syd_enabled_fs() -> bool {
+    stat("/dev/syd/sandbox/fs?") == 0
+}
+
+/// Enable Filesystem sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_enable_fs() -> c_int {
+    stat("/dev/syd/sandbox/fs:on")
+}
+
+/// Disable Filesystem sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_disable_fs() -> c_int {
+    stat("/dev/syd/sandbox/fs:off")
+}
+
+/// Checks if walk sandboxing is enabled.
+///
+/// Returns true if walk sandboxing is enabled, false otherwise.
+#[no_mangle]
+pub extern "C" fn syd_enabled_walk() -> bool {
+    stat("/dev/syd/sandbox/walk?") == 0
+}
+
+/// Enable walk sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_enable_walk() -> c_int {
+    stat("/dev/syd/sandbox/walk:on")
+}
+
+/// Disable walk sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_disable_walk() -> c_int {
+    stat("/dev/syd/sandbox/walk:off")
+}
+
 /// Checks if stat sandboxing is enabled.
 ///
 /// Returns true if stat sandboxing is enabled, false otherwise.
@@ -1051,6 +1122,28 @@ pub extern "C" fn syd_disable_tpe() -> c_int {
     stat("/dev/syd/sandbox/tpe:off")
 }
 
+/// Set the default action for Filesystem sandboxing.
+#[no_mangle]
+pub extern "C" fn syd_default_fs(action: action_t) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+    stat(&format!("/dev/syd/default/fs:{action}"))
+}
+
+/// Set the default action for Walk Sandboxing.
+#[no_mangle]
+pub extern "C" fn syd_default_walk(action: action_t) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+    stat(&format!("/dev/syd/default/walk:{action}"))
+}
+
 /// Set the default action for Stat Sandboxing.
 #[no_mangle]
 pub extern "C" fn syd_default_stat(action: action_t) -> c_int {
@@ -1395,7 +1488,7 @@ pub extern "C" fn syd_default_tpe(action: action_t) -> c_int {
 /// Adds a request to the _ioctl_(2) denylist.
 #[no_mangle]
 pub extern "C" fn syd_ioctl_deny(request: u64) -> c_int {
-    stat(&format!("/dev/syd/ioctl/deny+{request}"))
+    stat(&format!("/dev/syd/deny/ioctl+{request}"))
 }
 
 /// Adds an entry to the Integrity Force map for Force Sandboxing.
@@ -1479,6 +1572,98 @@ pub extern "C" fn syd_force_clr() -> c_int {
     stat("/dev/syd/force^")
 }
 
+/// Adds to the given actionlist of Filesystem sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_fs_add(action: action_t, name: *const c_char) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+
+    // Call magic function with add operator.
+    esyd(&format!("{action}/fs"), name, b'+')
+}
+
+/// Removes the first instance from the end of the given actionlist of
+/// Filesystem sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_fs_del(action: action_t, name: *const c_char) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+
+    // Call magic function with del operator.
+    esyd(&format!("{action}/fs"), name, b'-')
+}
+
+/// Removes all matching patterns from the given actionlist of Filesystem sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_fs_rem(action: action_t, name: *const c_char) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+
+    // Call magic function with rem operator.
+    esyd(&format!("{action}/fs"), name, b'^')
+}
+
+/// Adds to the given actionlist of walk sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_walk_add(action: action_t, glob: *const c_char) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+
+    // Call magic function with add operator.
+    esyd(&format!("{action}/walk"), glob, b'+')
+}
+
+/// Removes the first instance from the end of the given actionlist of
+/// walk sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_walk_del(action: action_t, glob: *const c_char) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+
+    // Call magic function with del operator.
+    esyd(&format!("{action}/walk"), glob, b'-')
+}
+
+/// Removes all matching patterns from the given actionlist of walk sandboxing.
+///
+/// Returns 0 on success, negated errno on failure.
+#[no_mangle]
+pub extern "C" fn syd_walk_rem(action: action_t, glob: *const c_char) -> c_int {
+    // Convert action_t enum to corresponding action string.
+    let action = match Action::try_from(action) {
+        Ok(action) => action,
+        Err(_) => return -EINVAL,
+    };
+
+    // Call magic function with rem operator.
+    esyd(&format!("{action}/walk"), glob, b'^')
+}
+
 /// Adds to the given actionlist of stat sandboxing.
 ///
 /// Returns 0 on success, negated errno on failure.
@@ -1663,52 +1848,6 @@ pub extern "C" fn syd_exec_rem(action: action_t, glob: *const c_char) -> c_int {
     esyd(&format!("{action}/exec"), glob, b'^')
 }
 
-/// Adds to the given actionlist of ioctl sandboxing.
-///
-/// Returns 0 on success, negated errno on failure.
-#[no_mangle]
-pub extern "C" fn syd_ioctl_add(action: action_t, glob: *const c_char) -> c_int {
-    // Convert action_t enum to corresponding action string.
-    let action = match Action::try_from(action) {
-        Ok(action) => action,
-        Err(_) => return -EINVAL,
-    };
-
-    // Call magic function with add operator.
-    esyd(&format!("{action}/ioctl"), glob, b'+')
-}
-
-/// Removes the first instance from the end of the given actionlist of
-/// ioctl sandboxing.
-///
-/// Returns 0 on success, negated errno on failure.
-#[no_mangle]
-pub extern "C" fn syd_ioctl_del(action: action_t, glob: *const c_char) -> c_int {
-    // Convert action_t enum to corresponding action string.
-    let action = match Action::try_from(action) {
-        Ok(action) => action,
-        Err(_) => return -EINVAL,
-    };
-
-    // Call magic function with del operator.
-    esyd(&format!("{action}/ioctl"), glob, b'-')
-}
-
-/// Removes all matching patterns from the given actionlist of ioctl sandboxing.
-///
-/// Returns 0 on success, negated errno on failure.
-#[no_mangle]
-pub extern "C" fn syd_ioctl_rem(action: action_t, glob: *const c_char) -> c_int {
-    // Convert action_t enum to corresponding action string.
-    let action = match Action::try_from(action) {
-        Ok(action) => action,
-        Err(_) => return -EINVAL,
-    };
-
-    // Call magic function with rem operator.
-    esyd(&format!("{action}/ioctl"), glob, b'^')
-}
-
 /// Adds to the given actionlist of create sandboxing.
 ///
 /// Returns 0 on success, negated errno on failure.
