@@ -2,7 +2,7 @@
 // pandora: syd's Dump Inspector & Profile Writer
 // pandora.rs: Main entry point
 //
-// Copyright (c) 2021, 2024, 2025 Ali Polatel <alip@exherbo.org>
+// Copyright (c) 2021, 2024, 2025 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -15,7 +15,7 @@ use std::{
     fmt::Write as FmtWrite,
     fs::{metadata, File, OpenOptions},
     hash::{Hash, Hasher},
-    io::{self, stdin, BufRead, BufReader, Read, Write as IoWrite},
+    io::{self, stderr, stdin, BufRead, BufReader, Read, Write as IoWrite},
     iter::FromIterator,
     net::{IpAddr, SocketAddrV4, SocketAddrV6},
     os::{
@@ -120,12 +120,12 @@ const CAP_ORDER: &[&str] = &[
     "cnet",
     "snet",
     // base capabilities
+    "fs",
     "walk",
     "stat",
     "read",
     "write",
     "exec",
-    "ioctl",
     "create",
     "delete",
     "rename",
@@ -150,28 +150,28 @@ const CAP_ORDER: &[&str] = &[
     "net/sendfd",
 ];
 
-#[allow(clippy::disallowed_types)]
+#[expect(clippy::disallowed_types)]
 type PandoraMap<K, V> = std::collections::HashMap<K, V, ahash::RandomState>;
-#[allow(clippy::disallowed_types)]
+#[expect(clippy::disallowed_types)]
 type PandoraSet<K> = std::collections::HashSet<K, ahash::RandomState>;
 
 // write! which returns Errno.
 macro_rules! w {
     ($out:expr) => {
-        write!($out).or(Err(Errno::EIO))
+        retry_on_intr(|| write!($out).or(Err(Errno::EIO)))
     };
     ($out:expr, $($arg:tt)*) => {
-        write!($out, $($arg)*).or(Err(Errno::EIO))
+        retry_on_intr(|| write!($out, $($arg)*).or(Err(Errno::EIO)))
     };
 }
 
 // writeln! which returns Errno.
 macro_rules! wln {
     ($out:expr) => {
-        writeln!($out).map_err(err2no)
+        retry_on_intr(|| writeln!($out).map_err(err2no))
     };
     ($out:expr, $($arg:tt)*) => {
-        writeln!($out, $($arg)*).map_err(err2no)
+        retry_on_intr(|| writeln!($out, $($arg)*).map_err(err2no))
     };
 }
 
@@ -264,14 +264,74 @@ impl<'de> Deserialize<'de> for Capability {
     }
 }
 
+#[derive(Clone, Debug)]
+enum IoctlEntry {
+    Name(String),
+    Val(u64),
+}
+
+impl<'de> Deserialize<'de> for IoctlEntry {
+    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        struct IoctlEntryVisitor;
+
+        impl<'de> Visitor<'de> for IoctlEntryVisitor {
+            type Value = IoctlEntry;
+
+            fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
+                formatter.write_str("integer or string")
+            }
+
+            fn visit_u64<E>(self, value: u64) -> Result<Self::Value, E>
+            where
+                E: serde::de::Error,
+            {
+                Ok(IoctlEntry::Val(value))
+            }
+
+            fn visit_i64<E>(self, value: i64) -> Result<Self::Value, E>
+            where
+                E: serde::de::Error,
+            {
+                if value < 0 {
+                    Err(E::custom("negative integer for ioctl"))
+                } else {
+                    Ok(IoctlEntry::Val(value as u64))
+                }
+            }
+
+            fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
+            where
+                E: serde::de::Error,
+            {
+                Ok(IoctlEntry::Name(value.to_owned()))
+            }
+
+            fn visit_string<E>(self, value: String) -> Result<Self::Value, E>
+            where
+                E: serde::de::Error,
+            {
+                Ok(IoctlEntry::Name(value))
+            }
+        }
+
+        deserializer.deserialize_any(IoctlEntryVisitor)
+    }
+}
+
 #[derive(Clone, Debug)]
 enum Access {
     Path {
         ctx: String,
         cap: Option<Capability>,
-        ctl: Option<Vec<String>>,
+        fs: Option<String>,
         path: String,
-        args: Option<Vec<u64>>,
+    },
+    Ioctl {
+        ctx: String,
+        ctl: Vec<IoctlEntry>,
     },
     InetAddr {
         ctx: String,
@@ -323,7 +383,8 @@ impl<'de> Deserialize<'de> for Access {
                 let mut addr: Option<String> = None;
                 let mut unix_: Option<String> = None;
 
-                let mut ctl: Option<Vec<String>> = None;
+                let mut ctl: Option<Vec<IoctlEntry>> = None;
+                let mut fs: Option<String> = None;
                 let mut cmd: Option<String> = None;
                 let mut argv: Option<Vec<String>> = None;
                 let mut time_: Option<String> = None;
@@ -376,6 +437,12 @@ impl<'de> Deserialize<'de> for Access {
                             }
                             ctl = map.next_value()?;
                         }
+                        "fs" => {
+                            if fs.is_some() {
+                                return Err(serde::de::Error::duplicate_field("fs"));
+                            }
+                            fs = map.next_value()?;
+                        }
                         "cmd" => {
                             if cmd.is_some() {
                                 return Err(serde::de::Error::duplicate_field("cmd"));
@@ -421,13 +488,9 @@ impl<'de> Deserialize<'de> for Access {
 
                 // Now decide which variant to build based on which fields we have:
                 if let Some(path) = path {
-                    Ok(Access::Path {
-                        ctx,
-                        cap,
-                        ctl,
-                        path,
-                        args,
-                    })
+                    Ok(Access::Path { ctx, cap, fs, path })
+                } else if let Some(ctl) = ctl {
+                    Ok(Access::Ioctl { ctx, ctl })
                 } else if let Some(addr) = addr {
                     Ok(Access::InetAddr { ctx, cap, addr })
                 } else if let Some(unix) = unix_ {
@@ -452,7 +515,7 @@ impl<'de> Deserialize<'de> for Access {
 // lookup_addr is borrowed from Syd to avoid depending on Syd.
 // syd::dns::lookup_addr:
 // Performs a reverse DNS lookup for the given IP address, returning a hostname or an error.
-#[allow(clippy::cast_possible_truncation)]
+#[expect(clippy::cast_possible_truncation)]
 fn lookup_addr(addr: IpAddr) -> Result<String, Errno> {
     let addr = match addr {
         IpAddr::V4(v4) => SockaddrStorage::from(SocketAddrV4::new(v4, 0)),
@@ -661,6 +724,7 @@ struct InspectOpts {
 struct AccessMap {
     magic: PandoraMap<String, PandoraSet<String>>,
     force: PandoraSet<String>,
+    fs_types: PandoraSet<String>,
     ioctl: PandoraSet<u64>,
     ioctl_names: StringPatriciaSet,
 }
@@ -684,20 +748,23 @@ impl AccessMap {
             Ok(json) => json,
             Err(err) => {
                 if env::var_os("PANDORA_DEBUG").is_some() {
-                    eprintln!(
+                    let _ = wln!(
+                        stderr(),
                         "{} {}",
                         style("pandora:").bold().magenta(),
                         style("skip invalid JSON!").bold().yellow()
                     );
-                    eprintln!(
+                    let _ = wln!(
+                        stderr(),
                         "\t{} {}",
                         style("LINE:").bold().cyan(),
                         style(line).bold().red()
                     );
-                    eprintln!(
+                    let _ = wln!(
+                        stderr(),
                         "\t{} {}",
                         style("ERROR:").bold().cyan(),
-                        style(err).bold().red()
+                        style(err.to_string()).bold().red()
                     );
                 }
                 return None;
@@ -706,14 +773,9 @@ impl AccessMap {
 
         match json {
             Access::Path {
-                ctx,
-                cap,
-                ctl,
-                path,
-                args,
-                ..
+                ctx, cap, fs, path, ..
             } if ctx == "access" => {
-                let capabilities = match cap {
+                let mut capabilities = match cap {
                     None => return None,
                     Some(Capability::One(cap)) => {
                         let mut caps = PandoraSet::default();
@@ -727,12 +789,9 @@ impl AccessMap {
                     self.force.insert(path.clone());
                 }
 
-                if capabilities.contains("ioctl") {
-                    if let Some(reqs) = ctl {
-                        self.ioctl_names.extend(reqs);
-                    } else if let Some(args) = args {
-                        self.ioctl.insert(args[1]);
-                    }
+                if let Some(fs_type) = fs {
+                    self.fs_types.insert(fs_type);
+                    capabilities.remove("fs");
                 }
 
                 let path = process_path(&path, path_limit).to_string();
@@ -767,6 +826,18 @@ impl AccessMap {
                 let unix = process_path(&unix, u8::MAX).to_string();
                 self.magic.entry(unix).or_default().extend(capabilities);
             }
+            Access::Ioctl { ctx, ctl, .. } if ctx == "access" => {
+                for req in ctl {
+                    match req {
+                        IoctlEntry::Val(val) => {
+                            self.ioctl.insert(val);
+                        }
+                        IoctlEntry::Name(name) => {
+                            self.ioctl_names.insert(name);
+                        }
+                    }
+                }
+            }
             Access::InetAddr { ctx, cap, addr, .. } if ctx == "access" => {
                 let capabilities = match cap {
                     None => return None,
@@ -789,22 +860,26 @@ impl AccessMap {
 
 fn command_profile(opts: ProfileOpts) -> Result<ExitCode, Errno> {
     if Path::new(&opts.output).exists() {
-        eprintln!(
+        wln!(
+            stderr(),
             "{} error creating output file: `{}' already exists!",
             style("pandora:").bold().magenta(),
-            style(opts.output).bold().yellow(),
-        );
+            style(opts.output.clone()).bold().yellow(),
+        )
+        .unwrap();
         return Ok(ExitCode::from(1));
     }
 
     let (fd_rd, fd_rw) = match pipe2(OFlag::O_CLOEXEC) {
         Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
         Err(error) => {
-            eprintln!(
+            wln!(
+                stderr(),
                 "{} error creating pipe: {}!",
                 style("pandora:").bold().magenta(),
-                style(format!("{error}")).bold().red()
-            );
+                style(error.to_string()).bold().red()
+            )
+            .unwrap();
             return Ok(ExitCode::from(1));
         }
     };
@@ -817,7 +892,8 @@ fn command_profile(opts: ProfileOpts) -> Result<ExitCode, Errno> {
     syd.env("SYD_LOG_FD", &log_fd);
     if env::var_os("PANDORA_DEBUG").is_some() {
         let ino = fstatx(&fd_rw, STATX_INO).map(|stx| stx.stx_ino)?;
-        eprintln!(
+        let _ = wln!(
+            stderr(),
             "{} syd log fd set to pipe {} with inode {}.",
             style("pandora:").bold().magenta(),
             style(&log_fd).bold().green(),
@@ -854,7 +930,8 @@ fn command_profile(opts: ProfileOpts) -> Result<ExitCode, Errno> {
             .name("pandora_mon".to_string())
             .spawn(move || {
                 thread::sleep(cmd_timeout);
-                eprintln!(
+                let _ = wln!(
+                    stderr(),
                     "{} {}",
                     style("pandora:").bold().magenta(),
                     style("Timeout expired, terminating process...")
@@ -940,8 +1017,8 @@ fn parse_main_opts() -> Result<MainOpts, lexopt::Error> {
 
     match first_arg {
         None => Ok(MainOpts::Help),
-        Some(Short('h')) => Ok(MainOpts::Help),
-        Some(Short('V')) => Ok(MainOpts::Version),
+        Some(Short('h') | Long("help")) => Ok(MainOpts::Help),
+        Some(Short('V') | Long("version")) => Ok(MainOpts::Version),
         Some(ref arg @ Value(ref cmd)) => match cmd.as_bytes() {
             b"profile" => {
                 let prof = parse_profile_opts(parser)?;
@@ -1021,12 +1098,12 @@ fn parse_profile_opts(mut parser: lexopt::Parser) -> Result<ProfileOpts, lexopt:
     while let Some(arg) = parser.next()? {
         match arg {
             // -h => subcommand help
-            Short('h') => {
+            Short('h') | Long("help") => {
                 print_help_profile();
                 std::process::exit(0);
             }
             // -V => version
-            Short('V') => {
+            Short('V') | Long("version") => {
                 print_version();
                 std::process::exit(0);
             }
@@ -1159,11 +1236,11 @@ fn parse_inspect_opts(mut parser: lexopt::Parser) -> Result<InspectOpts, lexopt:
 
     while let Some(arg) = parser.next()? {
         match arg {
-            Short('h') => {
+            Short('h') | Long("help") => {
                 print_help_inspect();
                 std::process::exit(0);
             }
-            Short('V') => {
+            Short('V') | Long("version") => {
                 print_version();
                 std::process::exit(0);
             }
@@ -1300,7 +1377,7 @@ fn do_inspect(
     }
     let config = config.join("\n");
 
-    // Step 1: Print out the magic header.
+    // Print out the magic header.
     wln!(
         &mut output,
         "#
@@ -1326,8 +1403,16 @@ fn do_inspect(
         ));
     }
 
-    // Step 2: Print out all the sandbox rules from `magic`.
     wln!(&mut output, "###\n# Sandbox Rules\n###")?;
+
+    // Print filesystem sandboxing rules.
+    if !access.fs_types.is_empty() {
+        let mut fs_types = access.fs_types.into_iter().collect::<Vec<_>>();
+        fs_types.sort();
+        wln!(&mut output, "allow/fs+{}\n", fs_types.join(","))?;
+    }
+
+    // Print out all the sandbox rules from `magic`.
     let mut list = Vec::from_iter(access.magic);
     // Alphabetical sort.
     list.sort_by_key(|(path, _)| path.to_string());
@@ -1425,18 +1510,19 @@ fn do_inspect(
         wln!(&mut output, "allow/{}+{}", caps.join(","), elem)?;
     }
 
-    eprintln!(
+    let _ = wln!(
+        stderr(),
         "{} generated {} rules.",
         style("pandora:").bold().magenta(),
         style(len.to_string()).bold().yellow(),
     );
 
-    // Step 3: Print out all ioctl requests.
+    // Print out all ioctl requests.
     if !access.ioctl.is_empty() || !access.ioctl_names.is_empty() {
         wln!(&mut output, "\n###\n# Sandbox ioctl(2) Rules\n###")?;
         wln!(&mut output, "sandbox/ioctl:on\n")?;
 
-        // Step 3.1: Print unknown numeric ioctls.
+        // Print unknown numeric ioctls.
         let mut nums = Vec::with_capacity(access.ioctl.len());
         for &n in access.ioctl.iter() {
             nums.push(n);
@@ -1446,7 +1532,7 @@ fn do_inspect(
         let mut line = String::new();
         for chunk in nums.chunks(5) {
             line.clear();
-            line.push_str("ioctl/allow+");
+            line.push_str("allow/ioctl+");
             for (i, n) in chunk.iter().enumerate() {
                 if i > 0 {
                     line.push(',');
@@ -1456,7 +1542,7 @@ fn do_inspect(
             wln!(&mut output, "{line}")?;
         }
 
-        // Step 3.2: Print named ioctls.
+        // Print named ioctls.
         if !access.ioctl_names.is_empty() {
             // Bucket names by inferred prefix without double-storing keys.
             let mut groups: PandoraMap<String, Vec<String>> = PandoraMap::default();
@@ -1492,7 +1578,7 @@ fn do_inspect(
             for (_k, mut v) in grouped {
                 v.sort_unstable();
                 buf.clear();
-                buf.push_str("ioctl/allow+");
+                buf.push_str("allow/ioctl+");
                 for (i, name) in v.iter().enumerate() {
                     if i > 0 {
                         buf.push(',');
@@ -1504,7 +1590,7 @@ fn do_inspect(
         }
     }
 
-    // Step 4: Print Force entries if available,
+    // Print Force entries if available,
     // concurrency-limited parallel checksums + multiple progress bars.
     if !access.force.is_empty() {
         wln!(&mut output, "\n###\n# Executable Verification\n###")?;
@@ -1559,7 +1645,7 @@ fn do_inspect(
         let rules = Arc::new(Mutex::new(PandoraMap::<PathBuf, String>::default()));
 
         // Spawn concurrency worker threads to do the hashing
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         pool.install(|| {
             pbs.par_iter()
                 .for_each(|(path, pb)| match path2force(path, hash_function, pb) {
@@ -1596,7 +1682,7 @@ fn do_inspect(
         drop(pool);
         let rules = rules.lock().unwrap_or_else(|err| err.into_inner());
 
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         for path in &force {
             let rule = rules
                 .get(path)
@@ -1605,7 +1691,8 @@ fn do_inspect(
         }
         wln!(&mut output)?;
 
-        eprintln!(
+        let _ = wln!(
+            stderr(),
             "{} calculated {} checksums for {} executables.",
             style("pandora:").bold().magenta(),
             style(hash_function.to_string()).bold().cyan(),
@@ -1613,19 +1700,22 @@ fn do_inspect(
         );
     }
 
-    eprintln!(
+    let _ = wln!(
+        stderr(),
         "{} profile generation completed! \\o/",
         style("pandora:").bold().magenta(),
     );
 
-    eprintln!(
+    let _ = wln!(
+        stderr(),
         "{} profile has been written to `{}'.",
         style("pandora:").bold().magenta(),
         style(output_path).bold().yellow(),
     );
 
-    eprintln!(
-        "{} To use it, do: {} -P \"{}\" -- command args...",
+    let _ = wln!(
+        stderr(),
+        "{} To use it, do: {} -P \"{}\" -- {cmd}",
         style("pandora:").bold().magenta(),
         style("syd").bold().green(),
         style(output_path).bold().yellow(),
@@ -1652,7 +1742,7 @@ fn path2force(path: &PathBuf, func: HashAlgorithm, pb: &ProgressBar) -> std::io:
     };
 
     let open_how = safe_open_how(OFlag::O_RDONLY | OFlag::O_NOCTTY);
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let mut file = openat2(AT_FDCWD, path, open_how).map(File::from)?;
 
     let mut buffer = [0u8; 64 * 1024];
@@ -1707,25 +1797,29 @@ fn open_input(input: &str) -> Box<dyn BufRead> {
         "dmesg" | "syslog" => Box::new(BufReader::new(match Syslog::open() {
             Ok(syslog) => syslog,
             Err(err) => {
-                eprintln!(
+                wln!(
+                    stderr(),
                     "{} failed to access syslog: {}!",
                     style("pandora:").bold().magenta(),
-                    style(format!("{err}")).bold().red(),
-                );
+                    style(err.to_string()).bold().red(),
+                )
+                .unwrap();
                 exit(1);
             }
         })),
         path => Box::new(BufReader::new(
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             match OpenOptions::new().read(true).open(path) {
                 Ok(file) => file,
                 Err(err) => {
-                    eprintln!(
+                    wln!(
+                        stderr(),
                         "{} failed to open file {}: {}!",
                         style("pandora:").bold().magenta(),
                         style(path).bold().yellow(),
-                        style(err).bold().red(),
-                    );
+                        style(err.to_string()).bold().red(),
+                    )
+                    .unwrap();
                     exit(1);
                 }
             },
@@ -1738,16 +1832,18 @@ fn open_output(path_or_stdout: &str) -> Box<dyn std::io::Write> {
     match path_or_stdout {
         "-" => Box::new(std::io::BufWriter::new(std::io::stdout())),
         path => Box::new(std::io::BufWriter::new(
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             match OpenOptions::new().write(true).create_new(true).open(path) {
                 Ok(file) => file,
                 Err(err) => {
-                    eprintln!(
+                    wln!(
+                        stderr(),
                         "{} failed to open file {}: {}!",
                         style("pandora:").bold().magenta(),
                         style(path).bold().cyan(),
-                        style(err).bold().red(),
-                    );
+                        style(err.to_string()).bold().red(),
+                    )
+                    .unwrap();
                     exit(1);
                 }
             },
@@ -1989,7 +2085,7 @@ fn is_ascii_printable(byte: u8) -> bool {
 
 /// Converts capability to a priority number for sorting.
 #[inline]
-#[allow(clippy::ptr_arg)]
+#[expect(clippy::ptr_arg)]
 fn cap2prio(cap: &String) -> usize {
     let cap = cap.as_str().as_bytes();
 
@@ -1999,10 +2095,11 @@ fn cap2prio(cap: &String) -> usize {
     }
 }
 
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 fn print_help_main() {
     let nproc = env::var("PANDORA_NPROC").unwrap();
-    eprint!(
+    w!(
+        stderr(),
         r#"{PKG_NAME} {PKG_VERSION}
 {PKG_DESCRIPTION}
 Copyright (c) 2023, 2024, 2025 {PKG_AUTHORS}
@@ -2038,13 +2135,14 @@ Attaching poems encourages consideration tremendously.
 Homepage: https://sydbox.exherbo.org
 Repository: https://gitlab.exherbo.org/sydbox/
 "#,
-    );
+    ).unwrap();
 }
 
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 fn print_help_profile() {
     let nproc = env::var("PANDORA_NPROC").unwrap();
-    eprint!(
+    w!(
+        stderr(),
         r#"{PKG_NAME} {PKG_VERSION}
 Profile subcommand
 
@@ -2062,13 +2160,14 @@ Options:
                sha3-512 (default), sha3-384, sha3-256, sha1, md5, crc64, crc32
   -T <threads> Number of concurrency threads used for parallel hashing [default: {nproc}, env:PANDORA_NPROC]
 "#,
-    );
+    ).unwrap();
 }
 
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 fn print_help_inspect() {
     let nproc = env::var("PANDORA_NPROC").unwrap();
-    eprint!(
+    w!(
+        stderr(),
         r#"{PKG_NAME} {PKG_VERSION}
 Inspect subcommand
 
@@ -2084,11 +2183,11 @@ Options:
                sha3-512 (default), sha3-384, sha3-256, sha1, md5, crc64, crc32
   -T <threads> Number of concurrency threads used for parallel hashing [default: {nproc}, env:PANDORA_NPROC]
 "#,
-    );
+    ).unwrap();
 }
 
 fn print_version() {
-    eprintln!("{PKG_NAME}-{PKG_VERSION}");
+    wln!(stderr(), "{PKG_NAME}-{PKG_VERSION}").unwrap();
 }
 
 // Returns the longest prefix of `s` shared by ≥2 keys in `set`, else `s`.
@@ -2175,7 +2274,7 @@ fn parse_utf8_str(bytes: &[u8]) -> Result<String, Errno> {
     match std::str::from_utf8(bytes) {
         Ok(s) => Ok(s.to_owned()),
         Err(_) => {
-            eprintln!("ERROR: invalid UTF-8 data");
+            wln!(stderr(), "ERROR: invalid UTF-8 data")?;
             Err(Errno::EINVAL)
         }
     }
@@ -2343,3 +2442,20 @@ fn fstatx<Fd: AsFd>(fd: Fd, mask: libc::c_uint) -> Result<FileStatx, Errno> {
 
 // Want/got stx_ino.
 const STATX_INO: libc::c_uint = 0x00000100;
+
+// Retries a closure on `EAGAIN` and `EINTR` errors.
+//
+// This function will call the provided closure, and if the closure
+// returns `EAGAIN` or `EINTR` error, it will retry the operation until it
+// succeeds or fails with a different error.
+fn retry_on_intr<F, T>(mut f: F) -> Result<T, Errno>
+where
+    F: FnMut() -> Result<T, Errno>,
+{
+    loop {
+        match f() {
+            Err(Errno::EAGAIN | Errno::EINTR) => continue,
+            result => return result,
+        }
+    }
+}
