@@ -14,21 +14,20 @@ use std::{
 };
 
 use libseccomp::ScmpNotifResp;
-use nix::{
-    errno::Errno,
-    fcntl::{OFlag, ResolveFlag},
-};
+use nix::{errno::Errno, fcntl::OFlag};
 use serde::{Serialize, Serializer};
 
 use crate::{
+    compat::ResolveFlag,
     config::{PAGE_SIZE, PROC_FILE},
     elf::ExecutableFile,
     error,
-    fs::{fd_status_flags, safe_open_msym, CanonicalPath},
-    hook::UNotifyEventRequest,
+    fs::fd_status_flags,
     kernel::sandbox_path,
+    lookup::{safe_open_msym, CanonicalPath},
     path::XPathBuf,
-    proc::{proc_mem_limit, proc_statm},
+    proc::{proc_mem, proc_stat, proc_statm},
+    req::UNotifyEventRequest,
     sandbox::{Action, Capability, IntegrityError},
     warn,
 };
@@ -84,37 +83,54 @@ impl Serialize for MemSyscall {
 }
 
 pub(crate) fn sys_brk(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_mem_handler(request, MemSyscall::Brk)
+    let req = request.scmpreq;
+    let size = match proc_stat(req.pid()) {
+        Ok(stat) => req.data.args[0].saturating_sub(stat.startbrk),
+        Err(errno) => return request.fail_syscall(errno),
+    };
+    if size == 0 {
+        // SAFETY: System call wants to shrink memory.
+        // No pointer dereference in size check.
+        return unsafe { request.continue_syscall() };
+    }
+    syscall_mem_handler(request, MemSyscall::Brk, size)
 }
 
 pub(crate) fn sys_mmap(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_mem_handler(request, MemSyscall::Mmap)
+    let req = request.scmpreq;
+    syscall_mem_handler(request, MemSyscall::Mmap, req.data.args[1])
 }
 
 pub(crate) fn sys_mmap2(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_mem_handler(request, MemSyscall::Mmap2)
+    let req = request.scmpreq;
+    syscall_mem_handler(request, MemSyscall::Mmap2, req.data.args[1])
 }
 
 pub(crate) fn sys_mremap(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
     let old_size = req.data.args[1];
     let new_size = req.data.args[2];
-    if new_size < old_size {
+    let size = new_size.saturating_sub(old_size);
+    if size == 0 {
         // SAFETY: System call wants to shrink memory.
         // No pointer dereference in size check.
         return unsafe { request.continue_syscall() };
     }
-    syscall_mem_handler(request, MemSyscall::Mremap)
+    syscall_mem_handler(request, MemSyscall::Mremap, size)
 }
 
-#[allow(clippy::cognitive_complexity)]
-fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> ScmpNotifResp {
+#[expect(clippy::cognitive_complexity)]
+fn syscall_mem_handler(
+    request: UNotifyEventRequest,
+    syscall: MemSyscall,
+    size: u64,
+) -> ScmpNotifResp {
     let req = request.scmpreq;
     let caps = syscall.caps();
 
     // Get mem & vm max.
     let sandbox = request.get_sandbox();
-    let verbose = sandbox.verbose;
+    let log_scmp = sandbox.log_scmp();
     let caps = sandbox.getcaps(caps);
     let exec = caps.contains(Capability::CAP_EXEC);
     let force = caps.contains(Capability::CAP_FORCE);
@@ -123,15 +139,15 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
     let mem_max = sandbox.mem_max;
     let mem_vm_max = sandbox.mem_vm_max;
     let mem_act = sandbox.default_action(Capability::CAP_MEM);
-    let restrict_memory = !sandbox.flags.allow_unsafe_memory();
-    let restrict_stack = !sandbox.flags.allow_unsafe_stack();
+    let restrict_exec_memory = !sandbox.flags.allow_unsafe_exec_memory();
+    let restrict_exec_stack = !sandbox.flags.allow_unsafe_exec_stack();
     let restrict_append_only = sandbox.has_append() || sandbox.enabled(Capability::CAP_CRYPT);
 
     if !exec
         && !force
         && !tpe
-        && !restrict_memory
-        && !restrict_stack
+        && !restrict_exec_memory
+        && !restrict_exec_stack
         && !restrict_append_only
         && (!mem || (mem_max == 0 && mem_vm_max == 0))
     {
@@ -143,14 +159,14 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
     let name = syscall.to_string();
 
     let check_exec = syscall.is_mmap()
-        && (exec || force || tpe || restrict_memory || restrict_stack)
+        && (exec || force || tpe || restrict_exec_memory || restrict_exec_stack)
         && req.data.args[2] & PROT_EXEC != 0
         && req.data.args[3] & MAP_ANONYMOUS == 0;
     let check_append_only = restrict_append_only && req.data.args[3] & MAP_SHARED != 0;
 
     // Get the file descriptor before access check.
     let fd = if check_exec || check_append_only {
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         let remote_fd = req.data.args[4] as RawFd;
         if remote_fd < 0 {
             return request.fail_syscall(Errno::EBADF);
@@ -164,8 +180,8 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
         None
     };
 
-    #[allow(clippy::disallowed_methods)]
-    let oflags = if check_append_only || (check_exec && restrict_memory) {
+    #[expect(clippy::disallowed_methods)]
+    let oflags = if check_append_only || (check_exec && restrict_exec_memory) {
         fd_status_flags(fd.as_ref().unwrap()).ok()
     } else {
         None
@@ -190,8 +206,8 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
         // but set as PROT_READ|PROT_EXEC which breaks W^X!
         // We do not need to check for PROT_WRITE here as
         // this is already enforced at kernel-level when
-        // trace/allow_unsafe_memory:1 is not set at startup.
-        if restrict_memory {
+        // trace/allow_unsafe_exec_memory:1 is not set at startup.
+        if restrict_exec_memory {
             let deny = oflags
                 .map(|fl| fl.contains(OFlag::O_RDWR) || fl.contains(OFlag::O_WRONLY))
                 .unwrap_or(true);
@@ -201,7 +217,7 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
             }
         }
 
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         let mut path = match CanonicalPath::new_fd(fd.unwrap().into(), req.pid()) {
             Ok(path) => path,
             Err(errno) => return request.fail_syscall(errno),
@@ -225,12 +241,12 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
         // Step 3: Check for TPE sandboxing.
         if tpe {
             // MUST_PATH ensures path.dir is Some.
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let file = path.dir.as_ref().unwrap();
             let (action, msg) = sandbox.check_tpe(file, path.abs());
             if !matches!(action, Action::Allow | Action::Filter) {
                 let msg = msg.as_deref().unwrap_or("?");
-                if verbose {
+                if log_scmp {
                     error!("ctx": "trusted_path_execution",
                         "msg": format!("library load from untrusted path blocked: {msg}"),
                         "sys": &name, "path": &path,
@@ -257,26 +273,27 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
             }
         }
 
-        if force || restrict_stack {
+        if force || restrict_exec_stack {
             // The following checks require the contents of the file.
             // SAFETY:
             // 1. Reopen the file via `/proc/thread-self/fd` to avoid sharing the file offset.
             // 2. `path` is a remote-fd transfer which asserts `path.dir` is Some.
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let fd = path.dir.take().unwrap();
-            let pfd = XPathBuf::from_self_fd(fd.as_raw_fd());
 
-            let mut file =
-                match safe_open_msym(PROC_FILE(), &pfd, OFlag::O_RDONLY, ResolveFlag::empty())
-                    .map(File::from)
-                {
-                    Ok(file) => file,
-                    Err(_) => {
-                        return request.fail_syscall(Errno::EBADF);
-                    }
-                };
+            let mut file = match XPathBuf::from_self_fd(fd.as_raw_fd())
+                .and_then(|pfd| {
+                    safe_open_msym(PROC_FILE(), &pfd, OFlag::O_RDONLY, ResolveFlag::empty())
+                })
+                .map(File::from)
+            {
+                Ok(file) => file,
+                Err(_) => {
+                    return request.fail_syscall(Errno::EBADF);
+                }
+            };
 
-            if restrict_stack {
+            if restrict_exec_stack {
                 // Step 4: Check for non-executable stack.
                 // An execstack library that is dlopened into an executable
                 // that is otherwise mapped no-execstack can change the
@@ -287,18 +304,18 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
                     let exe = ExecutableFile::parse(&mut *file, true).or(Err(Errno::EACCES))?;
                     if matches!(exe, ExecutableFile::Elf { xs: true, .. }) {
                         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
-                            if verbose {
+                            if log_scmp {
                                 error!("ctx": "check_lib",
                                     "msg": "library load with executable stack blocked",
                                     "sys": &name, "path": path.abs(),
-                                    "tip": "configure `trace/allow_unsafe_stack:1'",
+                                    "tip": "configure `trace/allow_unsafe_exec_stack:1'",
                                     "lib": format!("{exe}"),
                                     "req": &request);
                             } else {
                                 error!("ctx": "check_lib",
                                     "msg": "library load with executable stack blocked",
                                     "sys": &name, "path": path.abs(),
-                                    "tip": "configure `trace/allow_unsafe_stack:1'",
+                                    "tip": "configure `trace/allow_unsafe_exec_stack:1'",
                                     "lib": format!("{exe}"),
                                     "pid": request.scmpreq.pid);
                             }
@@ -316,7 +333,7 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
 
             if force {
                 // Step 5: Check for Force sandboxing.
-                if restrict_stack && file.rewind().is_err() {
+                if restrict_exec_stack && file.rewind().is_err() {
                     drop(sandbox); // release the read-lock.
                     return request.fail_syscall(Errno::EBADF);
                 }
@@ -325,7 +342,7 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
                 let deny = match result {
                     Ok(action) => {
                         if !matches!(action, Action::Allow | Action::Filter) {
-                            if verbose {
+                            if log_scmp {
                                 warn!("ctx": "verify_lib", "act": action,
                                     "sys": &name, "path": path.abs(),
                                     "tip": format!("configure `force+{}:<checksum>'", path.abs()),
@@ -350,7 +367,7 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
                         }
                     }
                     Err(IntegrityError::Sys(errno)) => {
-                        if verbose {
+                        if log_scmp {
                             error!("ctx": "verify_lib",
                                 "msg": format!("system error during library checksum calculation: {errno}"),
                                 "sys": &name, "path": path.abs(),
@@ -371,7 +388,7 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
                         found,
                     }) => {
                         if action != Action::Filter {
-                            if sandbox.verbose {
+                            if log_scmp {
                                 error!("ctx": "verify_lib", "act": action,
                                     "msg": format!("library checksum mismatch: {found} is not {expected}"),
                                     "sys": &name, "path": path.abs(),
@@ -426,17 +443,17 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
             Ok(statm) => statm.size.saturating_mul(*PAGE_SIZE),
             Err(errno) => return request.fail_syscall(errno),
         };
-        if mem_vm_cur >= mem_vm_max {
+        if mem_vm_cur.saturating_add(size) >= mem_vm_max {
             if mem_act != Action::Filter {
-                if verbose {
+                if log_scmp {
                     warn!("ctx": "access", "cap": Capability::CAP_MEM, "act": mem_act,
                         "sys": &name, "mem_vm_max": mem_vm_max, "mem_vm_cur": mem_vm_cur,
-                        "tip": "increase `mem/vm_max'",
+                        "mem_size": size, "tip": "increase `mem/vm_max'",
                         "req": &request);
                 } else {
                     warn!("ctx": "access", "cap": Capability::CAP_MEM, "act": mem_act,
                         "sys": &name, "mem_vm_max": mem_vm_max, "mem_vm_cur": mem_vm_cur,
-                        "tip": "increase `mem/vm_max'",
+                        "mem_size": size, "tip": "increase `mem/vm_max'",
                         "pid": request.scmpreq.pid);
                 }
             }
@@ -457,45 +474,42 @@ fn syscall_mem_handler(request: UNotifyEventRequest, syscall: MemSyscall) -> Scm
 
     // Check PSS
     if mem_max > 0 {
-        match proc_mem_limit(req.pid(), mem_max) {
-            Ok(false) => {
-                // SAFETY: No pointer dereference in security check.
-                unsafe { request.continue_syscall() }
-            }
-            Ok(true) => {
-                if mem_act != Action::Filter {
-                    if verbose {
-                        warn!("ctx": "access", "cap": Capability::CAP_MEM, "act": mem_act,
-                            "sys": &name, "mem_max": mem_max,
-                            "tip": "increase `mem/max'",
-                            "req": &request);
-                    } else {
-                        warn!("ctx": "access", "cap": Capability::CAP_MEM, "act": mem_act,
-                            "sys": &name, "mem_max": mem_max,
-                            "tip": "increase `mem/max'",
-                            "pid": request.scmpreq.pid);
-                    }
-                }
-                match mem_act {
-                    // Allow cannot happen.
-                    Action::Warn => {
-                        // SAFETY: No pointer dereference in security check.
-                        unsafe { request.continue_syscall() }
-                    }
-                    Action::Deny | Action::Filter => request.fail_syscall(Errno::ENOMEM),
-                    Action::Panic => panic!(),
-                    Action::Exit => std::process::exit(libc::ENOMEM),
-                    _ => {
-                        // Stop|Kill
-                        let _ = request.kill(mem_act);
-                        request.fail_syscall(Errno::ENOMEM)
-                    }
+        let mem_cur = match proc_mem(req.pid()) {
+            Ok(mem_cur) => mem_cur,
+            Err(errno) => return request.fail_syscall(errno),
+        };
+        if mem_cur.saturating_add(size) >= mem_max {
+            if mem_act != Action::Filter {
+                if log_scmp {
+                    warn!("ctx": "access", "cap": Capability::CAP_MEM, "act": mem_act,
+                        "sys": &name, "mem_max": mem_max, "mem_cur": mem_cur,
+                        "mem_size": size, "tip": "increase `mem/max'",
+                        "req": &request);
+                } else {
+                    warn!("ctx": "access", "cap": Capability::CAP_MEM, "act": mem_act,
+                        "sys": &name, "mem_max": mem_max, "mem_cur": mem_cur,
+                        "mem_size": size, "tip": "increase `mem/max'",
+                        "pid": request.scmpreq.pid);
                 }
             }
-            Err(errno) => request.fail_syscall(errno),
+            return match mem_act {
+                // Allow cannot happen.
+                Action::Warn => {
+                    // SAFETY: No pointer dereference in security check.
+                    unsafe { request.continue_syscall() }
+                }
+                Action::Deny | Action::Filter => request.fail_syscall(Errno::ENOMEM),
+                Action::Panic => panic!(),
+                Action::Exit => std::process::exit(libc::ENOMEM),
+                _ => {
+                    // Stop|Kill
+                    let _ = request.kill(mem_act);
+                    request.fail_syscall(Errno::ENOMEM)
+                }
+            };
         }
-    } else {
-        // SAFETY: No pointer dereference in security check.
-        unsafe { request.continue_syscall() }
     }
+
+    // SAFETY: No pointer dereference in security check.
+    unsafe { request.continue_syscall() }
 }
