@@ -9,11 +9,13 @@
 use libseccomp::ScmpNotifResp;
 use nix::errno::Errno;
 
-use crate::{config::SHM_UNSAFE_MASK, hook::UNotifyEventRequest, sandbox::Action, warn};
+use crate::{
+    compat::SHM_EXEC, config::SHM_UNSAFE_MASK, req::UNotifyEventRequest, sandbox::Action, warn,
+};
 
 const O_CREAT: u64 = libc::O_CREAT as u64;
 
-const SHM_X: u64 = libc::SHM_EXEC as u64;
+const SHM_X: u64 = SHM_EXEC as u64;
 
 const IPC_SET: u64 = libc::IPC_SET as u64;
 
@@ -109,7 +111,7 @@ fn syscall_shmat_handler(request: UNotifyEventRequest, shmflg: u64) -> ScmpNotif
     request.fail_syscall(Errno::EACCES)
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 fn syscall_msgctl_handler(request: UNotifyEventRequest, op: u64) -> ScmpNotifResp {
     let op = op & 0xff;
     let req = request.scmpreq;
@@ -130,7 +132,7 @@ fn syscall_msgctl_handler(request: UNotifyEventRequest, op: u64) -> ScmpNotifRes
     request.fail_syscall(Errno::EACCES)
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 fn syscall_semctl_handler(request: UNotifyEventRequest, op: u64) -> ScmpNotifResp {
     let op = op & 0xff;
     let req = request.scmpreq;
@@ -151,7 +153,7 @@ fn syscall_semctl_handler(request: UNotifyEventRequest, op: u64) -> ScmpNotifRes
     request.fail_syscall(Errno::EACCES)
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 fn syscall_shmctl_handler(request: UNotifyEventRequest, op: u64) -> ScmpNotifResp {
     let op = op & 0xff;
     let req = request.scmpreq;
@@ -229,7 +231,7 @@ fn syscall_shmget_handler(request: UNotifyEventRequest, flg: u64) -> ScmpNotifRe
     request.fail_syscall(Errno::EACCES)
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub(crate) fn sys_mq_open(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
