@@ -9,29 +9,34 @@
 use std::{
     fs::File,
     io::BufReader,
-    os::fd::{AsFd, AsRawFd},
+    os::{fd::AsFd, unix::ffi::OsStrExt},
 };
 
+use data_encoding::HEXLOWER;
 use libseccomp::ScmpNotifResp;
-use nix::{errno::Errno, NixPath};
+use nix::{errno::Errno, fcntl::AtFlags, NixPath};
 
 use crate::{
-    compat::{fstatat64, statx, STATX_BASIC_STATS, STATX_MODE, STATX_TYPE},
-    config::{MAGIC_PREFIX, MMAP_MIN_ADDR},
-    confine::{scmp_arch_bits, EOWNERDEAD},
-    fs::{is_sidechannel_device, parse_fd, CanonicalPath, FileInfo, FileType, FsFlags},
+    compat::{
+        fstatat64, statx, FileStat64, FileStatx, FileStatxTimestamp, STATX_BASIC_STATS, STATX_MODE,
+        STATX_TYPE,
+    },
+    config::{API_VERSION, ENV_ID, MAGIC_PREFIX},
+    confine::{is_valid_ptr, scmp_arch_bits, EOWNERDEAD},
+    fs::parse_fd,
     hash::SydHashSet,
-    hook::{SysArg, SysFlags, UNotifyEventRequest},
-    kernel::sandbox_path,
+    kernel::{sandbox_path, to_atflags},
+    lookup::{CanonicalPath, FileInfo, FileType, FsFlags},
     path::XPath,
-    sandbox::Capability,
+    req::{SysArg, SysFlags, UNotifyEventRequest},
+    sandbox::{Capability, Flags},
 };
 
 pub(crate) fn sys_stat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -39,6 +44,7 @@ pub(crate) fn sys_stat(request: UNotifyEventRequest) -> ScmpNotifResp {
 
     let arg = SysArg {
         path: Some(0),
+        flags: SysFlags::CHECK_MAGIC,
         fsflags: FsFlags::MUST_PATH,
         ..Default::default()
     };
@@ -50,12 +56,13 @@ pub(crate) fn sys_stat64(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
     let arg = SysArg {
         path: Some(0),
+        flags: SysFlags::CHECK_MAGIC,
         fsflags: FsFlags::MUST_PATH,
         ..Default::default()
     };
@@ -67,7 +74,7 @@ pub(crate) fn sys_fstat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -85,7 +92,7 @@ pub(crate) fn sys_fstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -101,7 +108,7 @@ pub(crate) fn sys_lstat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -109,6 +116,7 @@ pub(crate) fn sys_lstat(request: UNotifyEventRequest) -> ScmpNotifResp {
 
     let arg = SysArg {
         path: Some(0),
+        flags: SysFlags::CHECK_MAGIC,
         fsflags: FsFlags::MUST_PATH | FsFlags::NO_FOLLOW_LAST,
         ..Default::default()
     };
@@ -120,12 +128,13 @@ pub(crate) fn sys_lstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
     let arg = SysArg {
         path: Some(0),
+        flags: SysFlags::CHECK_MAGIC,
         fsflags: FsFlags::MUST_PATH | FsFlags::NO_FOLLOW_LAST,
         ..Default::default()
     };
@@ -136,22 +145,35 @@ pub(crate) fn sys_lstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
 pub(crate) fn sys_statx(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
+    // SAFETY: Reject undefined/invalid flags.
+    // const AT_STATX_SYNC_AS_STAT: AtFlags = AtFlags::empty();
+    const AT_STATX_FORCE_SYNC: AtFlags = AtFlags::from_bits_retain(0x2000);
+    const AT_STATX_DONT_SYNC: AtFlags = AtFlags::from_bits_retain(0x4000);
+    let atflags = match to_atflags(
+        req.data.args[2],
+        AtFlags::AT_EMPTY_PATH
+            | AtFlags::AT_SYMLINK_NOFOLLOW
+            | AtFlags::AT_NO_AUTOMOUNT
+            | AT_STATX_FORCE_SYNC
+            | AT_STATX_DONT_SYNC,
+    ) {
+        Ok(atflags) => atflags,
+        Err(errno) => return request.fail_syscall(errno),
+    };
+
     // Return EFAULT here for invalid pointers.
-    if req.data.args[4] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[4]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
-    let empty = req.data.args[2] & libc::AT_EMPTY_PATH as u64 != 0;
-    let follow = req.data.args[2] & libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
-
     let mut flags = SysFlags::empty();
     let mut fsflags = FsFlags::MUST_PATH;
-
-    if empty {
+    if atflags.contains(AtFlags::AT_EMPTY_PATH) {
         flags |= SysFlags::EMPTY_PATH;
+    } else {
+        flags |= SysFlags::CHECK_MAGIC;
     }
-
-    if !follow {
+    if atflags.contains(AtFlags::AT_SYMLINK_NOFOLLOW) {
         fsflags |= FsFlags::NO_FOLLOW_LAST;
     }
 
@@ -170,20 +192,27 @@ pub(crate) fn sys_newfstatat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[2] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[2]) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
-    let empty = req.data.args[3] & libc::AT_EMPTY_PATH as u64 != 0;
-    let follow = req.data.args[3] & libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
+    // SAFETY: Reject undefined/invalid flags.
+    let atflags = match to_atflags(
+        req.data.args[3],
+        AtFlags::AT_EMPTY_PATH | AtFlags::AT_SYMLINK_NOFOLLOW,
+    ) {
+        Ok(atflags) => atflags,
+        Err(errno) => return request.fail_syscall(errno),
+    };
+
     let mut flags = SysFlags::empty();
     let mut fsflags = FsFlags::MUST_PATH;
-
-    if empty {
+    if atflags.contains(AtFlags::AT_EMPTY_PATH) {
         flags |= SysFlags::EMPTY_PATH;
+    } else {
+        flags |= SysFlags::CHECK_MAGIC;
     }
-
-    if !follow {
+    if atflags.contains(AtFlags::AT_SYMLINK_NOFOLLOW) {
         fsflags |= FsFlags::NO_FOLLOW_LAST;
     }
 
@@ -198,7 +227,7 @@ pub(crate) fn sys_newfstatat(request: UNotifyEventRequest) -> ScmpNotifResp {
     syscall_stat_handler(request, arg, 2, false)
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 fn syscall_stat_handler(
     request: UNotifyEventRequest,
     arg: SysArg,
@@ -207,24 +236,30 @@ fn syscall_stat_handler(
 ) -> ScmpNotifResp {
     syscall_handler!(request, |request: UNotifyEventRequest| {
         let req = request.scmpreq;
-        let is_fd = arg.path.is_none();
         let sandbox = request.get_sandbox();
-        let is_lock = sandbox.locked_for(req.pid());
 
-        let is_crypt = sandbox.enabled(Capability::CAP_CRYPT);
-        let is_stat = sandbox.enabled(Capability::CAP_STAT);
+        let has_crypt = sandbox.enabled(Capability::CAP_CRYPT);
+        let has_stat = sandbox.enabled(Capability::CAP_STAT);
+
+        // Read the remote path and check for magic path as necessary.
+        let (mut path, magic, empty_path) = request.read_path(&sandbox, arg)?;
+        let is_fd = empty_path || arg.path.is_none();
 
         // Check for chroot.
         if sandbox.is_chroot() {
             return Err(if is_fd { Errno::EACCES } else { Errno::ENOENT });
         }
 
-        // Read the remote path.
-        // If lock is on do not check for magic path.
-        let (mut path, magic) = request.read_path(&sandbox, arg, !is_lock)?;
-
+        let restrict_stat_bdev = !sandbox.flags.allow_unsafe_stat_bdev();
+        let restrict_stat_cdev = !sandbox.flags.allow_unsafe_stat_cdev();
         let mut ghost = false;
-        if !is_lock && magic {
+        let caps = *sandbox.state;
+        let sflags = *sandbox.flags;
+        if magic {
+            if sandbox.locked_for(req.pid()) {
+                // Sandbox is locked, access denied.
+                return Err(Errno::ENOENT);
+            }
             drop(sandbox); // release the read-lock.
 
             // Handle magic prefix (ie /dev/syd)
@@ -290,40 +325,42 @@ fn syscall_stat_handler(
             drop(sandbox); // release the write-lock.
         } else {
             // Handle fstat for files with encryption in progress.
-            let mut crypt_stat = false;
-            if is_crypt && is_fd {
+            #[expect(clippy::disallowed_methods)]
+            if is_fd && has_crypt {
                 // SAFETY: SysArg.path is None asserting dirfd is Some fd!=AT_FDCWD.
-                #[allow(clippy::disallowed_methods)]
                 let fd = path.dir.as_ref().unwrap();
+                // SAFETY: has_crypt asserts crypt_map is Some.
+                let files = request.cache.crypt_map.as_ref().unwrap();
+
                 if let Ok(info) = FileInfo::from_fd(fd) {
-                    #[allow(clippy::disallowed_methods)]
-                    let files = request.crypt_map.as_ref().unwrap();
-                    for (enc_path, map) in
-                        files.read().unwrap_or_else(|err| err.into_inner()).iter()
-                    {
+                    let files = files.0.lock().unwrap_or_else(|err| err.into_inner());
+                    for (enc_path, map) in files.iter() {
                         if info == map.info {
                             // Found underlying encrypted file for the memory fd.
                             // Note, we only ever attempt to encrypt regular files.
-                            path =
-                                CanonicalPath::new(enc_path.clone(), FileType::Reg, arg.fsflags)?;
-                            crypt_stat = true;
+                            path = CanonicalPath::new_crypt(
+                                path.dir.take().unwrap(),
+                                enc_path.clone(),
+                            );
                             break;
                         }
                     }
-                }
+                } // Lock is released here.
+            }
+
+            // Return correct stat information for !memfd:syd/ paths.
+            // This prefix is internal to Syd and sandbox process cannot
+            // create memory file descriptors with this name prefix.
+            if is_fd && path.is_memory_fd() && path.abs().starts_with(b"!memfd:syd") {
+                let mut p = path.take();
+                p.drain(0..b"!memfd:syd".len());
+                path = CanonicalPath::new_mask(&p, &p)?;
             }
 
             // SAFETY:
             // 1. Allow access to fd-only calls.
             // 2. Allow access to files with encryption in progress.
-            // 3. Allow access to /memfd:syd-*. This prefix is internal
-            //    to Syd and sandbox process cannot create memory file
-            //    descriptors with this name prefix.
-            if is_stat
-                && !crypt_stat
-                && arg.path.is_some()
-                && !path.abs().starts_with(b"/memfd:syd-")
-            {
+            if has_stat && arg.path.is_some() {
                 sandbox_path(
                     Some(&request),
                     &sandbox,
@@ -357,7 +394,7 @@ fn syscall_stat_handler(
         let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
         let mut flags = libc::AT_EMPTY_PATH;
 
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         if arg_stat == 4 {
             // statx
 
@@ -387,19 +424,24 @@ fn syscall_stat_handler(
             }
 
             // Note, unlike statfs, stat does not EINTR.
-            let mut statx = statx(fd, "", flags, mask)?;
+            let mut statx = statx(fd, c"", flags, mask)?;
 
             // SAFETY: Check if the file is a sidechannel device and
             // update its access and modification times to match the
             // creation time if it is. This prevents timing attacks on
             // block or character devices like /dev/ptmx using stat.
-            if is_sidechannel_device(statx.stx_mode.into()) {
-                statx.stx_atime = statx.stx_ctime;
-                statx.stx_mtime = statx.stx_ctime;
+            if restrict_stat_bdev || restrict_stat_cdev {
+                let filetype = FileType::from(libc::mode_t::from(statx.stx_mode));
+                if (restrict_stat_bdev && filetype.is_block_device())
+                    || (restrict_stat_cdev && filetype.is_char_device())
+                {
+                    statx.stx_atime = statx.stx_ctime;
+                    statx.stx_mtime = statx.stx_ctime;
+                }
             }
 
             // SAFETY: Restore mask, type and mode, see the comment above.
-            #[allow(clippy::cast_possible_truncation)]
+            #[expect(clippy::cast_possible_truncation)]
             if !basic_stx {
                 if (orig_mask & STATX_TYPE) == 0 {
                     statx.stx_mode &= !libc::S_IFMT as u16;
@@ -411,10 +453,15 @@ fn syscall_stat_handler(
                 }
             }
 
+            // If magic path, mask values for easy id.
+            if magic {
+                magic_statx(&mut statx, caps, sflags);
+            }
+
             // SAFETY: The following block creates an immutable byte
             // slice representing the memory of `statx`. We ensure that
             // the slice covers the entire memory of `statx` using
-            // `std::mem::size_of_val`. Since `statx` is a stack
+            // `size_of_val`. Since `statx` is a stack
             // variable and we're only borrowing its memory for the
             // duration of the slice, there's no risk of `statx` being
             // deallocated while the slice exists. Additionally, we
@@ -423,7 +470,7 @@ fn syscall_stat_handler(
             let statx = unsafe {
                 std::slice::from_raw_parts(
                     std::ptr::addr_of!(statx) as *const u8,
-                    std::mem::size_of_val(&statx),
+                    size_of_val(&statx),
                 )
             };
             let addr = req.data.args[4];
@@ -435,17 +482,27 @@ fn syscall_stat_handler(
 
             // SAFETY: In libc we trust.
             // Note, unlike statfs, stat does not EINTR.
-            let mut stat = fstatat64(Some(fd.as_raw_fd()), "", flags)?;
+            let mut stat = fstatat64(fd, c"", flags)?;
 
             // SAFETY: Check if the file is a sidechannel device and
             // update its access and modification times to match the
             // creation time if it is. This prevents timing attacks on
             // block or character devices like /dev/ptmx using stat.
-            if is_sidechannel_device(stat.st_mode) {
-                stat.st_atime = stat.st_ctime;
-                stat.st_mtime = stat.st_ctime;
-                stat.st_atime_nsec = stat.st_ctime_nsec;
-                stat.st_mtime_nsec = stat.st_ctime_nsec;
+            if restrict_stat_bdev || restrict_stat_cdev {
+                let filetype = FileType::from(stat.st_mode);
+                if (restrict_stat_bdev && filetype.is_block_device())
+                    || (restrict_stat_cdev && filetype.is_char_device())
+                {
+                    stat.st_atime = stat.st_ctime;
+                    stat.st_mtime = stat.st_ctime;
+                    stat.st_atime_nsec = stat.st_ctime_nsec;
+                    stat.st_mtime_nsec = stat.st_ctime_nsec;
+                }
+            }
+
+            // If magic path, mask values for easy id.
+            if magic {
+                magic_stat(&mut stat, caps, sflags);
             }
 
             let addr = req.data.args[arg_stat];
@@ -456,7 +513,7 @@ fn syscall_stat_handler(
                     // SAFETY: The following block creates an immutable
                     // byte slice representing the memory of `stat`.  We
                     // ensure that the slice covers the entire memory of
-                    // `stat` using `std::mem::size_of_val`. Since
+                    // `stat` using `size_of_val`. Since
                     // `stat` is a stack variable and we're only
                     // borrowing its memory for the duration of the
                     // slice, there's no risk of `stat` being
@@ -466,7 +523,7 @@ fn syscall_stat_handler(
                     let stat = unsafe {
                         std::slice::from_raw_parts(
                             std::ptr::addr_of!(stat32) as *const u8,
-                            std::mem::size_of_val(&stat32),
+                            size_of_val(&stat32),
                         )
                     };
                     request.write_mem(stat, addr)?;
@@ -474,7 +531,7 @@ fn syscall_stat_handler(
                     // SAFETY: The following block creates an immutable
                     // byte slice representing the memory of `stat`.  We
                     // ensure that the slice covers the entire memory of
-                    // `stat` using `std::mem::size_of_val`. Since
+                    // `stat` using `size_of_val`. Since
                     // `stat` is a stack variable and we're only
                     // borrowing its memory for the duration of the
                     // slice, there's no risk of `stat` being
@@ -484,7 +541,7 @@ fn syscall_stat_handler(
                     let stat = unsafe {
                         std::slice::from_raw_parts(
                             std::ptr::addr_of!(stat) as *const u8,
-                            std::mem::size_of_val(&stat),
+                            size_of_val(&stat),
                         )
                     };
                     request.write_mem(stat, addr)?;
@@ -503,3 +560,105 @@ fn syscall_stat_handler(
         Ok(request.return_syscall(0))
     })
 }
+
+fn magic_stat(stat: &mut FileStat64, caps: Capability, flags: Flags) {
+    stat.st_ino = magic_ino();
+    stat.st_nlink = caps.nlink().into();
+    stat.st_mode = magic_mode(caps, flags).into();
+    stat.st_rdev = API_VERSION.dev();
+    stat.st_atime = 505958400; // must match libsyd!
+    stat.st_ctime = -2036448000; // ditto!
+    stat.st_mtime = -842745600; // ditto!
+}
+
+fn magic_statx(statx: &mut FileStatx, caps: Capability, flags: Flags) {
+    statx.stx_ino = magic_ino();
+    statx.stx_nlink = caps.nlink();
+    statx.stx_mode = magic_mode(caps, flags);
+    statx.stx_rdev_major = API_VERSION.major().into();
+    statx.stx_rdev_minor = API_VERSION.minor().into();
+    statx.stx_atime = FileStatxTimestamp {
+        tv_sec: 505958400, // must match libsyd!
+        ..Default::default()
+    };
+    statx.stx_ctime = FileStatxTimestamp {
+        tv_sec: -2036448000, // ditto!
+        ..Default::default()
+    };
+    statx.stx_mtime = FileStatxTimestamp {
+        tv_sec: -842745600, // ditto!
+        ..Default::default()
+    };
+}
+
+// Inode = `xxd -r -p < /etc/machine-id | od -An -t u8 | head -n1`
+fn magic_ino() -> u64 {
+    let id = if let Some(id) = std::env::var_os(ENV_ID) {
+        id
+    } else {
+        return 0;
+    };
+
+    if id.len() < 16 {
+        return 0;
+    }
+
+    let id = &id.as_bytes()[..16];
+    let id = if let Ok(id) = HEXLOWER.decode(id) {
+        id
+    } else {
+        return 0;
+    };
+
+    assert_eq!(id.len(), 8);
+    u64::from_ne_bytes(id.try_into().unwrap_or_default())
+}
+
+#[allow(clippy::cognitive_complexity)]
+#[allow(clippy::cast_possible_truncation)]
+fn magic_mode(caps: Capability, flags: Flags) -> u16 {
+    // Start with file type = character device
+    let mut mode: u16 = libc::S_IFCHR as u16;
+
+    // Special bits
+    if flags.contains(Flags::FL_UNSHARE_MOUNT) {
+        mode |= libc::S_ISVTX as u16;
+    }
+    if flags.contains(Flags::FL_UNSHARE_USER) {
+        mode |= libc::S_ISUID as u16;
+    }
+    if flags.contains(Flags::FL_UNSHARE_NET) {
+        mode |= libc::S_ISGID as u16;
+    }
+
+    // Permission bits for owner/group/other
+    if caps.contains(Capability::CAP_READ) {
+        mode |= libc::S_IRUSR as u16;
+    }
+    if caps.contains(Capability::CAP_WRITE) {
+        mode |= libc::S_IWUSR as u16;
+    }
+    if caps.contains(Capability::CAP_EXEC) {
+        mode |= libc::S_IXUSR as u16;
+    }
+    if caps.contains(Capability::CAP_STAT) {
+        mode |= libc::S_IRGRP as u16;
+    }
+    if caps.contains(Capability::CAP_PROXY) {
+        mode |= libc::S_IWGRP as u16;
+    }
+    if caps.contains(Capability::CAP_TPE) {
+        mode |= libc::S_IXGRP as u16;
+    }
+    if caps.contains(Capability::CAP_LOCK) {
+        mode |= libc::S_IROTH as u16;
+    }
+    if caps.contains(Capability::CAP_CRYPT) {
+        mode |= libc::S_IWOTH as u16;
+    }
+    if caps.contains(Capability::CAP_FORCE) {
+        mode |= libc::S_IXOTH as u16;
+    }
+
+    mode
+}
