@@ -12,9 +12,9 @@ use libseccomp::ScmpNotifResp;
 use nix::{errno::Errno, unistd::symlinkat, NixPath};
 
 use crate::{
-    fs::FsFlags,
-    hook::{RemoteProcess, SysArg, UNotifyEventRequest},
     kernel::sandbox_path,
+    lookup::FsFlags,
+    req::{RemoteProcess, SysArg, UNotifyEventRequest},
     sandbox::Capability,
 };
 
@@ -56,24 +56,25 @@ fn syscall_symlink_handler(
 
     let process = RemoteProcess::new(request.scmpreq.pid());
 
-    // SAFETY: Read remote path, process will be validated
-    // using the seccomp request ID before use.
-    let target = unsafe { process.remote_path(req.data.args[0]) }?;
-    if !request.is_valid() {
-        return Err(Errno::ESRCH);
-    }
+    // Read remote path, request will be validated by remote_path.
+    let target = process.remote_path(req.data.args[0], Some(&request))?;
 
-    // SAFETY: symlink() returns ENOENT if target is an empty string.
+    // symlink() returns ENOENT if target is an empty string.
     if target.is_empty() {
         return Err(Errno::ENOENT);
     }
 
     // Read remote path.
     let sandbox = request.get_sandbox();
-    let (path, _) = request.read_path(&sandbox, arg, false)?;
+    let (path, _, _) = request.read_path(&sandbox, arg)?;
 
     // Check for access.
     let hide = sandbox.enabled(Capability::CAP_STAT);
+    let name = if arg.dirfd.is_some() {
+        "symlinkat"
+    } else {
+        "symlink"
+    };
     sandbox_path(
         Some(&request),
         &sandbox,
@@ -81,10 +82,16 @@ fn syscall_symlink_handler(
         path.abs(),
         Capability::CAP_SYMLINK,
         hide,
-        "symlink",
+        name,
     )?;
     drop(sandbox); // release the read-lock.
 
+    // SAFETY: Path hiding is done:
+    // Now it is safe to return EEXIST if linkpath exists.
+    if path.base.is_empty() {
+        return Err(Errno::EEXIST);
+    }
+
     // All done, call underlying system call.
     symlinkat(
         &target,
