@@ -7,22 +7,27 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
-    ffi::CStr,
+    borrow::Cow,
     mem::MaybeUninit,
     os::fd::{AsFd, AsRawFd},
 };
 
 use libseccomp::ScmpNotifResp;
-use nix::{errno::Errno, fcntl::AtFlags, unistd::fchdir, NixPath};
+use nix::{errno::Errno, fcntl::AtFlags, NixPath};
 
 use crate::{
-    compat::{getxattrat, listxattrat, removexattrat, setxattrat, XattrArgs},
+    compat::{
+        getxattrat, listxattrat, removexattrat, setxattrat, XattrArgs, XATTR_LIST_MAX,
+        XATTR_SIZE_MAX,
+    },
     config::ROOT_FILE,
-    fs::{denyxattr, filterxattr, FsFlags},
-    hook::{PathArgs, SysArg, SysFlags, UNotifyEventRequest},
+    cookie::safe_fchdir,
     kernel::{syscall_path_handler, to_atflags},
+    lookup::FsFlags,
     path::XPath,
+    req::{PathArgs, SysArg, SysFlags, UNotifyEventRequest},
     sandbox::SandboxGuard,
+    xattr::{denyxattr, filterxattr},
 };
 
 pub(crate) fn sys_getxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -69,29 +74,21 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         // SAFETY:
         // 1. SysArg has one element.
         // 2. SysArg.path is None asserting dir is Some.
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         let fd = path_args.0.as_ref().unwrap().dir.as_ref().unwrap();
 
         let req = request.scmpreq;
         let name = if req.data.args[1] != 0 {
-            const SIZ: usize = libc::PATH_MAX as usize;
-            let mut buf = Vec::new();
-            buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-            buf.resize(SIZ, 0);
-            request.read_mem(&mut buf, req.data.args[1])?;
-            Some(buf)
+            Some(request.read_xattr(req.data.args[1])?)
         } else {
             None
         };
-        let name = if let Some(ref name) = name {
-            CStr::from_bytes_until_nul(name)
-                .or(Err(Errno::E2BIG))?
-                .as_ptr()
-        } else {
-            std::ptr::null()
-        };
+        let name = name
+            .as_ref()
+            .map(|cstr| cstr.as_ptr())
+            .unwrap_or(std::ptr::null());
 
-        if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+        if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
             // SAFETY: Deny user.syd* extended attributes. name is either
             // NULL or a valid nul-terminated C-String.
             // SAFETY: Deny with ENODATA for stealth.
@@ -102,12 +99,10 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         // SAFETY: The size argument to the getxattr call
         // must not be fully trusted, it can be overly large,
         // and allocating a Vector of that capacity may overflow.
-        let len = usize::try_from(req.data.args[3]).or(Err(Errno::E2BIG))?;
-        let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX
+        let len = to_len_cap(req.data.args[3], XATTR_SIZE_MAX)?;
         let mut buf = if len > 0 {
-            let mut buf = Vec::new();
+            let mut buf: Vec<u8> = Vec::new();
             buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-            buf.resize(len, 0);
             Some(buf)
         } else {
             None
@@ -116,25 +111,31 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             Some(b) => b.as_mut_ptr(),
             None => std::ptr::null_mut(),
         };
+
+        #[expect(clippy::cast_sign_loss)]
         // SAFETY: In libc we trust.
-        let n = unsafe {
-            libc::fgetxattr(
-                fd.as_raw_fd(),
-                name,
-                ptr as *mut _ as *mut libc::c_void,
-                len,
-            )
+        let mut n = match Errno::result(unsafe {
+            libc::fgetxattr(fd.as_raw_fd(), name, ptr.cast(), len)
+        }) {
+            Ok(n) => n as usize,
+            Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+                // SAFETY: Avoid a well-behaving process from
+                // repeating calls to potentially exhaust memory.
+                // See tar's tests for an example.
+                return Err(Errno::E2BIG);
+            }
+            Err(errno) => return Err(errno),
         };
-        #[allow(clippy::cast_sign_loss)]
-        let n = if n == -1 {
-            return Err(Errno::last());
-        } else {
-            n as usize
-        };
-        if let Some(buf) = buf {
-            request.write_mem(&buf[..n], req.data.args[2])?;
+
+        if let Some(mut buf) = buf {
+            // SAFETY: n was returned by the host Linux kernel.
+            unsafe { buf.set_len(n) };
+            buf.shrink_to_fit();
+
+            n = request.write_mem(&buf, req.data.args[2])?;
         }
-        #[allow(clippy::cast_possible_wrap)]
+
+        #[expect(clippy::cast_possible_wrap)]
         Ok(request.return_syscall(n as i64))
     })
 }
@@ -217,7 +218,7 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
     }];
     syscall_path_handler(request, "lsetxattr", argv, |path_args, request, sandbox| {
         // SAFETY: SysArg has one element.
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         let path = path_args.0.as_ref().unwrap();
 
         let base = if path.base.is_empty() {
@@ -229,24 +230,16 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         let req = request.scmpreq;
 
         let name = if req.data.args[1] != 0 {
-            const SIZ: usize = libc::PATH_MAX as usize;
-            let mut buf = Vec::new();
-            buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-            buf.resize(SIZ, 0);
-            request.read_mem(&mut buf, req.data.args[1])?;
-            Some(buf)
+            Some(request.read_xattr(req.data.args[1])?)
         } else {
             None
         };
-        let name = if let Some(ref name) = name {
-            CStr::from_bytes_until_nul(name)
-                .or(Err(Errno::E2BIG))?
-                .as_ptr()
-        } else {
-            std::ptr::null()
-        };
+        let name = name
+            .as_ref()
+            .map(|cstr| cstr.as_ptr())
+            .unwrap_or(std::ptr::null());
 
-        if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+        if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
             // SAFETY: Deny user.syd* extended attributes. name is either
             // NULL or a valid nul-terminated C-String.
             // SAFETY: Deny with ENODATA for stealth.
@@ -260,27 +253,24 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         let (buf, len) = if req.data.args[3] == 0 {
             (None, 0)
         } else {
-            let len = usize::try_from(req.data.args[3]).or(Err(Errno::E2BIG))?;
-            let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-            let mut buf = Vec::new();
-            buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-            buf.resize(len, 0);
-            request.read_mem(&mut buf, req.data.args[2])?;
+            let len = to_len_val(req.data.args[3], XATTR_SIZE_MAX)?;
+            let buf = request.read_vec(req.data.args[2], len)?;
+            let len = buf.len();
             (Some(buf), len)
         };
         let buf = buf.as_ref().map_or(std::ptr::null(), |b| b.as_ptr()) as *const libc::c_void;
 
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         let flags = req.data.args[4] as libc::c_int;
 
         match &path.dir {
             Some(fd) => {
                 // SAFETY: We use fchdir which is TOCTOU-free!
-                fchdir(fd)?;
+                safe_fchdir(fd)?;
             }
             None => {
                 // SAFETY: `/` is never a symlink!
-                fchdir(ROOT_FILE())?;
+                safe_fchdir(ROOT_FILE())?;
             }
         };
 
@@ -350,7 +340,7 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             // SAFETY:
             // 1. SysArg has one element.
             // 2. SysArg.path is None asserting dir is Some.
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let fd = path_args.0.as_ref().unwrap().dir.as_ref().unwrap();
 
             let req = request.scmpreq;
@@ -358,12 +348,10 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             // SAFETY: The size argument to the flistxattr call
             // must not be fully trusted, it can be overly large,
             // and allocating a Vector of that capacity may overflow.
-            let len = usize::try_from(req.data.args[2]).or(Err(Errno::E2BIG))?;
-            let len = len.min(10240); // Cap count at 10240.
+            let len = to_len_cap(req.data.args[2], XATTR_LIST_MAX)?;
             let mut buf = if len > 0 {
                 let mut buf = Vec::new();
                 buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-                buf.resize(len, 0);
                 Some(buf)
             } else {
                 None
@@ -373,29 +361,37 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
                 .map_or(std::ptr::null_mut(), |b| b.as_mut_ptr())
                 as *mut libc::c_char;
 
+            #[expect(clippy::cast_sign_loss)]
             // SAFETY: In libc we trust.
-            let n = unsafe { libc::flistxattr(fd.as_raw_fd(), ptr, len) };
-            #[allow(clippy::cast_sign_loss)]
-            let n = if n == -1 {
-                return Err(Errno::last());
-            } else {
-                n as usize
+            let mut n = match Errno::result(unsafe { libc::flistxattr(fd.as_raw_fd(), ptr, len) }) {
+                Ok(n) => n as usize,
+                Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
+                    // SAFETY: Avoid a well-behaving process from
+                    // repeating calls to potentially exhaust memory.
+                    // See tar's tests for an example.
+                    return Err(Errno::E2BIG);
+                }
+                Err(errno) => return Err(errno),
             };
-            let n = if let Some(buf) = buf {
-                // SAFETY: Filter out attributes that start with "user.syd".
-                // SAFETY: Deny only if the Sandbox is locked for the process.
-                let buf = if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
-                    filterxattr(&buf[..n], n)?
+
+            if let Some(mut buf) = buf {
+                // SAFETY: n was returned by the host Linux kernel.
+                unsafe { buf.set_len(n) };
+                buf.shrink_to_fit();
+
+                // SAFETY:
+                // Filter out attributes that start with "user.syd".
+                // Deny only if the Sandbox is locked for the process.
+                let buf = if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+                    Cow::Owned(filterxattr(&buf, n)?)
                 } else {
-                    buf
+                    Cow::Borrowed(&buf)
                 };
 
-                request.write_mem(&buf, req.data.args[1])?;
-                buf.len()
-            } else {
-                n
-            };
-            #[allow(clippy::cast_possible_wrap)]
+                n = request.write_mem(&buf, req.data.args[1])?;
+            }
+
+            #[expect(clippy::cast_possible_wrap)]
             Ok(request.return_syscall(n as i64))
         },
     )
@@ -525,7 +521,7 @@ pub(crate) fn sys_lremovexattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         argv,
         |path_args, request, sandbox| {
             // SAFETY: SysArg has one element.
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let path = path_args.0.as_ref().unwrap();
 
             let base = if path.base.is_empty() {
@@ -537,24 +533,16 @@ pub(crate) fn sys_lremovexattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             let req = request.scmpreq;
 
             let name = if req.data.args[1] != 0 {
-                const SIZ: usize = libc::PATH_MAX as usize;
-                let mut buf = Vec::new();
-                buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-                buf.resize(SIZ, 0);
-                request.read_mem(&mut buf, req.data.args[1])?;
-                Some(buf)
+                Some(request.read_xattr(req.data.args[1])?)
             } else {
                 None
             };
-            let name = if let Some(ref name) = name {
-                CStr::from_bytes_until_nul(name)
-                    .or(Err(Errno::E2BIG))?
-                    .as_ptr()
-            } else {
-                std::ptr::null()
-            };
+            let name = name
+                .as_ref()
+                .map(|cstr| cstr.as_ptr())
+                .unwrap_or(std::ptr::null());
 
-            if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+            if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
                 // SAFETY: Deny user.syd* extended attributes.
                 // name is either NULL or a valid nul-terminated C-String.
                 // SAFETY: Deny with ENODATA for stealth.
@@ -565,11 +553,11 @@ pub(crate) fn sys_lremovexattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             match &path.dir {
                 Some(fd) => {
                     // SAFETY: We use fchdir which is TOCTOU-free!
-                    fchdir(fd)?
+                    safe_fchdir(fd)?
                 }
                 None => {
                     // SAFETY: `/` is never a symlink!
-                    fchdir(ROOT_FILE())?;
+                    safe_fchdir(ROOT_FILE())?;
                 }
             };
 
@@ -632,7 +620,7 @@ fn syscall_getxattr_handler(
     let req = request.scmpreq;
 
     // SAFETY: SysArg has one element.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
     let base = if path.base.is_empty() {
@@ -642,24 +630,16 @@ fn syscall_getxattr_handler(
     };
 
     let name = if req.data.args[1] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        request.read_mem(&mut buf, req.data.args[1])?;
-        Some(buf)
+        Some(request.read_xattr(req.data.args[1])?)
     } else {
         None
     };
-    let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
-    } else {
-        std::ptr::null()
-    };
+    let name = name
+        .as_ref()
+        .map(|cstr| cstr.as_ptr())
+        .unwrap_or(std::ptr::null());
 
-    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+    if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -670,14 +650,10 @@ fn syscall_getxattr_handler(
     // SAFETY: The size argument to the getxattr call
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    #[allow(clippy::cast_possible_truncation)]
-    let len = req.data.args[3] as usize;
-    let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-
+    let len = to_len_cap(req.data.args[3], XATTR_SIZE_MAX)?;
     let mut buf = if len > 0 {
-        let mut buf = Vec::new();
+        let mut buf: Vec<u8> = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -691,23 +667,36 @@ fn syscall_getxattr_handler(
     match &path.dir {
         Some(fd) => {
             // SAFETY: We use fchdir which is TOCTOU-free!
-            fchdir(fd)?;
+            safe_fchdir(fd)?;
         }
-        None => fchdir(ROOT_FILE())?,
+        None => safe_fchdir(ROOT_FILE())?,
     };
 
     let res = base
         // SAFETY: We do not resolve symbolic links here!
         .with_nix_path(|cstr| unsafe { libc::lgetxattr(cstr.as_ptr(), name, ptr.cast(), len) })?;
 
-    #[allow(clippy::cast_sign_loss)]
-    let n = Errno::result(res)? as usize;
+    #[expect(clippy::cast_sign_loss)]
+    let mut n = match Errno::result(res) {
+        Ok(n) => n as usize,
+        Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
-    if let Some(buf) = buf {
-        request.write_mem(&buf[..n], req.data.args[2])?;
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        n = request.write_mem(&buf, req.data.args[2])?;
     }
 
-    #[allow(clippy::cast_possible_wrap)]
+    #[expect(clippy::cast_possible_wrap)]
     Ok(request.return_syscall(n as i64))
 }
 
@@ -720,7 +709,7 @@ fn syscall_getxattrat_handler(
     let req = request.scmpreq;
 
     // SAFETY: SysArg has one element.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
     let base = if path.base.is_empty() {
@@ -730,10 +719,10 @@ fn syscall_getxattrat_handler(
     };
 
     // Read struct xattr_args which holds the return pointer, buffer size and flags.
-    let mut args = MaybeUninit::<XattrArgs>::uninit();
+    let mut args = MaybeUninit::<XattrArgs>::zeroed();
 
     // SAFETY: Ensure size of XattrArgs matches with user argument.
-    if req.data.args[5] != std::mem::size_of::<XattrArgs>() as u64 {
+    if req.data.args[5] != size_of::<XattrArgs>() as u64 {
         return Err(Errno::EINVAL);
     }
 
@@ -741,14 +730,11 @@ fn syscall_getxattrat_handler(
     // We don't read uninitialized memory, and after `read_mem` fills it,
     // we're good to assume it's valid.
     let buf = unsafe {
-        std::slice::from_raw_parts_mut(
-            args.as_mut_ptr().cast::<u8>(),
-            std::mem::size_of::<XattrArgs>(),
-        )
+        std::slice::from_raw_parts_mut(args.as_mut_ptr().cast::<u8>(), size_of::<XattrArgs>())
     };
 
     // Read the remote data structure.
-    request.read_mem(buf, req.data.args[4])?;
+    request.read_mem(buf, req.data.args[4], buf.len())?;
 
     // SAFETY: read_mem() has initialized `args` if it succeeded.
     let args = unsafe { args.assume_init() };
@@ -759,24 +745,16 @@ fn syscall_getxattrat_handler(
     }
 
     let name = if req.data.args[3] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        request.read_mem(&mut buf, req.data.args[3])?;
-        Some(buf)
+        Some(request.read_xattr(req.data.args[3])?)
     } else {
         None
     };
-    let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
-    } else {
-        std::ptr::null()
-    };
+    let name = name
+        .as_ref()
+        .map(|cstr| cstr.as_ptr())
+        .unwrap_or(std::ptr::null());
 
-    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+    if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -787,14 +765,10 @@ fn syscall_getxattrat_handler(
     // SAFETY: The size element of the struct xattr_args
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    #[allow(clippy::cast_possible_truncation)]
-    let len = args.size as usize;
-    let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-
+    let len = to_len_cap(args.size.into(), XATTR_SIZE_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -805,7 +779,7 @@ fn syscall_getxattrat_handler(
         None => return Err(Errno::EBADF),
     };
 
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let mut my_args = XattrArgs {
         value: match buf.as_mut() {
             Some(b) => b.as_mut_ptr() as *mut libc::c_void as u64,
@@ -818,13 +792,27 @@ fn syscall_getxattrat_handler(
     // SAFETY:
     // 1. `name` is a valid raw pointer (may be NULL)!
     // 2. We do not resolve symbolic links here!
-    let n = unsafe { getxattrat(fd, base, name, &mut my_args, AtFlags::AT_SYMLINK_NOFOLLOW) }?;
+    let mut n =
+        match unsafe { getxattrat(fd, base, name, &mut my_args, AtFlags::AT_SYMLINK_NOFOLLOW) } {
+            Ok(n) => n,
+            Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+                // SAFETY: Avoid a well-behaving process from
+                // repeating calls to potentially exhaust memory.
+                // See tar's tests for an example.
+                return Err(Errno::E2BIG);
+            }
+            Err(errno) => return Err(errno),
+        };
 
-    if let Some(buf) = buf {
-        request.write_mem(&buf[..n], args.value)?;
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        n = request.write_mem(&buf, args.value)?;
     }
 
-    #[allow(clippy::cast_possible_wrap)]
+    #[expect(clippy::cast_possible_wrap)]
     Ok(request.return_syscall(n as i64))
 }
 
@@ -837,29 +825,21 @@ fn syscall_setxattr_handler(
     // SAFETY:
     // 1. SysArg has one element.
     // 2. `/` is not permitted -> EACCES.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let fd = args.0.as_ref().unwrap().dir.as_ref().ok_or(Errno::EACCES)?;
 
     let req = request.scmpreq;
     let name = if req.data.args[1] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        request.read_mem(&mut buf, req.data.args[1])?;
-        Some(buf)
+        Some(request.read_xattr(req.data.args[1])?)
     } else {
         None
     };
-    let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
-    } else {
-        std::ptr::null()
-    };
+    let name = name
+        .as_ref()
+        .map(|cstr| cstr.as_ptr())
+        .unwrap_or(std::ptr::null());
 
-    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+    if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with EACCES to denote access violation.
@@ -873,25 +853,19 @@ fn syscall_setxattr_handler(
     let (buf, len) = if req.data.args[3] == 0 {
         (None, 0)
     } else {
-        let len = usize::try_from(req.data.args[3]).or(Err(Errno::E2BIG))?;
-        let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-        let mut buf = Vec::new();
-        buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
-        request.read_mem(&mut buf, req.data.args[2])?;
+        let len = to_len_val(req.data.args[3], XATTR_SIZE_MAX)?;
+        let buf = request.read_vec(req.data.args[2], len)?;
+        let len = buf.len();
         (Some(buf), len)
     };
     let buf = buf.as_ref().map_or(std::ptr::null(), |b| b.as_ptr()) as *const libc::c_void;
 
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let flags = req.data.args[4] as libc::c_int;
 
     // SAFETY: In libc we trust.
-    if unsafe { libc::fsetxattr(fd.as_raw_fd(), name, buf, len, flags) } == 0 {
-        Ok(request.return_syscall(0))
-    } else {
-        Err(Errno::last())
-    }
+    Errno::result(unsafe { libc::fsetxattr(fd.as_raw_fd(), name, buf, len, flags) })
+        .map(|_| request.return_syscall(0))
 }
 
 /// A helper function to handle setxattrat syscall.
@@ -903,7 +877,7 @@ fn syscall_setxattrat_handler(
     let req = request.scmpreq;
 
     // SAFETY: SysArg has one element.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
     let base = if path.base.is_empty() {
@@ -913,10 +887,10 @@ fn syscall_setxattrat_handler(
     };
 
     // Read struct xattr_args which holds the extension name, buffer size and flags.
-    let mut args = MaybeUninit::<XattrArgs>::uninit();
+    let mut args = MaybeUninit::<XattrArgs>::zeroed();
 
     // SAFETY: Ensure size of XattrArgs matches with user argument.
-    if req.data.args[5] != std::mem::size_of::<XattrArgs>() as u64 {
+    if req.data.args[5] != size_of::<XattrArgs>() as u64 {
         return Err(Errno::EINVAL);
     }
 
@@ -924,37 +898,26 @@ fn syscall_setxattrat_handler(
     // We don't read uninitialized memory, and after `read_mem` fills it,
     // we're good to assume it's valid.
     let buf = unsafe {
-        std::slice::from_raw_parts_mut(
-            args.as_mut_ptr().cast::<u8>(),
-            std::mem::size_of::<XattrArgs>(),
-        )
+        std::slice::from_raw_parts_mut(args.as_mut_ptr().cast::<u8>(), size_of::<XattrArgs>())
     };
 
     // Read the remote data structure.
-    request.read_mem(buf, req.data.args[4])?;
+    request.read_mem(buf, req.data.args[4], buf.len())?;
 
     // SAFETY: read_mem() has initialized `args` if it succeeded.
     let args = unsafe { args.assume_init() };
 
     let name = if req.data.args[3] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        request.read_mem(&mut buf, req.data.args[3])?;
-        Some(buf)
+        Some(request.read_xattr(req.data.args[3])?)
     } else {
         None
     };
-    let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
-    } else {
-        std::ptr::null()
-    };
+    let name = name
+        .as_ref()
+        .map(|cstr| cstr.as_ptr())
+        .unwrap_or(std::ptr::null());
 
-    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+    if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with EACCES to denote access violation.
@@ -968,12 +931,9 @@ fn syscall_setxattrat_handler(
     let (buf, len) = if args.size == 0 {
         (None, 0)
     } else {
-        let len = usize::try_from(args.size).or(Err(Errno::E2BIG))?;
-        let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-        let mut buf = Vec::new();
-        buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
-        request.read_mem(&mut buf, args.value)?;
+        let len = to_len_val(args.size.into(), XATTR_SIZE_MAX)?;
+        let buf = request.read_vec(args.value, len)?;
+        let len = buf.len();
         (Some(buf), len)
     };
     let buf = buf.as_ref().map_or(std::ptr::null(), |b| b.as_ptr()) as *const libc::c_void;
@@ -983,7 +943,7 @@ fn syscall_setxattrat_handler(
         None => return Err(Errno::EBADF),
     };
 
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let my_args = XattrArgs {
         value: buf as u64,
         size: len as u32,
@@ -1006,7 +966,7 @@ fn syscall_listxattr_handler(
     let req = request.scmpreq;
 
     // SAFETY: SysArg has one element.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
     let base = if path.base.is_empty() {
@@ -1018,14 +978,10 @@ fn syscall_listxattr_handler(
     // SAFETY: The size argument to the llistxattr call
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    let len = usize::try_from(req.data.args[2])
-        .or(Err(Errno::E2BIG))?
-        .min(10240); // Cap count at 10240.
-
+    let len = to_len_cap(req.data.args[2], XATTR_LIST_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -1038,32 +994,45 @@ fn syscall_listxattr_handler(
     match &path.dir {
         Some(fd) => {
             // SAFETY: We use fchdir which is TOCTOU-free!
-            fchdir(fd)?;
+            safe_fchdir(fd)?;
         }
-        None => fchdir(ROOT_FILE())?,
+        None => safe_fchdir(ROOT_FILE())?,
     };
 
     let res = base
         // SAFETY: We do not resolve symbolic links here!
         .with_nix_path(|cstr| unsafe { libc::llistxattr(cstr.as_ptr(), ptr, len) })?;
 
-    #[allow(clippy::cast_sign_loss)]
-    let mut n = Errno::result(res)? as usize;
+    #[expect(clippy::cast_sign_loss)]
+    let mut n = match Errno::result(res) {
+        Ok(n) => n as usize,
+        Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
-    if let Some(buf) = buf {
-        // SAFETY: Filter out attributes that start with "user.syd".
-        // SAFETY: Deny only if the Sandbox is locked for the process.
-        let buf = if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
-            filterxattr(&buf[..n], n)?
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        // SAFETY:
+        // Filter out attributes that start with "user.syd".
+        // Deny only if the Sandbox is locked for the process.
+        let buf = if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+            Cow::Owned(filterxattr(&buf, n)?)
         } else {
-            buf[..n].to_vec()
+            Cow::Borrowed(&buf)
         };
 
-        request.write_mem(&buf, req.data.args[1])?;
-        n = buf.len();
+        n = request.write_mem(&buf, req.data.args[1])?;
     }
 
-    #[allow(clippy::cast_possible_wrap)]
+    #[expect(clippy::cast_possible_wrap)]
     Ok(request.return_syscall(n as i64))
 }
 
@@ -1076,7 +1045,7 @@ fn syscall_listxattrat_handler(
     let req = request.scmpreq;
 
     // SAFETY: SysArg has one element.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
     let base = if path.base.is_empty() {
@@ -1088,14 +1057,10 @@ fn syscall_listxattrat_handler(
     // SAFETY: The size argument to the llistxattr call
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    let len = usize::try_from(req.data.args[4])
-        .or(Err(Errno::E2BIG))?
-        .min(10240); // Cap count at 10240.
-
+    let len = to_len_cap(req.data.args[4], XATTR_LIST_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -1113,22 +1078,35 @@ fn syscall_listxattrat_handler(
     // SAFETY:
     // 1. `ptr` is a valid raw pointer (may be NULL)!
     // 2. We do not resolve symbolic links here!
-    let mut n = unsafe { listxattrat(fd, base, AtFlags::AT_SYMLINK_NOFOLLOW, ptr, len) }?;
+    let mut n = match unsafe { listxattrat(fd, base, AtFlags::AT_SYMLINK_NOFOLLOW, ptr, len) } {
+        Ok(n) => n,
+        Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
-    if let Some(buf) = buf {
-        // SAFETY: Filter out attributes that start with "user.syd".
-        // SAFETY: Deny only if the Sandbox is locked for the process.
-        let buf = if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
-            filterxattr(&buf[..n], n)?
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        // SAFETY:
+        // Filter out attributes that start with "user.syd".
+        // Deny only if the Sandbox is locked for the process.
+        let buf = if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+            Cow::Owned(filterxattr(&buf, n)?)
         } else {
-            buf[..n].to_vec()
+            Cow::Borrowed(&buf)
         };
 
-        request.write_mem(&buf, req.data.args[3])?;
-        n = buf.len();
+        n = request.write_mem(&buf, req.data.args[3])?;
     }
 
-    #[allow(clippy::cast_possible_wrap)]
+    #[expect(clippy::cast_possible_wrap)]
     Ok(request.return_syscall(n as i64))
 }
 
@@ -1141,30 +1119,22 @@ fn syscall_removexattr_handler(
     // SAFETY:
     // 1. SysArg has one element.
     // 2. `/` is not permitted -> EACCES.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let fd = args.0.as_ref().unwrap().dir.as_ref().ok_or(Errno::EACCES)?;
 
     let req = request.scmpreq;
 
     let name = if req.data.args[1] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        request.read_mem(&mut buf, req.data.args[1])?;
-        Some(buf)
+        Some(request.read_xattr(req.data.args[1])?)
     } else {
         None
     };
-    let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
-    } else {
-        std::ptr::null()
-    };
+    let name = name
+        .as_ref()
+        .map(|cstr| cstr.as_ptr())
+        .unwrap_or(std::ptr::null());
 
-    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+    if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes.
         // name is either NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -1173,11 +1143,8 @@ fn syscall_removexattr_handler(
     }
 
     // SAFETY: In libc we trust.
-    if unsafe { libc::fremovexattr(fd.as_raw_fd(), name) } == 0 {
-        Ok(request.return_syscall(0))
-    } else {
-        Err(Errno::last())
-    }
+    Errno::result(unsafe { libc::fremovexattr(fd.as_raw_fd(), name) })
+        .map(|_| request.return_syscall(0))
 }
 
 /// A helper function to handle removexattrat syscall.
@@ -1189,7 +1156,7 @@ fn syscall_removexattrat_handler(
     let req = request.scmpreq;
 
     // SAFETY: SysArg has one element.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
     let base = if path.base.is_empty() {
@@ -1199,24 +1166,16 @@ fn syscall_removexattrat_handler(
     };
 
     let name = if req.data.args[3] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        request.read_mem(&mut buf, req.data.args[3])?;
-        Some(buf)
+        Some(request.read_xattr(req.data.args[3])?)
     } else {
         None
     };
-    let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
-    } else {
-        std::ptr::null()
-    };
+    let name = name
+        .as_ref()
+        .map(|cstr| cstr.as_ptr())
+        .unwrap_or(std::ptr::null());
 
-    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+    if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes.
         // name is either NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -1235,3 +1194,18 @@ fn syscall_removexattrat_handler(
     unsafe { removexattrat(fd, base, name, AtFlags::AT_SYMLINK_NOFOLLOW) }
         .map(|_| request.return_syscall(0))
 }
+
+// Capping length converter, used by *{get,list}xattr*
+#[inline]
+fn to_len_cap(arg: u64, max: usize) -> Result<usize, Errno> {
+    Ok(usize::try_from(arg).or(Err(Errno::E2BIG))?.min(max))
+}
+
+// Validating length converter, used by *setxattr*
+#[inline]
+fn to_len_val(arg: u64, max: usize) -> Result<usize, Errno> {
+    match usize::try_from(arg).or(Err(Errno::ERANGE)) {
+        Ok(len) if len <= max => Ok(len),
+        _ => Err(Errno::ERANGE),
+    }
+}
