@@ -2,6 +2,10 @@
 
 #[cfg(test)]
 use std::convert::TryInto;
+use std::{
+    fmt::{self, Display, Formatter},
+    io::Error,
+};
 
 use bitflags::bitflags;
 #[cfg(test)]
@@ -42,8 +46,8 @@ use crate::landlock::{uapi, Access, CompatError};
 ///
 /// In a nutshell, test the access rights you request on a kernel that support them and
 /// on a kernel that doesn't support them.
-#[cfg_attr(test, derive(Debug, EnumIter, EnumCountMacro))]
-#[derive(Copy, Clone, PartialEq, Eq, PartialOrd)]
+#[cfg_attr(test, derive(EnumIter, EnumCountMacro))]
+#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd)]
 #[non_exhaustive]
 pub enum ABI {
     /// Kernel not supporting Landlock, either because it is not built with Landlock
@@ -146,21 +150,124 @@ fn known_abi() {
     assert!(!ABI::is_known(last_i + 1));
 }
 
+impl Display for ABI {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        match self {
+            ABI::Unsupported => write!(f, "unsupported"),
+            v => (*v as u32).fmt(f),
+        }
+    }
+}
+
+/// Status of Landlock support for the running system.
+///
+/// This enum is used to represent the status of the Landlock support for the system where the code
+/// is executed. It can indicate whether Landlock is available or not.
+///
+/// # Warning
+///
+/// Sandboxed programs should only use this data to log or provide information to users,
+/// not to change their behavior according to this status.  Indeed, the `Ruleset` and the other
+/// types are designed to handle the compatibility in a simple and safe way.
+#[derive(Copy, Clone, Debug, PartialEq, Eq)]
+pub enum LandlockStatus {
+    /// Landlock is supported but not enabled (`EOPNOTSUPP`).
+    NotEnabled,
+    /// Landlock is not implemented (i.e. not built into the running kernel: `ENOSYS`).
+    NotImplemented,
+    /// Landlock is available and supported up to the given ABI.
+    ///
+    /// `Option<i32>` contains the raw ABI value if it's greater than the greatest known ABI,
+    /// which would mean that the running kernel is newer than the Landlock crate.
+    Available(ABI, Option<i32>),
+}
+
+impl LandlockStatus {
+    // Must remain private to avoid inconsistent behavior using such unknown-at-build-time ABI
+    // e.g., AccessFs::from_all(ABI::new_current())
+    //
+    // This should not be Default::default() because the returned value would may not be the same
+    // for all users.
+    fn current() -> Self {
+        // Landlock ABI version starts at 1 but errno is only set for negative values.
+        let v = unsafe {
+            uapi::landlock_create_ruleset(
+                std::ptr::null(),
+                0,
+                uapi::LANDLOCK_CREATE_RULESET_VERSION,
+            )
+        };
+        if v < 0 {
+            // The only possible error values should be EOPNOTSUPP and ENOSYS.
+            match Error::last_os_error().raw_os_error() {
+                Some(libc::EOPNOTSUPP) => Self::NotEnabled,
+                _ => Self::NotImplemented,
+            }
+        } else {
+            let abi = ABI::from(v);
+            Self::Available(abi, (v != abi as i32).then_some(v))
+        }
+    }
+}
+
+// Test against the running kernel.
+#[test]
+fn test_current_landlock_status() {
+    let status = LandlockStatus::current();
+    if *TEST_ABI == ABI::Unsupported {
+        assert_eq!(status, LandlockStatus::NotImplemented);
+    } else {
+        assert!(matches!(status, LandlockStatus::Available(abi, _) if abi == *TEST_ABI));
+        if std::env::var(TEST_ABI_ENV_NAME).is_ok() {
+            // We cannot reliably check for unknown kernel.
+            assert!(matches!(status, LandlockStatus::Available(_, None)));
+        }
+    }
+}
+
+impl From<LandlockStatus> for ABI {
+    fn from(status: LandlockStatus) -> Self {
+        match status {
+            // The only possible error values should be EOPNOTSUPP and ENOSYS,
+            // but let's convert all kind of errors as unsupported.
+            LandlockStatus::NotEnabled | LandlockStatus::NotImplemented => ABI::Unsupported,
+            LandlockStatus::Available(abi, _) => abi,
+        }
+    }
+}
+
+// This is only useful to tests and should not be exposed publicly because
+// the mapping can only be partial.
 #[cfg(test)]
-lazy_static! {
-    static ref TEST_ABI: ABI = match std::env::var("LANDLOCK_CRATE_TEST_ABI") {
-        Ok(s) => {
-            let n = s.parse::<i32>().unwrap();
+impl From<ABI> for LandlockStatus {
+    fn from(abi: ABI) -> Self {
+        match abi {
+            // Convert to ENOSYS because of check_ruleset_support() and ruleset_unsupported() tests.
+            ABI::Unsupported => Self::NotImplemented,
+            _ => Self::Available(abi, None),
+        }
+    }
+}
+
+#[cfg(test)]
+static TEST_ABI_ENV_NAME: &str = "LANDLOCK_CRATE_TEST_ABI";
+
+#[cfg(test)]
+static TEST_ABI: std::sync::LazyLock<ABI> = std::sync::LazyLock::new(|| {
+    use std::os::unix::ffi::OsStrExt;
+
+    match std::env::var_os(TEST_ABI_ENV_NAME) {
+        Some(s) => {
+            let n = btoi::btoi::<i32>(s.as_bytes()).expect("LANDLOCK_CRATE_TEST_ABI");
             if ABI::is_known(n) || n == 0 {
                 ABI::from(n)
             } else {
                 panic!("Unknown ABI: {n}");
             }
         }
-        Err(std::env::VarError::NotPresent) => ABI::new_current(),
-        Err(e) => panic!("Failed to read LANDLOCK_CRATE_TEST_ABI: {e}"),
-    };
-}
+        None => LandlockStatus::current().into(),
+    }
+});
 
 #[cfg(test)]
 pub(crate) fn can_emulate(mock: ABI, partial_support: ABI, full_support: Option<ABI>) -> bool {
@@ -175,17 +282,21 @@ pub(crate) fn can_emulate(mock: ABI, partial_support: ABI, full_support: Option<
 
 #[cfg(test)]
 pub(crate) fn get_errno_from_landlock_status() -> Option<i32> {
-    use std::io::Error;
-
-    match ABI::new_current() {
-        ABI::Unsupported => match Error::last_os_error().raw_os_error() {
-            // Returns ENOSYS when the kernel is not built with Landlock support,
-            // or EOPNOTSUPP when Landlock is supported but disabled at boot time.
-            ret @ Some(libc::ENOSYS | libc::EOPNOTSUPP) => ret,
-            // Other values can only come from bogus seccomp filters or debug tampering.
-            _ => unreachable!(),
-        },
-        _ => None,
+    match LandlockStatus::current() {
+        LandlockStatus::NotImplemented | LandlockStatus::NotEnabled => {
+            match Error::last_os_error().raw_os_error() {
+                // Returns ENOSYS when the kernel is not built with Landlock support,
+                // or EOPNOTSUPP when Landlock is supported but disabled at boot time.
+                ret @ Some(libc::ENOSYS | libc::EOPNOTSUPP) => ret,
+                // Other values can only come from bogus seccomp filters or debugging tampering.
+                ret => {
+                    eprintln!("Current kernel should support this Landlock ABI according to $LANDLOCK_CRATE_TEST_ABI");
+                    eprintln!("Unexpected result: {ret:?}");
+                    unreachable!();
+                }
+            }
+        }
+        LandlockStatus::Available(_, _) => None,
     }
 }
 
@@ -197,7 +308,7 @@ fn current_kernel_abi() {
     // Landlock ABI version known by this crate is automatically set.
     // From Linux 5.13 to 5.18, you need to run: LANDLOCK_CRATE_TEST_ABI=1 cargo test
     let test_abi = *TEST_ABI;
-    let current_abi = ABI::new_current();
+    let current_abi = LandlockStatus::current().into();
     println!(
         "Current kernel version: {}",
         std::fs::read_to_string("/proc/version")
@@ -210,8 +321,7 @@ fn current_kernel_abi() {
 
 // CompatState is not public outside this crate.
 /// Returned by ruleset builder.
-#[cfg_attr(test, derive(Debug))]
-#[derive(Copy, Clone, PartialEq, Eq)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq)]
 pub enum CompatState {
     /// Initial undefined state.
     Init,
@@ -278,29 +388,35 @@ fn compat_state_update_2() {
     assert_eq!(state, CompatState::Partial);
 }
 
-#[cfg_attr(test, derive(Debug, PartialEq))]
-#[derive(Copy, Clone)]
+#[cfg_attr(test, derive(PartialEq))]
+#[derive(Copy, Clone, Debug)]
 pub(crate) struct Compatibility {
-    abi: ABI,
+    status: LandlockStatus,
     pub(crate) level: Option<CompatLevel>,
     pub(crate) state: CompatState,
 }
 
-impl From<ABI> for Compatibility {
-    fn from(abi: ABI) -> Self {
+impl From<LandlockStatus> for Compatibility {
+    fn from(status: LandlockStatus) -> Self {
         Compatibility {
-            abi,
+            status,
             level: Default::default(),
             state: CompatState::Init,
         }
     }
 }
 
+#[cfg(test)]
+impl From<ABI> for Compatibility {
+    fn from(abi: ABI) -> Self {
+        Self::from(LandlockStatus::from(abi))
+    }
+}
+
 impl Compatibility {
     // Compatibility is a semi-opaque struct.
-    #[allow(clippy::new_without_default)]
     pub(crate) fn new() -> Self {
-        ABI::new_current().into()
+        LandlockStatus::current().into()
     }
 
     pub(crate) fn update(&mut self, state: CompatState) {
@@ -308,7 +424,11 @@ impl Compatibility {
     }
 
     pub(crate) fn abi(&self) -> ABI {
-        self.abi
+        self.status.into()
+    }
+
+    pub(crate) fn status(&self) -> LandlockStatus {
+        self.status
     }
 }
 
@@ -461,7 +581,7 @@ pub trait Compatible: Sized + private::OptionCompatLevelMut {
 }
 
 #[test]
-#[allow(deprecated)]
+#[expect(deprecated)]
 fn deprecated_set_best_effort() {
     use crate::landlock::{CompatLevel, Compatible, Ruleset};
 
