@@ -1,9 +1,34 @@
 // SPDX-License-Identifier: Apache-2.0 OR MIT
 
+// Use architecture-specific bindings for native x86_64 and x86 architectures.
+// They contain minimal Landlock-only bindings with layout tests.
+#[allow(dead_code)]
+#[allow(non_camel_case_types)]
+#[allow(non_snake_case)]
+#[allow(non_upper_case_globals)]
+#[expect(clippy::arithmetic_side_effects)]
+#[cfg(target_arch = "x86_64")]
+#[path = "landlock_x86_64.rs"]
+mod landlock;
+
 #[allow(dead_code)]
 #[allow(non_camel_case_types)]
 #[allow(non_snake_case)]
 #[allow(non_upper_case_globals)]
+#[expect(clippy::arithmetic_side_effects)]
+#[cfg(target_arch = "x86")]
+#[path = "landlock_i686.rs"]
+mod landlock;
+
+// For all other architectures, use the architecture-agnostic landlock_all.rs
+// bindings without layout tests.
+#[allow(dead_code)]
+#[allow(non_camel_case_types)]
+#[allow(non_snake_case)]
+#[allow(non_upper_case_globals)]
+#[expect(clippy::arithmetic_side_effects)]
+#[cfg(not(any(target_arch = "x86_64", target_arch = "x86")))]
+#[path = "landlock_all.rs"]
 mod landlock;
 
 #[rustfmt::skip]
