@@ -14,26 +14,29 @@
 use std::{ops::RangeInclusive, str::FromStr};
 
 use btoi::{btoi, btoi_radix};
+use dur::Duration;
 use fixedbitset::FixedBitSet;
 use memchr::arch::all::is_prefix;
-use nix::{errno::Errno, mount::MsFlags};
+use nix::errno::Errno;
 use nom::{
     branch::alt,
     bytes::complete::{tag, tag_no_case, take_while1},
     character::complete::{char, digit1, one_of},
     combinator::{all_consuming, map, opt, recognize},
+    error::{Error, ErrorKind},
     multi::separated_list1,
     sequence::preceded,
     Finish, IResult, Parser,
 };
 
 use crate::{
-    confine::SydMsFlags,
+    confine::SydMountAttrFlags,
     hash::SydHashSet,
     landlock::{AccessFs, AccessNet},
     landlock_policy::LandlockPolicy,
+    mount::api::MountAttrFlags,
     path::XPathBuf,
-    sandbox::{Action, BindMount, Capability},
+    sandbox::{Action, Capability},
 };
 
 // Valid Netlink families.
@@ -70,8 +73,14 @@ const NETLINK_FAMILIES: &[&str] = &[
 pub(crate) struct BindCmd {
     // One of '+', '-', '^'
     pub(crate) op: char,
-    // Parsed BindMount
-    pub(crate) mount: BindMount,
+    // Mount source.
+    pub(crate) src: String,
+    // Mount destination.
+    pub(crate) dst: String,
+    // Mount flags.
+    pub(crate) opt: MountAttrFlags,
+    // Optional filesystem specific data.
+    pub(crate) dat: Option<String>,
 }
 
 // Represents a parsed "force" command.
@@ -142,6 +151,14 @@ impl TryFrom<char> for LandlockOp {
     }
 }
 
+/// Parsed filesystem sandbox rule.
+#[derive(Clone, Debug, PartialEq, Eq)]
+pub(crate) struct FsCmd {
+    pub(crate) action: Action,
+    pub(crate) op: char,
+    pub(crate) fs_type: String,
+}
+
 /// Parsed "allow/lock" command.
 #[derive(Debug, PartialEq, Eq)]
 pub struct LandlockCmd {
@@ -202,11 +219,11 @@ impl NetlinkCmd {
 //   * If `<opt>` is present, split on commas: known flags -> MsFlags; unknown accumulate into `dat`.
 pub(crate) fn parse_bind_cmd(command: &str) -> Result<BindCmd, Errno> {
     // Inner nom parser: returns (remaining, (op, src_str, dst_str, opt_str?))
-    #[allow(clippy::type_complexity)]
+    #[expect(clippy::type_complexity)]
     fn inner(input: &str) -> IResult<&str, (char, &str, &str, Option<&str>)> {
         // Sequence: "bind", one of '+','-','^', <src> (no ':'), ":", <dst> (no ':'), optional ":" + <opt>
         (
-            nom::bytes::complete::tag("bind"),
+            tag("bind"),
             alt((char('+'), char('-'), char('^'))),
             take_while1(|c| c != ':'), // src
             char(':'),                 // consume colon
@@ -219,17 +236,8 @@ pub(crate) fn parse_bind_cmd(command: &str) -> Result<BindCmd, Errno> {
 
     match inner(command).finish() {
         Ok(("", (op, src, dst, opt))) => {
-            let src = XPathBuf::from(src);
-            let dst = XPathBuf::from(dst);
-
-            // Validate src/dst.
-            // Deny if dst is not absolute or any ".." in src/dst
-            if dst.is_relative() || src.has_parent_dot() || dst.has_parent_dot() {
-                return Err(Errno::EINVAL);
-            }
-
-            // Build MsFlags and dat from opt.
-            let mut flags = MsFlags::empty();
+            // Build MountAttrFlags and dat from opt.
+            let mut flags = MountAttrFlags::empty();
             let mut dat = Vec::new();
 
             if let Some(opt) = opt {
@@ -253,7 +261,7 @@ pub(crate) fn parse_bind_cmd(command: &str) -> Result<BindCmd, Errno> {
                     }
 
                     // Try to parse as a mount flag.
-                    if let Some(syd_flag) = SydMsFlags::from_name(flag) {
+                    if let Some(syd_flag) = SydMountAttrFlags::from_name(flag) {
                         flags |= syd_flag.0;
                     } else {
                         // Add to auxiliary data.
@@ -268,17 +276,21 @@ pub(crate) fn parse_bind_cmd(command: &str) -> Result<BindCmd, Errno> {
             let dat = if dat.is_empty() {
                 None
             } else {
-                Some(XPathBuf::from(dat))
+                Some(String::from_utf8(dat).or(Err(Errno::EINVAL))?)
             };
 
-            let mount = BindMount {
-                opt: flags,
+            // Source and destination will be expanded as necessary,
+            // so we perform no further handling here.
+            let src = src.to_string();
+            let dst = dst.to_string();
+
+            Ok(BindCmd {
+                op,
                 src,
                 dst,
                 dat,
-            };
-
-            Ok(BindCmd { op, mount })
+                opt: flags,
+            })
         }
         _ => Err(Errno::EINVAL),
     }
@@ -336,10 +348,7 @@ pub(crate) fn parse_force_cmd(input: &str) -> Result<ForceCmd, Errno> {
             match Action::from_str(act) {
                 Ok(act) => Some(act),
                 Err(_) => {
-                    return Err(nom::Err::Error(nom::error::Error::new(
-                        input,
-                        nom::error::ErrorKind::Fail,
-                    )));
+                    return Err(nom::Err::Error(Error::new(input, ErrorKind::Fail)));
                 }
             }
         } else {
@@ -527,6 +536,43 @@ pub fn parse_landlock_cmd(input: &str) -> Result<LandlockCmd, Errno> {
     }
 }
 
+// Parse a filesystem rule command string, returning `FsCmd` or `Errno::EINVAL`.
+//
+// Format: `<action>/fs<op><fs_type>`
+// - `<action>`: one of allow, deny, filter, warn, stop, abort, kill, panic, exit
+// - `<op>`: one of '+','-','^'
+// - `<fs_type>`: non-empty string.
+//
+// Returns Err(EINVAL) on any parse or validation failure.
+pub(crate) fn parse_fs_cmd(input: &str) -> Result<FsCmd, Errno> {
+    // Inner parser: action "/" op fs_type.
+    fn inner(input: &str) -> IResult<&str, (&str, char, &str)> {
+        (
+            take_while1(|c| c != '/'),
+            char('/'),
+            tag("fs"),
+            one_of("+-^"),
+            take_while1(|_| true), // fs_type (rest of line, must be non-empty).
+        )
+            .map(|(act, _slash, _caps, op, fs_type)| (act, op, fs_type))
+            .parse(input)
+    }
+
+    match inner(input).finish() {
+        Ok(("", (act, op, fs_type))) => {
+            let action = Action::from_str(act).map_err(|_| Errno::EINVAL)?;
+            let fs_type = fs_type.to_string();
+
+            Ok(FsCmd {
+                action,
+                op,
+                fs_type,
+            })
+        }
+        _ => Err(Errno::EINVAL),
+    }
+}
+
 // Parse a seccomp rule command string, returning `ScmpCmd` or `Errno::EINVAL`.
 //
 // Format: `<action>/<caps><op><pat>`
@@ -544,7 +590,7 @@ pub fn parse_landlock_cmd(input: &str) -> Result<LandlockCmd, Errno> {
 // Returns Err(EINVAL) on any parse or validation failure.
 pub(crate) fn parse_scmp_cmd(input: &str) -> Result<ScmpCmd, Errno> {
     // Inner parser: action "/" caps op pat.
-    #[allow(clippy::type_complexity)]
+    #[expect(clippy::type_complexity)]
     fn inner(input: &str) -> IResult<&str, (&str, &str, char, &str)> {
         (
             take_while1(|c| c != '/'),
@@ -574,6 +620,12 @@ pub(crate) fn parse_scmp_cmd(input: &str) -> Result<ScmpCmd, Errno> {
                 }
 
                 let cap = Capability::from_str(cap)?;
+
+                // Reject ioctl which is for requests not paths.
+                if cap.contains(Capability::CAP_IOCTL) {
+                    return Err(Errno::EINVAL);
+                }
+
                 filter.insert(cap);
             }
 
@@ -582,9 +634,9 @@ pub(crate) fn parse_scmp_cmd(input: &str) -> Result<ScmpCmd, Errno> {
                 return Err(Errno::EINVAL);
             }
 
-            // IP address arguments are only valid for the `inet` set.
-            let maybe_addr = filter.intersects(Capability::CAP_INET)
-                && filter.difference(Capability::CAP_INET).is_empty();
+            // IP address arguments are only valid for the `net` set.
+            let maybe_addr = filter.intersects(Capability::CAP_NET)
+                && filter.difference(Capability::CAP_NET).is_empty();
 
             // Pattern resolution.
             let pat = if maybe_addr {
@@ -687,10 +739,7 @@ fn netlink_parser(input: &str) -> IResult<&str, Vec<String>> {
     // Validate each.
     for &fam in &raw_list {
         if NETLINK_FAMILIES.binary_search(&fam).is_err() {
-            return Err(nom::Err::Error(nom::error::Error::new(
-                input,
-                nom::error::ErrorKind::Fail,
-            )));
+            return Err(nom::Err::Error(Error::new(input, ErrorKind::Fail)));
         }
     }
 
@@ -759,6 +808,18 @@ pub fn str2u64(value: &[u8]) -> Result<u64, Errno> {
     .or(Err(Errno::EINVAL))
 }
 
+/// Converts a string representation of a duration to a `Duration` value.
+///
+/// The string may be human formatted in which case it will be parsed by
+/// the dur crate.
+pub fn str2dur(value: &str) -> Result<Duration, Errno> {
+    if let Ok(value) = str2u64(value.as_bytes()) {
+        Ok(Duration::from_secs(value.into()))
+    } else {
+        value.parse::<Duration>().or(Err(Errno::EINVAL))
+    }
+}
+
 /// Converts a string representation of a number into a `u32` value.
 ///
 /// The string can be in hexadecimal (prefixed with "0x"), octal
@@ -784,10 +845,10 @@ mod tests {
         let cmd = "bind+/:/:ro";
         let bc = parse_bind_cmd(cmd).unwrap();
         assert_eq!(bc.op, '+');
-        assert_eq!(bc.mount.src, XPathBuf::from("/"));
-        assert_eq!(bc.mount.dst, XPathBuf::from("/"));
-        assert!(bc.mount.opt.contains(MsFlags::MS_RDONLY));
-        assert!(bc.mount.dat.is_none());
+        assert_eq!(bc.src, "/");
+        assert_eq!(bc.dst, "/");
+        assert!(bc.opt.contains(MountAttrFlags::MOUNT_ATTR_RDONLY));
+        assert!(bc.dat.is_none());
     }
 
     #[test]
@@ -795,10 +856,10 @@ mod tests {
         let cmd = "bind-/mnt/data:/data";
         let bc = parse_bind_cmd(cmd).unwrap();
         assert_eq!(bc.op, '-');
-        assert_eq!(bc.mount.src, XPathBuf::from("/mnt/data"));
-        assert_eq!(bc.mount.dst, XPathBuf::from("/data"));
-        assert!(bc.mount.opt.is_empty());
-        assert!(bc.mount.dat.is_none());
+        assert_eq!(bc.src, "/mnt/data");
+        assert_eq!(bc.dst, "/data");
+        assert!(bc.opt.is_empty());
+        assert!(bc.dat.is_none());
     }
 
     #[test]
@@ -807,13 +868,13 @@ mod tests {
             "bind^overlay:/tmp/target:lowerdir=/tmp/lower,upperdir=/tmp/upper,workdir=/tmp/work";
         let bc = parse_bind_cmd(cmd).unwrap();
         assert_eq!(bc.op, '^');
-        assert_eq!(bc.mount.src, XPathBuf::from("overlay"));
-        assert_eq!(bc.mount.dst, XPathBuf::from("/tmp/target"));
+        assert_eq!(bc.src, "overlay");
+        assert_eq!(bc.dst, "/tmp/target");
         // All three flags are unknown, so go into dat_buf
-        assert!(bc.mount.opt.is_empty());
+        assert!(bc.opt.is_empty());
         assert_eq!(
-            bc.mount.dat.unwrap(),
-            XPathBuf::from("lowerdir=/tmp/lower,upperdir=/tmp/upper,workdir=/tmp/work")
+            bc.dat.unwrap(),
+            "lowerdir=/tmp/lower,upperdir=/tmp/upper,workdir=/tmp/work",
         );
     }
 
@@ -822,12 +883,12 @@ mod tests {
         let cmd = "bind+tmpfs:/tmp:ro,nosuid,size=10M";
         let bc = parse_bind_cmd(cmd).unwrap();
         assert_eq!(bc.op, '+');
-        assert_eq!(bc.mount.src, XPathBuf::from("tmpfs"));
-        assert_eq!(bc.mount.dst, XPathBuf::from("/tmp"));
-        assert!(bc.mount.opt.contains(MsFlags::MS_RDONLY));
-        assert!(bc.mount.opt.contains(MsFlags::MS_NOSUID));
+        assert_eq!(bc.src, "tmpfs");
+        assert_eq!(bc.dst, "/tmp");
+        assert!(bc.opt.contains(MountAttrFlags::MOUNT_ATTR_RDONLY));
+        assert!(bc.opt.contains(MountAttrFlags::MOUNT_ATTR_NOSUID));
         // "size=10M" is unknown -> goes into dat_buf
-        assert_eq!(bc.mount.dat.unwrap(), XPathBuf::from("size=10M"));
+        assert_eq!(bc.dat.unwrap(), "size=10M");
     }
 
     #[test]
@@ -837,21 +898,6 @@ mod tests {
         assert_eq!(parse_bind_cmd("bind+:/dst:opt"), Err(Errno::EINVAL));
     }
 
-    #[test]
-    fn test_parse_bind_parse_bind_relative_dst_fails() {
-        // dst does not start with '/', fails
-        assert_eq!(parse_bind_cmd("bind+/src:relative:opt"), Err(Errno::EINVAL));
-    }
-
-    #[test]
-    fn test_parse_bind_parse_bind_parent_dot_fails() {
-        assert_eq!(parse_bind_cmd("bind+/src/../etc:/dst"), Err(Errno::EINVAL));
-        assert_eq!(
-            parse_bind_cmd("bind+/src:/dst/../../tmp"),
-            Err(Errno::EINVAL)
-        );
-    }
-
     #[test]
     fn test_parse_bind_parse_bind_bad_mod_fails() {
         assert_eq!(parse_bind_cmd("bind*src:/dst"), Err(Errno::EINVAL));
@@ -1431,7 +1477,7 @@ mod tests {
             }
         );
 
-        let cmd = parse_scmp_cmd("filter/ioctl,all,chdir^/tmp").unwrap();
+        let cmd = parse_scmp_cmd("filter/all,chdir^/tmp").unwrap();
         assert_eq!(
             cmd,
             ScmpCmd {
@@ -1619,7 +1665,7 @@ mod tests {
                 action: Action::Stop,
                 filter: Capability::CAP_NET,
                 op: '^',
-                pat: ScmpPattern::Path("1.2.3.4!80".into()),
+                pat: ScmpPattern::Addr("1.2.3.4!80".into()),
             }
         );
     }
