@@ -14,32 +14,36 @@ use std::{
     cmp::Ordering,
     collections::VecDeque,
     ffi::{CStr, OsStr, OsString},
-    ops::Deref,
+    ops::{Deref, RangeBounds},
     os::{
         fd::RawFd,
         unix::ffi::{OsStrExt, OsStringExt},
     },
-    path::{Path, PathBuf},
+    path::{Component, Path, PathBuf},
     sync::LazyLock,
 };
 
 use btoi::btoi;
+use itoa::Integer;
 use memchr::{
-    arch::all::{is_equal, is_prefix, is_suffix, memchr::One},
+    arch::all::{is_equal, is_prefix, is_suffix},
     memchr, memmem, memrchr,
 };
 use nix::{
     errno::Errno,
-    fcntl::{openat2, OFlag, OpenHow, ResolveFlag, AT_FDCWD},
+    fcntl::{OFlag, AT_FDCWD},
     libc::pid_t,
     unistd::Pid,
     NixPath,
 };
 
 use crate::{
+    compat::{openat2, OpenHow, ResolveFlag},
     config::MAGIC_PREFIX,
-    fs::{retry_on_eintr, tgkill, FileType},
+    fs::tgkill,
     log::log_untrusted_buf,
+    lookup::FileType,
+    retry::retry_on_eintr,
 };
 
 /// Generate a formatted `XPathBuf`.
@@ -65,7 +69,7 @@ pub(crate) fn dotdot_with_nul() -> u64 {
 }
 
 /// `PathBuf` for UNIX.
-// SAFETY: k1 == k2 ⇒ hash(k1) == hash(k2) always holds for our PartialEq impl.
+// SAFETY: k1 == k2 -> hash(k1) == hash(k2) always holds for our PartialEq impl.
 #[allow(clippy::derived_hash_with_manual_eq)]
 #[derive(Clone, Hash, Ord, PartialOrd)]
 pub struct XPathBuf(Vec<u8>);
@@ -405,16 +409,19 @@ impl XPath {
             return Ok(());
         };
         let proc_pid = if let Some(pid) = proc_pid {
-            pid
+            Pid::from_raw(pid)
         } else {
             return Ok(());
         };
+        if proc_pid == pid {
+            return Ok(());
+        }
 
         //
         // RESTRICTION 2: Protect readdir(/proc).
         //
         // SAFETY: Prevent /proc process tree traversal.
-        if is_proc && proc_pid != pid.as_raw() {
+        if is_proc {
             return Err(Errno::ENOENT);
         }
         // END OF RESTRICTION 2
@@ -426,7 +433,6 @@ impl XPath {
         //
         // Step 1: Protect Syd thread group.
         let syd_pid = Pid::this();
-        let proc_pid = Pid::from_raw(proc_pid);
         if proc_pid == syd_pid {
             return Err(Errno::ENOENT);
         }
@@ -484,7 +490,7 @@ impl XPath {
     /// Enforcing these rules helps prevent security vulnerabilities
     /// arising from unexpected or malicious filenames, such as command
     /// injection, denial of service, or arbitrary file manipulation.
-    #[allow(clippy::arithmetic_side_effects)]
+    #[expect(clippy::arithmetic_side_effects)]
     pub fn check_name(&self) -> Result<(), Errno> {
         let (_, name) = self.split();
         let name = name.as_bytes();
@@ -551,10 +557,39 @@ impl XPath {
         Ok(())
     }
 
+    /// Convert /proc/${pid} to /proc/self as necessary.
+    ///
+    /// This must never used in syscalls only in access check.
+    pub fn replace_proc_self<'a>(&'a self, pid: Pid) -> Cow<'a, Self> {
+        let p = if let Some(p) = self.split_prefix(b"/proc") {
+            p
+        } else {
+            return Cow::Borrowed(self);
+        };
+
+        let mut buf = itoa::Buffer::new();
+        let pid = buf.format(pid.as_raw());
+        let p = if let Some(p) = p.split_prefix(pid.as_bytes()) {
+            p
+        } else {
+            return Cow::Borrowed(self);
+        };
+
+        let mut pdir = XPathBuf::from("/proc/self");
+        // SAFETY: We do not use XPathBuf::push here,
+        // and allow slash in path. This is safe against
+        // path traversals because the path is guaranteed
+        // to be a canonicalized path.
+        pdir.append_byte(b'/');
+        pdir.append_bytes(p.as_bytes());
+
+        Cow::Owned(pdir)
+    }
+
     /// Returns a path that, when joined onto `base`, yields `self`.
     ///
     /// Expects normalized, canonical path.
-    #[allow(clippy::arithmetic_side_effects)]
+    #[expect(clippy::arithmetic_side_effects)]
     pub fn split_prefix(&self, base: &[u8]) -> Option<&Self> {
         let mut len = base.len();
         if len == 0 {
@@ -590,7 +625,7 @@ impl XPath {
     /// - The function efficiently finds the last `/` in the path and splits at that point.
     /// - Trailing slashes are included in the filename to indicate directory paths.
     /// - For the root path `/`, both parent and filename are the original path reference.
-    #[allow(clippy::arithmetic_side_effects)]
+    #[expect(clippy::arithmetic_side_effects)]
     pub fn split(&self) -> (&Self, &Self) {
         // Special cases for the empty and root paths.
         let bytes = match self.get(0) {
@@ -641,7 +676,7 @@ impl XPath {
     pub fn extension(&self) -> Option<&Self> {
         let dot = memrchr(b'.', self.as_bytes())?;
         // dot==Some means len>=1.
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         if dot < self.0.len() - 1 {
             Some(Self::from_bytes(&self.as_bytes()[dot + 1..]))
         } else {
@@ -655,7 +690,7 @@ impl XPath {
     }
 
     /// Determines the length of the parent path.
-    #[allow(clippy::arithmetic_side_effects)]
+    #[expect(clippy::arithmetic_side_effects)]
     pub fn parent_len(&self) -> usize {
         // Special cases for the empty and root paths.
         let bytes = match self.get(0) {
@@ -715,6 +750,33 @@ impl XPath {
         }
     }
 
+    /// Returns a relative path by stripping the root component.
+    ///
+    /// # Safety
+    ///
+    /// This function does NOT assume `self` is normalized and performs
+    /// a small amount of normalization. Therefore this function is safe
+    /// to use with non-normalized paths.
+    pub fn strip_root(&self) -> XPathBuf {
+        let src = self.as_path();
+        let mut dst = XPathBuf::new();
+
+        for comp in src.components() {
+            if matches!(
+                comp,
+                Component::RootDir | Component::Prefix(_) | Component::CurDir
+            ) {
+                continue;
+            }
+            if !dst.is_empty() {
+                dst.append_byte(b'/');
+            }
+            dst.append_bytes(comp.as_os_str().as_bytes());
+        }
+
+        dst
+    }
+
     /// Returns a path that, when joined onto `base`, yields `self`.
     ///
     /// # Safety
@@ -724,7 +786,7 @@ impl XPath {
     /// # Errors
     ///
     /// If `base` is not a prefix of self (i.e., `starts_with` returns
-    /// `false`), returns `Err`.
+    /// `false`), returns `None`.
     pub fn strip_prefix(&self, base: &[u8]) -> Option<&Self> {
         if !self.starts_with(base) {
             return None;
@@ -752,8 +814,7 @@ impl XPath {
     /// This function iterates through the bytes of the path from end to
     /// start, and determines whether the last component before any
     /// slashes is a dot.
-    #[allow(clippy::arithmetic_side_effects)]
-    #[allow(clippy::if_same_then_else)]
+    #[expect(clippy::arithmetic_side_effects)]
     pub fn ends_with_dot(&self) -> bool {
         let bytes = self.as_bytes();
 
@@ -791,12 +852,27 @@ impl XPath {
 
     /// Returns true if the path ends with a slash.
     pub fn ends_with_slash(&self) -> bool {
-        !self.is_rootfs() && self.last() == Some(b'/')
+        self.last() == Some(b'/') && !self.is_rootfs()
     }
 
     /// Check if path has a parent dir component, ie `..`.
     pub fn has_parent_dot(&self) -> bool {
-        self.contains(b"/..") || self.is_equal(b"..")
+        let bytes = self.as_bytes();
+
+        #[allow(clippy::arithmetic_side_effects)]
+        for index in memmem::Finder::new(b"..").find_iter(bytes) {
+            let is_dotdot = if index == 0 {
+                true
+            } else {
+                bytes[index - 1] == b'/'
+            } && (index + 2 == bytes.len() || bytes[index + 2] == b'/');
+
+            if is_dotdot {
+                return true;
+            }
+        }
+
+        false
     }
 
     /// Check if path starts with the `MAGIC_PREFIX`.
@@ -863,23 +939,31 @@ impl XPath {
             .unwrap_or(false)
     }
 
-    /// Check if path points to machine-id(5), /etc/hostid or /var/adm/hostid.
-    pub fn is_machine_id(&self) -> bool {
-        for f in [
-            &b"/etc/machine-id"[..],
-            &b"/etc/hostid"[..],
-            &b"/var/adm/hostid"[..],
-        ] {
-            if self.is_equal(f) {
-                return true;
-            }
-        }
-        false
+    /// Check if path points to `/proc/version`.
+    pub fn is_proc_version(&self) -> bool {
+        self.is_equal(b"/proc/version")
     }
 
-    /// Check if path points to a proc_pid_status(5) file.
-    pub fn is_proc_pid_status(&self) -> bool {
-        self.is_proc_pid() && self.ends_with(b"/status")
+    /// Check if path points to `/proc/sys/kernel/osrelease`.
+    pub fn is_proc_osrelease(&self) -> bool {
+        self.is_equal(b"/proc/sys/kernel/osrelease")
+    }
+
+    /// Check if path points to machine-id(5), /etc/hostid or /var/adm/hostid.
+    pub fn is_machine_id(&self) -> bool {
+        const MACHINE_ID: &[&[u8]] = &[
+            b"/etc/machine-id",
+            b"/etc/hostid",
+            b"/var/adm/hostid",
+            b"/sys/class/dmi/id/product_uuid",
+            b"/sys/devices/virtual/dmi/id/product_uuid",
+        ];
+        MACHINE_ID.iter().any(|f| self.is_equal(f))
+    }
+
+    /// Check if path points to `/proc/self/status` or friends.
+    pub fn is_proc_status(&self) -> bool {
+        self.starts_with(b"/proc") && self.ends_with(b"/status")
     }
 
     /// Check if path points to the `/proc/self` link.
@@ -893,7 +977,7 @@ impl XPath {
     }
 
     /// Check if path exists.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     pub fn exists(&self, follow: bool) -> bool {
         let flags = if self.is_empty() {
             return false;
@@ -943,42 +1027,24 @@ impl XPath {
 
     /// Checks if the path consists only of "." components.
     pub fn is_dot(&self) -> bool {
-        let bytes = self.as_bytes();
-
-        if bytes.is_empty() {
+        // Empty or absolute paths are not "."-only.
+        if self.is_empty() || self.is_absolute() {
             return false;
         }
 
-        let mut has_component = false;
-        let mut start = 0;
-
-        // Iterate over '/' positions without allocating.
-        #[allow(clippy::arithmetic_side_effects)]
-        for sep in One::new(b'/').iter(bytes) {
-            if sep > start {
-                // Non-empty component [start, sep).
-                let component = &bytes[start..sep];
-                has_component = true;
-                if component != b"." {
-                    return false;
-                }
-            }
-
-            // Move past '/'.
-            start = sep + 1;
-        }
-
-        // Handle the trailing component after the last '/',
-        // or the whole slice if none.
-        if start < bytes.len() {
-            let component = &bytes[start..];
-            has_component = true;
-            if component != b"." {
-                return false;
+        // Every non-empty component must be exactly ".".
+        let mut saw_component = false;
+        for comp in self
+            .as_bytes()
+            .split(|&c| c == b'/')
+            .filter(|s| !s.is_empty())
+        {
+            saw_component = true;
+            if comp != b"." {
+                return false; // rejects "..", ".hidden", "a", etc.
             }
         }
-
-        has_component
+        saw_component
     }
 
     /// Determine whether path is equal to the given string.
@@ -1006,6 +1072,16 @@ impl XPath {
         memchr(c, self.as_bytes()).is_some()
     }
 
+    /// Return the index of the given substring in path.
+    pub fn find(&self, sub: &[u8]) -> Option<usize> {
+        memmem::find(self.as_bytes(), sub)
+    }
+
+    /// Return the index of the given character in path.
+    pub fn find_char(&self, c: u8) -> Option<usize> {
+        memchr(c, self.as_bytes())
+    }
+
     /// Returns the first character of the path.
     /// Empty path returns None.
     pub fn first(&self) -> Option<u8> {
@@ -1047,6 +1123,21 @@ impl XPath {
         &self.0
     }
 
+    /// Create an `XPath` from a byte slice.
+    ///
+    /// Use the bytes up to the first nul byte if one exists;
+    /// otherwise use the whole byte slice.
+    #[inline]
+    pub fn from_bytes_until_nul(slice: &[u8]) -> &XPath {
+        // Find first nul byte if any.
+        // Use the whole slice otherwise.
+        let nullx = memchr(0, slice).unwrap_or(slice.len());
+        let slice = &slice[..nullx];
+
+        // SAFETY: XPath has repr(transparent)
+        unsafe { std::mem::transmute(slice) }
+    }
+
     /// Create a new `XPath` from a byte slice.
     pub const fn from_bytes(slice: &[u8]) -> &XPath {
         // SAFETY: XPath has repr(transparent)
@@ -1093,7 +1184,7 @@ impl XPathBuf {
 
         let mut write_pos = 0;
         let mut read_pos = 0;
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         while read_pos < len {
             if self.0[read_pos] == b'/' {
                 // Write a single slash.
@@ -1156,7 +1247,7 @@ impl XPathBuf {
         let new_len = new.len();
         let len = self.0.len();
 
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         match new_len.cmp(&old_len) {
             Ordering::Equal => {
                 // Same length: overwrite the prefix.
@@ -1191,28 +1282,123 @@ impl XPathBuf {
         Ok(())
     }
 
-    /// Create a path from the given PID.
-    pub fn from_pid(pid: Pid) -> Self {
+    /// Remove the given range of bytes from the path buffer, dropping them.
+    ///
+    /// Panics on out-of-bounds.
+    pub fn drain<R: RangeBounds<usize>>(&mut self, range: R) {
+        self.0.drain(range);
+    }
+
+    /// Create a root magiclink path from the given PID.
+    pub fn from_root(pid: Pid) -> Result<Self, Errno> {
         let mut buf = itoa::Buffer::new();
-        buf.format(pid.as_raw()).as_bytes().into()
+        let mut pfd = Vec::new();
+        let len = pid_t::MAX_STR_LEN
+            .checked_add("/root".len())
+            .ok_or(Errno::EOVERFLOW)?;
+        pfd.try_reserve(len).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(pid.as_raw()).as_bytes());
+        pfd.extend_from_slice(b"/root");
+        pfd.shrink_to_fit();
+        Ok(pfd.into())
+    }
+
+    /// Create a exe magiclink path from the given PID.
+    pub fn from_exe(pid: Pid) -> Result<Self, Errno> {
+        let mut buf = itoa::Buffer::new();
+        let mut pfd = Vec::new();
+        let len = pid_t::MAX_STR_LEN
+            .checked_add("/exe".len())
+            .ok_or(Errno::EOVERFLOW)?;
+        pfd.try_reserve(len).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(pid.as_raw()).as_bytes());
+        pfd.extend_from_slice(b"/exe");
+        pfd.shrink_to_fit();
+        Ok(pfd.into())
+    }
+
+    /// Create a cwd magiclink path from the given PID.
+    pub fn from_cwd(pid: Pid) -> Result<Self, Errno> {
+        let mut buf = itoa::Buffer::new();
+        let mut pfd = Vec::new();
+        let len = pid_t::MAX_STR_LEN
+            .checked_add("/cwd".len())
+            .ok_or(Errno::EOVERFLOW)?;
+        pfd.try_reserve(len).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(pid.as_raw()).as_bytes());
+        pfd.extend_from_slice(b"/cwd");
+        pfd.shrink_to_fit();
+        Ok(pfd.into())
+    }
+
+    /// Create a path from the given PID.
+    pub fn from_pid(pid: Pid) -> Result<Self, Errno> {
+        let mut buf = itoa::Buffer::new();
+        let mut pfd = Vec::new();
+        pfd.try_reserve(pid_t::MAX_STR_LEN).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(pid.as_raw()).as_bytes());
+        pfd.shrink_to_fit();
+        Ok(pfd.into())
+    }
+
+    /// Create a path from the given TGID and TID.
+    pub fn from_task(tgid: Pid, tid: Pid) -> Result<Self, Errno> {
+        let mut buf = itoa::Buffer::new();
+        let mut pfd = Vec::new();
+        let len = pid_t::MAX_STR_LEN
+            .checked_mul(2)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_add("/task/".len())
+            .ok_or(Errno::EOVERFLOW)?;
+        pfd.try_reserve(len).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(tgid.as_raw()).as_bytes());
+        pfd.extend_from_slice(b"/task/");
+        pfd.extend_from_slice(buf.format(tid.as_raw()).as_bytes());
+        pfd.shrink_to_fit();
+        Ok(pfd.into())
     }
 
     /// Create a path from the given FD.
-    pub fn from_fd(fd: RawFd) -> Self {
+    pub fn from_fd(fd: RawFd) -> Result<Self, Errno> {
         let mut buf = itoa::Buffer::new();
-        buf.format(fd).as_bytes().into()
+        let mut pfd = Vec::new();
+        pfd.try_reserve(RawFd::MAX_STR_LEN).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(fd).as_bytes());
+        Ok(pfd.into())
+    }
+
+    /// Create a path from the given PID and FD.
+    pub fn from_pid_fd(pid: Pid, fd: RawFd) -> Result<Self, Errno> {
+        let mut buf = itoa::Buffer::new();
+        let mut pfd = Vec::new();
+        let len = pid_t::MAX_STR_LEN
+            .checked_add(RawFd::MAX_STR_LEN)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_add("/fd/".len())
+            .ok_or(Errno::EOVERFLOW)?;
+        pfd.try_reserve(len).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(buf.format(pid.as_raw()).as_bytes());
+        pfd.extend_from_slice(b"/fd/");
+        pfd.extend_from_slice(buf.format(fd).as_bytes());
+        pfd.shrink_to_fit();
+        Ok(pfd.into())
     }
 
     /// Create a path for the given self-FD.
     ///
     /// Used for _procfs_(5) indirection.
-    pub fn from_self_fd(fd: RawFd) -> Self {
+    pub fn from_self_fd(fd: RawFd) -> Result<Self, Errno> {
         // SAFETY:
         // Use /proc/thread-self rather than /proc/self
         // because CLONE_FILES may be in effect!
-        let mut pfd = Self::from("thread-self/fd");
+        const LEN: usize = "thread-self/fd".len() + 1 + RawFd::MAX_STR_LEN;
+        let mut pfd = Vec::new();
+        pfd.try_reserve(LEN).or(Err(Errno::ENOMEM))?;
+        pfd.extend_from_slice(b"thread-self/fd");
+        let mut pfd: Self = pfd.into();
         pfd.push_fd(fd);
-        pfd
+        pfd.shrink_to_fit();
+        Ok(pfd)
     }
 
     /// Append the formatted FD as a new component.
@@ -1228,8 +1414,22 @@ impl XPathBuf {
     }
 
     /// Append a path component, managing separators correctly.
+    ///
+    /// Panics if `path` is not a single component.
     pub fn push(&mut self, path: &[u8]) {
-        if path.first() == Some(&b'/') {
+        // SAFETY: Guard against path traversals, allow leading/trailing slash.
+        let path = XPath::from_bytes(path);
+        match path.find_char(b'/') {
+            None | Some(0) => {}
+            Some(n) if n == path.len().saturating_sub(1) => {}
+            _ => unreachable!("BUG: Path traversal detected for `{path}'"),
+        }
+        assert!(
+            !path.has_parent_dot(),
+            "BUG: Path traversal detected for `{path}'"
+        );
+
+        if path.first() == Some(b'/') {
             // Absolute path replaces pbuf.
             self.0.clear();
         } else if self.last().map(|c| c != b'/').unwrap_or(true) {
@@ -1237,7 +1437,7 @@ impl XPathBuf {
             self.append_byte(b'/');
         }
         // Append new path part.
-        self.append_bytes(path);
+        self.append_bytes(path.as_bytes());
     }
 
     /// Remove the last path component.
@@ -1253,7 +1453,7 @@ impl XPathBuf {
     /// 2. Path must not have a trailing slash!
     #[inline]
     pub unsafe fn pop_unchecked(&mut self) {
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         if let Some(idx) = memrchr(b'/', &self.as_bytes()[1..]) {
             self.0.truncate(idx + 1);
         } else if self.0.len() > 1 {
@@ -1307,25 +1507,27 @@ impl XPathBuf {
         self.0.shrink_to_fit()
     }
 
+    /// Try to clone an `XPathBuf`, returning `Err(Errno::ENOMEM)` on allocation errors.
+    pub fn try_clone(&self) -> Result<Self, Errno> {
+        let mut vec = Vec::new();
+        vec.try_reserve(self.len()).or(Err(Errno::ENOMEM))?;
+        vec.extend_from_slice(self.as_bytes());
+        Ok(vec.into())
+    }
+
     /// Reserve capacity for additional more bytes to be inserted.
     /// May reserve more space to avoid frequent allocations.
     pub fn try_reserve(&mut self, additional: usize) -> Result<(), Errno> {
         self.0.try_reserve(additional).or(Err(Errno::ENOMEM))
     }
 
-    /// Create a new, empty `XPathBuf`.
-    pub fn empty() -> Self {
-        Self(vec![])
-    }
-
-    /// Construct an empty `XPathBuf` with capacity pre-allocated.
-    pub fn with_capacity(n: usize) -> Self {
-        Self(Vec::with_capacity(n))
-    }
-
-    /// Report capacity of path.
-    pub fn capacity(&self) -> usize {
-        self.0.capacity()
+    /// Resizes the `XPathBuf` in place so that `len` is equal to `new_len`.
+    ///
+    /// If `new_len` is greater than `len`, the `XPathBuf` is extended by the difference,
+    /// with each additional slot filled with value. If `new_len` is less than `len`, the
+    /// `XPathBuf` is simply truncated.
+    pub fn resize(&mut self, new_len: usize, value: u8) {
+        self.0.resize(new_len, value)
     }
 
     /// Creates an owned `XPathBuf` with path adjoined to `self`.
@@ -1411,6 +1613,30 @@ impl XPathBuf {
     pub unsafe fn set_len(&mut self, new_len: usize) {
         self.0.set_len(new_len)
     }
+
+    /// Construct an empty `XPathBuf` with capacity pre-allocated.
+    pub fn with_capacity(n: usize) -> Self {
+        Self(Vec::with_capacity(n))
+    }
+
+    /// Report capacity of path.
+    pub fn capacity(&self) -> usize {
+        self.0.capacity()
+    }
+
+    /// Constructs a new, empty `XPathBuf`.
+    ///
+    /// The internal vector will not allocate until elements are pushed onto it.
+    pub const fn empty() -> Self {
+        Self::new()
+    }
+
+    /// Constructs a new, empty `XPathBuf`.
+    ///
+    /// The internal vector will not allocate until elements are pushed onto it.
+    pub const fn new() -> XPathBuf {
+        Self(Vec::new())
+    }
 }
 
 /// Logs an untrusted Path, escaping it as hex if it contains control
@@ -1575,8 +1801,8 @@ mod tests {
             ("././", true),
             ("./././", true),
             ("././././", true),
-            ("/././", true),
-            ("/./././", true),
+            ("/././", false),
+            ("/./././", false),
             (".//././", true),
             ("", false),
             ("/", false),
@@ -1632,12 +1858,151 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_has_parent_dot() {
+        const TEST_CASES: &[(&[u8], bool)] = &[
+            (b"/home/user/..", true),
+            (b"/home/user/abc/..", true),
+            (b"/path/to/dir/..", true),
+            (b"/home/user/abc/xyz/..", true),
+            (b"/dir/..", true),
+            (b"/dir/..abc", false),
+            (b"/../", true),
+            (b"..", true),
+            (b"/..", true),
+            (b"/../../", true),
+            (b"/..file/..", true),
+            (b"/home/user/abc", false),
+            (b"/home/user/abc/xyz", false),
+            (b"/dir/abc/xyz/..file", false),
+            (b"abc/..file", false),
+            (b"dir/abc/..xyz", false),
+            (b"dir/abc/xyz/..xyz", false),
+            (b"/../", true),
+            (b"/abc/../", true),
+            (b"/abc/../xyz", true),
+            (b"/..", true),
+            (b"/dir/..file", false),
+            (b"/..file", false),
+            (b"abc/..xyz", false),
+            (b"file/..name", false),
+            (b"/..", true),
+            (b"", false),
+            (b".", false),
+            (b".a", false),
+            (b"..", true),
+            (b"...", false),
+            (b"/long/path/with/some/other/..component", false),
+            (b"/long/path/with/..other", false),
+            (b"/this/is/a/path/to/..", true),
+            (b"/home/abc/def/..", true),
+            (b"/dir/..abc/..", true),
+            (b"/path/to/../../..", true),
+            (b"/path/to/abc/../../../xyz", true),
+            (b"/../file", true),
+            (b"/file/..", true),
+            (b"/..test", false),
+            (b"/..test/file", false),
+            (b"/some/dir/..test/file", false),
+            (b"/path/with/..file", false),
+            (b"/home/user", false),
+            (b"/usr/local/bin", false),
+            (b"/test/dir", false),
+            (b"/..", true),
+            (b"/../../", true),
+            (b"/..file/..", true),
+            (b"../dir/abc/..", true),
+            (b"dir/..abc/xyz/..", true),
+            (b"/path/to/..test/dir/../file/..", true),
+            (b"/path/..to/../file", true),
+            (b"/..test/file", false),
+            (b"/test/dir/..file/..", true),
+            (b"/some/dir/..test/file/..", true),
+            (b"/path/with/..file", false),
+        ];
+
+        for (idx, &(path, expected)) in TEST_CASES.iter().enumerate() {
+            let path = XPath::from_bytes(path);
+            assert_eq!(
+                path.has_parent_dot(),
+                expected,
+                "FAIL: `{path}' at index:{idx} expected has_parent_dot:{expected}"
+            );
+        }
+    }
+
+    #[test]
+    fn test_strip_root() {
+        let cases = [
+            // Basic relative & empty
+            ("", ""),
+            (".", ""),
+            ("./", ""),
+            ("./.", ""),
+            ("././", ""),
+            ("foo", "foo"),
+            ("foo/bar", "foo/bar"),
+            ("foo/./bar", "foo/bar"),
+            ("./foo/bar", "foo/bar"),
+            // Leading root & redundant separators / dots
+            ("/", ""),
+            ("//", ""),
+            ("///", ""),
+            ("/.", ""),
+            ("/./", ""),
+            ("/./.", ""),
+            ("/././", ""),
+            ("/foo", "foo"),
+            ("/foo/", "foo"),
+            ("//foo", "foo"),
+            ("///foo", "foo"),
+            ("/./foo", "foo"),
+            ("/././foo", "foo"),
+            ("/foo/bar", "foo/bar"),
+            ("/foo/bar/", "foo/bar"),
+            ("//foo//bar///", "foo/bar"),
+            ("/./foo/./bar/.", "foo/bar"),
+            ("///.//././foo/bar", "foo/bar"),
+            // Dots inside segment names must be preserved
+            ("/foo.bar", "foo.bar"),
+            ("/foo..bar", "foo..bar"),
+            // Parent directory segments must be preserved (no over-normalization)
+            ("../foo", "../foo"),
+            ("../../foo", "../../foo"),
+            ("foo/../bar", "foo/../bar"),
+            ("/..", ".."),
+            ("/../", ".."),
+            ("/../foo", "../foo"),
+            ("/foo/..", "foo/.."),
+            ("/foo/../bar", "foo/../bar"),
+            ("/foo/../../bar", "foo/../../bar"),
+            ("/../../etc/passwd", "../../etc/passwd"),
+            ("/foo/../../../etc/passwd", "foo/../../../etc/passwd"),
+            // "Weird" but valid segments
+            ("/ foo", " foo"),
+            ("/\u{00A0}foo", "\u{00A0}foo"), // non-breaking space
+        ];
+
+        for &(input, expected) in &cases {
+            let path = XPath::from_bytes(input.as_bytes());
+            let stripped = path.strip_root();
+            let expected_path = XPath::from_bytes(expected.as_bytes());
+
+            assert_eq!(
+                stripped.as_xpath(),
+                expected_path,
+                "strip_root failed: input={input:?}, expected={expected:?}, got={:?}",
+                stripped
+            );
+        }
+    }
+
     #[test]
     fn test_path_check_file_type() {
-        assert!(XPathBuf::from("/proc")
+        xpath!("/proc")
             .check(Pid::from_raw(1), Some(&FileType::Dir), None, true, true)
-            .is_ok());
-        assert!(XPathBuf::from("/proc")
+            .unwrap();
+        xpath!("/proc")
             .check(
                 Pid::from_raw(1),
                 Some(&FileType::Dir),
@@ -1645,8 +2010,8 @@ mod tests {
                 true,
                 true,
             )
-            .is_ok());
-        assert!(XPathBuf::from("/proc")
+            .unwrap();
+        xpath!("/proc")
             .check(
                 Pid::from_raw(1),
                 Some(&FileType::Reg),
@@ -1654,38 +2019,38 @@ mod tests {
                 true,
                 true,
             )
-            .is_ok());
-        assert!(XPathBuf::from("/dev/null")
+            .unwrap();
+        xpath!("/dev/null")
             .check(Pid::from_raw(1), Some(&FileType::Chr), None, true, true)
-            .is_ok());
-        assert!(XPathBuf::from("/dev/log")
+            .unwrap();
+        xpath!("/dev/log")
             .check(Pid::from_raw(1), Some(&FileType::Sock), None, true, true)
-            .is_ok());
-        assert!(XPathBuf::from("/dev/fifo")
+            .unwrap();
+        xpath!("/dev/fifo")
             .check(Pid::from_raw(1), Some(&FileType::Fifo), None, true, true)
-            .is_ok());
-        assert!(XPathBuf::from("/dev/sda1")
+            .unwrap();
+        xpath!("/dev/sda1")
             .check(Pid::from_raw(1), Some(&FileType::Blk), None, true, true)
-            .is_err());
-        assert!(XPathBuf::from("/dev/lmao")
+            .unwrap_err();
+        xpath!("/dev/lmao")
             .check(Pid::from_raw(1), Some(&FileType::Unk), None, true, true)
-            .is_err());
+            .unwrap_err();
     }
 
     #[test]
     fn test_path_check_procfs() {
         let this = Pid::from_raw(128);
         let that = Pid::from_raw(256);
-        assert!(XPathBuf::from("/proc")
+        xpath!("/proc")
             .check(
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{this}")),
                 true,
-                true
+                true,
             )
-            .is_ok());
-        assert!(XPathBuf::from(format!("/proc/{this}"))
+            .unwrap();
+        xpath!("/proc/{this}")
             .check(
                 this,
                 Some(&FileType::Reg),
@@ -1693,8 +2058,8 @@ mod tests {
                 true,
                 true,
             )
-            .is_ok());
-        assert!(XPathBuf::from(format!("/proc/{this}"))
+            .unwrap();
+        xpath!("/proc/{this}")
             .check(
                 this,
                 Some(&FileType::Dir),
@@ -1702,26 +2067,26 @@ mod tests {
                 true,
                 true,
             )
-            .is_ok());
-        assert!(XPathBuf::from(format!("/proc/{this}/task"))
+            .unwrap();
+        xpath!("/proc/{this}/task")
             .check(
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{this}")),
                 true,
-                true
+                true,
             )
-            .is_ok());
-        assert!(XPathBuf::from("/proc")
+            .unwrap();
+        xpath!("/proc")
             .check(
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{that}")),
                 true,
-                true
+                true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{that}"))
+            .unwrap_err();
+        xpath!("/proc/{that}")
             .check(
                 this,
                 Some(&FileType::Reg),
@@ -1729,8 +2094,8 @@ mod tests {
                 true,
                 true,
             )
-            .is_ok());
-        assert!(XPathBuf::from(format!("/proc/{that}"))
+            .unwrap();
+        xpath!("/proc/{that}")
             .check(
                 this,
                 Some(&FileType::Dir),
@@ -1738,31 +2103,41 @@ mod tests {
                 true,
                 true,
             )
-            .is_ok());
-        assert!(XPathBuf::from(format!("/proc/{that}/task"))
+            .unwrap();
+        xpath!("/proc/{that}/task")
             .check(
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{that}")),
                 true,
-                true
+                true,
             )
-            .is_ok());
+            .unwrap();
     }
 
     #[test]
     fn test_path_check_procfs_syd_leader() {
         let syd = Pid::this();
-        assert!(XPathBuf::from("/proc")
+        let pid = Pid::from_raw(syd.as_raw() + 1);
+        xpath!("/proc")
             .check(
                 syd,
                 Some(&FileType::Dir),
                 Some(&xpath!("{syd}")),
                 true,
-                true
+                true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{syd}"))
+            .unwrap();
+        xpath!("/proc")
+            .check(
+                pid,
+                Some(&FileType::Dir),
+                Some(&xpath!("{syd}")),
+                true,
+                true,
+            )
+            .unwrap_err();
+        xpath!("/proc/{syd}")
             .check(
                 syd,
                 Some(&FileType::Reg),
@@ -1770,30 +2145,49 @@ mod tests {
                 true,
                 true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{syd}"))
+            .unwrap();
+        xpath!("/proc/{syd}")
             .check(
-                syd,
+                pid,
+                Some(&FileType::Reg),
+                Some(&XPath::from_bytes(b"")),
+                true,
+                true,
+            )
+            .unwrap_err();
+        xpath!("/proc/{syd}")
+            .check(
+                pid,
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
                 true,
                 true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{syd}/task"))
+            .unwrap_err();
+        xpath!("/proc/{syd}/task")
             .check(
                 syd,
                 Some(&FileType::Dir),
                 Some(&xpath!("{syd}")),
                 true,
-                true
+                true,
             )
-            .is_err());
+            .unwrap();
+        xpath!("/proc/{syd}/task")
+            .check(
+                pid,
+                Some(&FileType::Dir),
+                Some(&xpath!("{syd}")),
+                true,
+                true,
+            )
+            .unwrap_err();
     }
 
     #[test]
     fn test_path_check_procfs_syd_thread() {
         // Spawn a new thread.
+        let pid = Pid::this();
         let tid = {
             let (tx, rx) = mpsc::channel();
             thread::spawn(move || {
@@ -1802,16 +2196,16 @@ mod tests {
             });
             rx.recv().unwrap()
         };
-        assert!(XPathBuf::from("/proc")
+        xpath!("/proc")
             .check(
                 tid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{tid}")),
                 true,
-                true
+                true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{tid}"))
+            .unwrap();
+        xpath!("/proc/{tid}")
             .check(
                 tid,
                 Some(&FileType::Reg),
@@ -1819,8 +2213,8 @@ mod tests {
                 true,
                 true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{tid}"))
+            .unwrap();
+        xpath!("/proc/{tid}")
             .check(
                 tid,
                 Some(&FileType::Dir),
@@ -1828,16 +2222,52 @@ mod tests {
                 true,
                 true,
             )
-            .is_err());
-        assert!(XPathBuf::from(format!("/proc/{tid}/task"))
+            .unwrap();
+        xpath!("/proc/{tid}/task")
             .check(
                 tid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{tid}")),
                 true,
-                true
+                true,
             )
-            .is_err());
+            .unwrap();
+        xpath!("/proc")
+            .check(
+                pid,
+                Some(&FileType::Dir),
+                Some(&xpath!("{tid}")),
+                true,
+                true,
+            )
+            .unwrap_err();
+        xpath!("/proc/{tid}")
+            .check(
+                pid,
+                Some(&FileType::Reg),
+                Some(&XPath::from_bytes(b"")),
+                true,
+                true,
+            )
+            .unwrap_err();
+        xpath!("/proc/{tid}")
+            .check(
+                pid,
+                Some(&FileType::Dir),
+                Some(&XPath::from_bytes(b"")),
+                true,
+                true,
+            )
+            .unwrap_err();
+        xpath!("/proc/{tid}/task")
+            .check(
+                pid,
+                Some(&FileType::Dir),
+                Some(&xpath!("{tid}")),
+                true,
+                true,
+            )
+            .unwrap_err();
     }
 
     #[test]
@@ -1915,7 +2345,7 @@ mod tests {
     }
 
     #[test]
-    fn test_path_push() {
+    fn test_path_push_simple() {
         // Pushing a relative path extends the existing path.
         let mut path = XPathBuf::from("/tmp");
         path.push(b"file.bk");
@@ -1937,6 +2367,20 @@ mod tests {
         assert_eq!(path.as_os_str().as_bytes(), b"/tmp/");
     }
 
+    #[test]
+    #[should_panic]
+    fn test_path_push_dotdot_simple() {
+        let mut path = XPathBuf::from("/tmp");
+        path.push(b"..");
+    }
+
+    #[test]
+    #[should_panic]
+    fn test_path_push_dotdot_with_trailing_slash() {
+        let mut path = XPathBuf::from("/tmp");
+        path.push(b"../");
+    }
+
     #[test]
     fn test_path_split() {
         // Test typical path without trailing slash
@@ -2001,6 +2445,7 @@ mod tests {
             "valid_filename.txt",
             "hello_world",
             "File123",
+            "Makefile",
             "こんにちは", // Japanese characters
             "文件",       // Chinese characters
             "emoji😀",    // Starts with permitted character
