@@ -10,24 +10,23 @@ use std::{
 };
 
 use btoi::btoi;
-use libloading::os::unix::Symbol;
+use memchr::arch::all::is_prefix;
 use nix::{
     errno::Errno,
     libc,
     sys::signal::{kill, sigprocmask, SigSet, SigmaskHow, Signal},
-    unistd::{close, getpid, read, setsid, tcsetpgrp, write},
+    unistd::{execvp, getpid, read, setsid, tcsetpgrp, write},
 };
 
 use crate::{
-    caps::{securebits::set_keepcaps, Capability},
+    caps::Capability,
+    compat::{set_name, set_pdeathsig},
     config::*,
     confine::{confine_scmp_ioctl, confine_scmp_pwritev2, safe_drop_cap},
-    fs::retry_on_eintr,
+    fs::close,
     landlock::{CompatLevel, Compatible, RestrictSelfFlags, Ruleset, RulesetAttr, Scope},
-    unshare::{
-        error::ErrorCode as Err,
-        run::{ChildInfo, Exe},
-    },
+    retry::retry_on_eintr,
+    unshare::{error::ErrorCode as Err, run::ChildInfo},
 };
 
 unsafe fn fail_errno(code: Err, errno: i32) -> ! {
@@ -68,11 +67,40 @@ macro_rules! fail_errno_safe {
     };
 }
 
-#[allow(clippy::cognitive_complexity)]
-pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
+#[expect(clippy::cognitive_complexity)]
+pub fn child_after_clone(arg: *mut libc::c_void) -> ! {
+    // Set process name, ignore errors.
+    let _ = set_name(c"syd_exec");
+
     // SAFETY: arg is a valid ChildInfo structure.
     let mut child: Box<ChildInfo> = unsafe { Box::from_raw(arg as *mut ChildInfo) };
 
+    if let Some(&sig) = child.cfg.death_sig.as_ref() {
+        if let Err(errno) = set_pdeathsig(Some(sig)) {
+            fail_errno_safe!(child, Err::ParentDeathSignal, errno as i32);
+        }
+    }
+
+    if child.cfg.restore_sigmask {
+        // Reset blocking signals.
+        // Step 1: Reset the signal mask using pthread_sigmask.
+        unsafe {
+            let mut sigmask: libc::sigset_t = mem::zeroed();
+            libc::sigemptyset(&raw mut sigmask);
+            libc::pthread_sigmask(libc::SIG_SETMASK, &raw const sigmask, ptr::null_mut());
+        }
+        // Step 2: Unblock all signals using sigprocmask.
+        let sigmask = SigSet::all();
+        if let Err(errno) = sigprocmask(SigmaskHow::SIG_UNBLOCK, Some(&sigmask), None) {
+            fail_errno_safe!(child, Err::ResetSignal, errno as i32);
+        }
+
+        // Reset all signals to their default dispositions.
+        if let Err(errno) = crate::reset_signals() {
+            fail_errno_safe!(child, Err::ResetSignal, errno as i32);
+        }
+    }
+
     // Restriction 0: Change controlling terminal to PTY as necessary.
     if let Some(pty_fd) = child.pty_fd.take() {
         // SAFETY: pty_fd is a valid FD.
@@ -111,14 +139,12 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
     // Apply a Landlock scope sandbox to restrict
     // 1. Ptrace attach outside Landlock.
     // 2. Signal send outside Landlock.
-    // 3. UNIX abstract socket connect outside Landlock.
     // We leave path and network restrictions for Landlock
     // to be configured by the user using Lock sandboxing.
     // This is an added layer for hardening and it's best-effort.
     let mut ruleset = Ruleset::default();
     let ruleset_ref = &mut ruleset;
     ruleset_ref.set_compatibility(CompatLevel::BestEffort);
-    let _ = ruleset_ref.scope(Scope::AbstractUnixSocket);
     let _ = ruleset_ref.scope(Scope::Signal);
     if let Ok(ruleset) = ruleset.create() {
         let _ = ruleset.restrict_self(RestrictSelfFlags::empty());
@@ -160,15 +186,23 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
             None
         };
         if let Some(fd) = fd {
-            let _ = close(fd);
+            if fd >= 0 {
+                let _ = close(fd);
+            }
         }
     }
 
+    // Passthrough RUST_BACKTRACE to the sandbox process.
+    match env::var_os("SYD_RUST_BACKTRACE") {
+        Some(val) => env::set_var("RUST_BACKTRACE", val),
+        None => env::remove_var("RUST_BACKTRACE"),
+    }
+
     // SAFETY: Clean Syd environment variables from process environment.
     // Note, we have just used ENV_LOG_FD above and do not need it anymore.
     for (key, _) in env::vars_os() {
-        if key.as_bytes().starts_with(b"CARGO_BIN_EXE_syd")
-            || (key.as_bytes().starts_with(b"SYD_") && !key.as_bytes().starts_with(b"SYD_TEST_"))
+        if is_prefix(key.as_bytes(), b"CARGO_BIN_EXE_syd")
+            || (is_prefix(key.as_bytes(), b"SYD_") && !is_prefix(key.as_bytes(), b"SYD_TEST_"))
         {
             env::remove_var(key);
         }
@@ -183,34 +217,6 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
     drop(child.seccomp_pipefd.0 .1);
     drop(child.seccomp_pipefd.1 .0);
 
-    if let Some(&sig) = child.cfg.death_sig.as_ref() {
-        if let Err(errno) = Errno::result(unsafe {
-            libc::prctl(libc::PR_SET_PDEATHSIG, sig as libc::c_ulong, 0, 0, 0)
-        }) {
-            fail_errno_safe!(child, Err::ParentDeathSignal, errno as i32);
-        }
-    }
-
-    if child.cfg.restore_sigmask {
-        // Reset blocking signals.
-        // Step 1: Reset the signal mask using pthread_sigmask.
-        unsafe {
-            let mut sigmask: libc::sigset_t = mem::zeroed();
-            libc::sigemptyset(&raw mut sigmask);
-            libc::pthread_sigmask(libc::SIG_SETMASK, &raw const sigmask, ptr::null_mut());
-        }
-        // Step 2: Unblock all signals using sigprocmask.
-        let sigmask = SigSet::all();
-        if let Err(errno) = sigprocmask(SigmaskHow::SIG_UNBLOCK, Some(&sigmask), None) {
-            fail_errno_safe!(child, Err::ResetSignal, errno as i32);
-        }
-
-        // Reset all signals to their default dispositions.
-        if let Err(errno) = crate::reset_signals() {
-            fail_errno_safe!(child, Err::ResetSignal, errno as i32);
-        }
-    }
-
     if let Some(callback) = &child.pre_exec {
         if let Err(errno) = callback() {
             fail_errno_safe!(child, Err::PreExec, errno as i32);
@@ -269,7 +275,7 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         let fd = seccomp_fd.as_raw_fd().to_le_bytes();
         let mut nwrite = 0;
         while nwrite < fd.len() {
-            #[allow(clippy::arithmetic_side_effects)]
+            #[expect(clippy::arithmetic_side_effects)]
             match retry_on_eintr(|| write(&pipe_rw, &fd[nwrite..])) {
                 Ok(0) => {
                     // Parent died before reading.
@@ -348,31 +354,11 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         fail_safe!(child, Err::CapSet);
     }
 
-    // Ensure we keep the specified capabilities.
-    if set_keepcaps(true).is_err() {
-        fail_safe!(child, Err::CapSet);
+    // Exit immediately if export mode is in effect.
+    if env::var_os(ENV_DUMP_SCMP).is_some() {
+        unsafe { libc::_exit(0) };
     }
 
-    // TODO:
-    // Set the new secure bits:
-    // 1. SECBIT_EXEC_RESTRICT_FILE
-    // 2. SECBIT_EXEC_DENY_INTERACTIVE
-    // upon user configuration.
-    // See: https://docs.kernel.org/userspace-api/check_exec.html
-    // Note, we already use AT_EXECVE_CHECK.
-
-    match child.exe {
-        Exe::Library(lib) => unsafe {
-            let fun: Symbol<unsafe extern "C" fn() -> i32> = match lib.get(b"syd_main") {
-                Ok(fun) => fun,
-                Err(_) => nix::libc::_exit(nix::libc::EINVAL),
-            };
-            nix::libc::_exit(fun());
-        },
-        Exe::Program((filename, ref args)) => {
-            let args = &args[..];
-            unsafe { libc::execvp(filename, args.as_ptr()) };
-            fail_safe!(child, Err::Exec);
-        }
-    }
+    let Err(errno) = execvp(&child.exe_file, &child.exe_args);
+    fail_errno_safe!(child, Err::Exec, errno as i32);
 }
