@@ -8,6 +8,8 @@ pub struct Config {
     pub deny_tsc: bool,
     pub keep: bool,
     pub stop: bool,
+    pub ssb: bool,
+    pub append_only: bool,
 }
 
 impl Default for Config {
@@ -19,6 +21,8 @@ impl Default for Config {
             deny_tsc: false,
             keep: false,
             stop: false,
+            ssb: false,
+            append_only: false,
         }
     }
 }
