@@ -23,7 +23,7 @@ pub enum ErrorCode {
     SetTSC = 16,
 }
 
-/// Error runnning process
+/// Error running process
 ///
 /// This type has very large number of options and it's enum only to be
 /// compact. Probably you shouldn't match on the error cases but just format
@@ -118,7 +118,7 @@ impl Error {
             BeforeUnfreeze(_) => "error in before_unfreeze callback",
             PreExec(_) => "error in pre_exec callback",
             ProcessStop(_) => "error stopping process",
-            ResetSignal(_) => "error reseting signals",
+            ResetSignal(_) => "error resetting signals",
             Seccomp(_) => "error in seccomp filter load",
             SeccompFilterIoctl(_) => "error filtering ioctl requests with seccomp",
             SeccompFilterAppendOnly(_) => "error filtering pwritev2 requests with seccomp",
