@@ -1,6 +1,6 @@
 //
 // Syd: rock-solid application kernel
-// src/syd-cat.rs: Print out the rules of the given sandbox profile and exit
+// src/utils/syd-cat.rs: Print out the rules of the given sandbox profile and exit
 //
 // Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
 //
@@ -18,6 +18,22 @@ use syd::{
     sandbox::Sandbox,
 };
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 syd::main! {
     use lexopt::prelude::*;
 
@@ -63,13 +79,13 @@ syd::main! {
     }
 
     if optj && optJ {
-        eprintln!("-j and -J are mutually exclusive!");
+        eprintln!("syd-cat: -j and -J are mutually exclusive!");
         return Err(Errno::EINVAL.into());
     }
 
     if let Some(name) = name {
         if optj || optJ {
-            eprintln!("-p cannot be used with JSON output!");
+            eprintln!("syd-cat: -p cannot be used with JSON output!");
             return Err(Errno::EINVAL.into());
         }
         if name == "list" {
@@ -91,7 +107,7 @@ syd::main! {
         let ips_ext = XPath::from_bytes(b"ipset");
         let net_ext = XPath::from_bytes(b"netset");
 
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         if *fext == *syd_ext {
             syd.parse_config_file(&path)?;
         } else if *fext == *ips_ext || *fext == *net_ext {
@@ -144,13 +160,18 @@ fn help() {
 fn list() {
     println!("container");
     println!("core");
+    println!("cwd");
     println!("debug");
     println!("enforce");
     println!("firefox");
+    println!("fs");
+    println!("hide");
     println!("immutable");
     println!("landlock");
     println!("lib");
     println!("linux");
+    println!("ltp");
+    println!("nix");
     println!("noipv4");
     println!("noipv6");
     println!("nomem");
@@ -208,6 +229,15 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "ltp" => {
+            println!("# Syd profile: LTP");
+            println!("# Number of rules: {}", syd::config::PROFILE_LTP.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_LTP {
+                println!("{command}");
+            }
+        }
         "kvm" => {
             println!("# Syd profile: KVM");
             println!("# Number of rules: {}", syd::config::PROFILE_KVM.len());
@@ -229,6 +259,15 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "nix" => {
+            println!("# Syd profile: NIX");
+            println!("# Number of rules: {}", syd::config::PROFILE_NIX.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_NIX {
+                println!("{command}");
+            }
+        }
         "tty" => {
             println!("# Syd profile: TTY");
             println!("# Number of rules: {}", syd::config::PROFILE_TTY.len());
@@ -259,6 +298,24 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "cwd" | "pwd" => {
+            println!("# Syd profile: CWD");
+            println!("# Number of rules: {}", syd::config::PROFILE_CWD.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_CWD {
+                println!("{command}");
+            }
+        }
+        "hide" => {
+            println!("# Syd profile: Hide");
+            println!("# Number of rules: {}", syd::config::PROFILE_HIDE.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_HIDE {
+                println!("{command}");
+            }
+        }
         "noipv4" => {
             println!("# Syd profile: NoIpv4");
             println!("# Number of rules: {}", syd::config::PROFILE_NOIPV4.len());
@@ -370,6 +427,15 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "fs" => {
+            println!("# Syd profile: Filesystem");
+            println!("# Number of rules: {}", syd::config::PROFILE_FS.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_FS {
+                println!("{command}");
+            }
+        }
         "off" => {
             println!("# Syd profile: Off");
             println!("# Number of rules: {}", syd::config::PROFILE_OFF.len());
@@ -415,6 +481,18 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "nixstore" => {
+            println!("# Syd profile: Nix store");
+            println!(
+                "# Number of rules: {}",
+                syd::config::PROFILE_NIX_STORE.len()
+            );
+            println!("# Copyright (c) 2025 Emery Hemingway <emery+syd@slow.janky.email>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_NIX_STORE {
+                println!("{command}");
+            }
+        }
         _ => return Err(Errno::EINVAL.into()),
     }
 
