@@ -1,6 +1,6 @@
 //
 // Syd: rock-solid application kernel
-// src/syd-cpu.rs: Print the number of CPUs.
+// src/utils/syd-cpu.rs: Print the number of CPUs.
 //
 // Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
 //
@@ -8,9 +8,23 @@
 
 use std::process::ExitCode;
 
-use syd::err::SydResult;
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
 
-fn main() -> SydResult<ExitCode> {
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
+syd::main! {
     use lexopt::prelude::*;
 
     syd::set_sigpipe_dfl()?;
