@@ -1,6 +1,6 @@
 //
 // Syd: rock-solid application kernel
-// src/syd-hex.rs: Hexadecimal encode/decode standard input.
+// src/utils/syd-hex.rs: Hexadecimal encode/decode standard input.
 //
 // Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
 //
@@ -17,6 +17,22 @@ use data_encoding::{HEXLOWER, HEXLOWER_PERMISSIVE};
 use memchr::arch::all::is_equal;
 use nix::{errno::Errno, unistd::isatty};
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 syd::main! {
     use lexopt::prelude::*;
 
@@ -74,7 +90,7 @@ syd::main! {
     }
 
     // Lock stdin for efficient reading
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let mut input: Box<dyn BufRead> = match opt_input {
         None => Box::new(std::io::stdin().lock()),
         Some(path) if is_equal(path.as_bytes(), b"-") => Box::new(std::io::stdin().lock()),
