@@ -1,6 +1,6 @@
 //
 // Syd: rock-solid application kernel
-// src/syd-ls.rs: Print the names of the system calls which belong to the given set and exit
+// src/utils/syd-ls.rs: Print the names of the system calls which belong to the given set and exit
 //                If set is prctl, print the list of allowed prctl options
 //                If set is personality, print the list of allowed personalities.
 //
@@ -24,6 +24,22 @@ use syd::{
     compat::getdents64, config::DIRENT_BUF_SIZE, err::SydResult, hash::SydHashSet, path::mask_path,
 };
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 syd::main! {
     syd::set_sigpipe_dfl()?;
 
@@ -40,17 +56,21 @@ syd::main! {
             println!("Print the names of the system calls which belong to the given set and exit.");
             println!("If set is drop, print the list of capabilities that are dropped at startup.");
             println!("If set is env, print the list of unsafe environment variables.");
-            println!("If set is prctl, print the list of allowed prctl options.");
+            println!("If set is madvise, print the list of allowed madvise(2) advice.");
+            println!("If set is prctl, print the list of allowed prctl(2) options.");
             println!("If set is personality, print the list of allowed personalities.");
+            println!("If set is setsockopt, print the list of denied setsockopt(2) options.");
             println!("Available sets are:");
             println!("- cpu");
+            println!("- dead");
             println!("- deny");
+            println!("- deprecated");
             println!("- ebpf");
             println!("- futex");
             println!("- hook");
             println!("- keyring");
+            println!("- mount");
             println!("- nice");
-            println!("- noop");
             println!("- page_cache");
             println!("- perf");
             println!("- pkey");
@@ -93,6 +113,16 @@ syd::main! {
                 println!("{name}");
             }
         }
+        Some("dead") => {
+            for name in syd::config::DEAD_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("deprecated") => {
+            for name in syd::config::DEPRECATED_SYSCALLS {
+                println!("{name}");
+            }
+        }
         Some("ebpf") => {
             for name in syd::config::EBPF_SYSCALLS {
                 println!("{name}");
@@ -113,8 +143,8 @@ syd::main! {
                 println!("{name}");
             }
         }
-        Some("noop") => {
-            for name in syd::config::NOOP_SYSCALLS {
+        Some("mount") => {
+            for name in syd::config::MOUNT_SYSCALLS {
                 println!("{name}");
             }
         }
@@ -179,8 +209,54 @@ syd::main! {
                 println!("{name}");
             }
         }
+        Some("madvise") => {
+            for (name, _) in syd::config::ALLOW_MADVISE {
+                println!("{name}");
+            }
+        }
         Some("prctl") => {
-            for (name, _) in syd::config::ALLOWLIST_PRCTL {
+            for (name, _) in syd::config::ALLOW_PRCTL {
+                println!("{name}");
+            }
+        }
+        Some("setsockopt") => {
+            for (level, optname) in syd::config::DENY_SETSOCKOPT {
+                println!("{level:#x}:{optname:#x}");
+            }
+        }
+        // Syscall sets for Syd threads, undocumented.
+        Some("syd_emu") => {
+            for name in syd::config::EMU_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("syd_int") => {
+            for name in syd::config::INT_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("syd_out") => {
+            for name in syd::config::OUT_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("syd_ipc") => {
+            for name in syd::config::IPC_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("syd_aes") => {
+            for name in syd::config::AES_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("syd_main") => {
+            for name in syd::config::MAIN_SYSCALLS {
+                println!("{name}");
+            }
+        }
+        Some("syd_oci") => {
+            for name in syd::config::OCI_SYSCALLS {
                 println!("{name}");
             }
         }
@@ -195,7 +271,7 @@ syd::main! {
 
 fn readdir_cwd() -> SydResult<()> {
     // Open a file descriptor to the current directory.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let cwd = OpenOptions::new()
         .read(true)
         .custom_flags(libc::O_DIRECTORY)
