@@ -12,7 +12,7 @@ use std::{
     process::{exit, ExitCode},
 };
 
-use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
+use libseccomp::{scmp_cmp, ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     fcntl::{fcntl, splice, FcntlArg, OFlag, SpliceFFlags},
@@ -20,24 +20,24 @@ use nix::{
     sched::{unshare, CloneFlags},
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
-        prctl::set_dumpable,
+        resource::Resource,
         signal::{signal, sigprocmask, SigHandler, SigmaskHow, Signal},
         signalfd::{SfdFlags, SigSet, SignalFd},
         termios::{cfmakeraw, tcgetattr, tcsetattr, OutputFlags, SetArg, Termios},
     },
-    unistd::{chdir, chroot, close, pipe2},
+    unistd::{chdir, chroot, pipe2},
 };
 use syd::{
-    compat::epoll_ctl_safe,
-    confine::{confine_mdwe, extend_ioctl},
+    compat::{epoll_ctl_safe, set_dumpable},
+    config::VDSO_SYSCALLS,
+    confine::{confine_mdwe, confine_rlimit_zero, confine_scmp_madvise, extend_ioctl},
     err::SydResult,
-    fs::{closeexcept, duprand, set_exclusive, set_nonblock, winsize_get, winsize_set},
+    fs::{close, closeexcept, set_exclusive, set_nonblock},
     ignore_signals,
-    ioctl::{
-        TCGETS, TCGETS2, TCSETS, TCSETS2, TCSETSF, TCSETSF2, TCSETSW, TCSETSW2, TIOCGWINSZ,
-        TIOCSWINSZ,
-    },
+    io::{winsize_get, winsize_set},
+    ioctl::IoctlMap,
     landlock_policy::LandlockPolicy,
+    rng::duprand,
     IgnoreSignalOpts,
 };
 
@@ -52,7 +52,7 @@ syd::main! {
     // Parse options.
     let opts = parse_options()?;
 
-    // Ignore all signals except SIG{KILL,STOP,PIPE,CHLD}.
+    // Ignore all signals except SIG{CHLD,KILL,STOP}.
     // This is used to ensure we can deny {rt_,}sigreturn(2) to mitigate SROP.
     ignore_signals(IgnoreSignalOpts::empty())?;
 
@@ -61,7 +61,7 @@ syd::main! {
     // 2. The PID fd and the PTY fd passed by the Syd process.
     // Nothing can be done on closeexcept errors.
     // We do it early here so FD randomization doesn't effect performance.
-    #[allow(clippy::cast_sign_loss)]
+    #[expect(clippy::cast_sign_loss)]
     {
         let fd1 = opts.fpid.as_raw_fd() as libc::c_uint;
         let fd2 = opts.fpty.as_raw_fd() as libc::c_uint;
@@ -205,7 +205,7 @@ syd::main! {
     );
 
     if let Some(tio) = tio {
-        let _ = tcsetattr(&fstd_rd, SetArg::TCSAFLUSH, &tio);
+        let _ = tcsetattr(&fstd_rd, SetArg::TCSANOW, &tio);
     }
 
     Ok(match result {
@@ -215,7 +215,7 @@ syd::main! {
 }
 
 /// Run the PTY forwarder.
-#[allow(clippy::too_many_arguments)]
+#[expect(clippy::too_many_arguments)]
 fn run_pty_forwarder<
     F1: AsFd,
     F2: AsFd,
@@ -393,11 +393,12 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
         "exit_group",
         // can handle signals limitedly.
         "sigaltstack",
-        // can handle memory limitedly.
+        // can {{dr}e,}allocate memory.
+        // mmap{,2} and mprotect are further confined to disable PROT_EXEC.
         "brk",
+        //"madvise", advice are confined.
+        "mremap",
         "munmap",
-        // can generate random bytes.
-        "getrandom",
         // can close files.
         "close",
         // can do I/O with splice.
@@ -411,14 +412,17 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     ];
 
     // Default allowlist.
-    for name in allow_call {
+    for name in allow_call.iter().chain(VDSO_SYSCALLS) {
         if let Ok(syscall) = ScmpSyscall::from_name(name) {
             ctx.add_rule(ScmpAction::Allow, syscall)?;
         }
     }
 
+    // Allow safe madvise(2) advice.
+    confine_scmp_madvise(&mut ctx)?;
+
     // Allow read(2) to the signal fd only.
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     let syscall = ScmpSyscall::from_name("read").unwrap();
     ctx.add_rule_conditional(
         ScmpAction::Allow,
@@ -429,35 +433,33 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     // Allow ioctl(2) requests:
     // 1. TCGETS{,2}, aka tcgetattr(3)
     // 2. TCSETS{,2}, aka tcsetattr(3) with TCSANOW
-    // 3. TCSETSW{,2}, aka tcsetattr(3) with TCSADRAIN
-    // 4. TCSETSF{,2}, aka tcsetattr(3) with TCSAFLUSH
-    // 5. TIOCGWINSZ, aka winsize_get
-    // 6. TIOCSWINSZ, aka winsize_set
+    // 3. TIOCGWINSZ, aka winsize_get
+    // 4. TIOCSWINSZ, aka winsize_set
     //
     // For *2, we use hardcoded versions because not all libcs define them.
     // They are portable as `struct termios2` has the same size
     // across 32-bit and 64-bit architectures.
-    #[allow(clippy::disallowed_methods)]
-    let syscall = ScmpSyscall::from_name("ioctl").unwrap();
-    #[allow(clippy::unnecessary_cast)]
-    for request in [
-        *TCGETS,
-        *TCGETS2,
-        *TCSETS,
-        *TCSETS2,
-        *TCSETSW,
-        *TCSETSW2,
-        *TCSETSF,
-        *TCSETSF2,
-        *TIOCGWINSZ,
-        *TIOCSWINSZ,
-    ] {
-        let request = if let Some(request) = request {
-            request as u64
-        } else {
-            continue;
-        };
+    let arch = ScmpArch::native();
+    let ioctl = IoctlMap::new(None, true);
+    let names = [
+        "TCGETS",
+        "TCSETS",
+        "TCGETS2",
+        "TCSETS2",
+        "TIOCGWINSZ",
+        "TIOCSWINSZ",
+    ];
+    let mut iotty = Vec::with_capacity(names.len());
+    for name in names {
+        if let Some(req) = ioctl.get_value(name, arch) {
+            #[allow(clippy::unnecessary_cast)]
+            iotty.push(req as u64);
+        }
+    }
 
+    #[expect(clippy::disallowed_methods)]
+    let syscall = ScmpSyscall::from_name("ioctl").unwrap();
+    for request in iotty {
         ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == request)])?;
         if let Some(request) = extend_ioctl(request) {
             ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == request)])?;
@@ -476,6 +478,25 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
         }
     }
 
+    // Allow safe prctl(2) operations.
+    let sysname = "prctl";
+    if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
+        let op = libc::PR_SET_VMA as u64;
+        ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == op)])?;
+    }
+
+    // Prevent executable memory.
+    const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
+    for name in ["mmap", "mmap2", "mprotect"] {
+        #[expect(clippy::disallowed_methods)]
+        let syscall = ScmpSyscall::from_name(name).unwrap();
+        ctx.add_rule_conditional(
+            ScmpAction::Allow,
+            syscall,
+            &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
+        )?;
+    }
+
     // We will ignore unshare errors next step and here we keep
     // with the expectation that we're inside the safe directory.
     chdir("/proc/self/fdinfo")?;
@@ -518,10 +539,12 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
         // Set the process dumpable attribute to not-dumpable.
         let _ = set_dumpable(false);
 
-        // Deny reading the timestamp counter (x86 only).
-        // SAFETY: Our nix version does not have a wrapper for SET_TSC yet.
-        #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-        let _ = unsafe { libc::prctl(libc::PR_SET_TSC, libc::PR_TSC_SIGSEGV) };
+        // Set nfiles, nprocs, and filesize rlimits to zero.
+        confine_rlimit_zero(&[
+            Resource::RLIMIT_FSIZE,
+            Resource::RLIMIT_NOFILE,
+            Resource::RLIMIT_NPROC,
+        ])?;
     }
 
     if print_rules {
@@ -635,14 +658,14 @@ fn refresh_pty<Fd1: AsFd, Fd2: AsFd>(src: Fd1, dst: Fd2) -> Option<Termios> {
     let orig_tio = tio.clone();
 
     // Inherit host terminal settings for PTY.
-    let _ = tcsetattr(&dst, SetArg::TCSAFLUSH, &tio);
+    let _ = tcsetattr(&dst, SetArg::TCSANOW, &tio);
 
     // Set raw mode for input TTY.
     // Enable output processing.
     cfmakeraw(&mut tio);
     tio.output_flags.insert(OutputFlags::OPOST);
 
-    let _ = tcsetattr(&src, SetArg::TCSAFLUSH, &tio);
+    let _ = tcsetattr(&src, SetArg::TCSANOW, &tio);
 
     Some(orig_tio)
 }
