@@ -1,6 +1,6 @@
 //
 // Syd: rock-solid application kernel
-// src/syd-sh.rs: confined shell
+// src/utils/syd-sh.rs: confined shell
 //
 // Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
 //
@@ -9,27 +9,72 @@
 use std::{
     env,
     fs::File,
-    io::{Read, Seek, SeekFrom, Write},
-    os::fd::AsRawFd,
+    io::{Read, Seek, SeekFrom, Stdin, Write},
+    os::fd::{AsFd, AsRawFd, BorrowedFd},
     process::{exit, ExitCode},
-    time::Duration,
 };
 
+use dur::Duration;
 use linefeed::{Interface, ReadResult};
 use nix::{
     errno::Errno,
     unistd::{isatty, Gid, Uid},
 };
 use syd::{
+    compat::MFdFlags,
     config::*,
+    cookie::safe_memfd_create,
     debug,
-    fs::{create_memfd, safe_copy_if_exists, seal_memfd, set_cloexec, MFD_ALLOW_SEALING},
+    fs::{seal_memfd_all, set_cloexec},
     get_user_home, get_user_name, human_size,
-    path::{XPath, XPathBuf},
+    io::ReadFd,
+    lookup::safe_copy_if_exists,
+    path::XPathBuf,
     syslog::LogLevel,
     wordexp::*,
 };
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
+enum Input {
+    File(File),
+    Stdin(Stdin),
+}
+
+impl Read for Input {
+    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
+        match self {
+            Input::File(f) => f.read(buf),
+            Input::Stdin(s) => s.read(buf),
+        }
+    }
+}
+
+impl AsFd for Input {
+    fn as_fd(&self) -> BorrowedFd<'_> {
+        match self {
+            Input::File(f) => f.as_fd(),
+            Input::Stdin(s) => s.as_fd(),
+        }
+    }
+}
+
+impl ReadFd for Input {}
+
 syd::main! {
     use lexopt::prelude::*;
 
@@ -87,7 +132,9 @@ syd::main! {
     // Create a memory fd to write input into,
     // and pass to the internal /bin/sh invoked
     // by wordexp(3).
-    let mut file = create_memfd(b"syd-sh\0", MFD_ALLOW_SEALING).map(File::from)?;
+    let mut file = safe_memfd_create(
+        c"syd-sh",
+        MFdFlags::MFD_ALLOW_SEALING | MFdFlags::MFD_CLOEXEC).map(File::from)?;
     debug!("ctx": "sh",
         "msg": format!("created memory-file {} with close-on-exec flag set",
             file.as_raw_fd()));
@@ -106,10 +153,10 @@ syd::main! {
 
     // Handle system-wide configuration.
     if optl {
-        safe_copy_if_exists(&mut file, XPath::from_bytes(b"/etc/syd/init_login.sh"))?;
+        safe_copy_if_exists(&mut file, "/etc/syd/init_login.sh")?;
         file.write_all(b"\n")?;
     }
-    safe_copy_if_exists(&mut file, XPath::from_bytes(b"/etc/syd/init.sh"))?;
+    safe_copy_if_exists(&mut file, "/etc/syd/init.sh")?;
     file.write_all(b"\n")?;
 
     // Handle user-specific configuration.
@@ -154,7 +201,7 @@ syd::main! {
             eprintln!("+ {input}");
         }
 
-        seal_memfd(&file)?;
+        seal_memfd_all(&file)?;
         debug!("ctx": "sh",
             "msg": format!("sealed memory-file {} against grows, shrinks and writes",
                 file.as_raw_fd()));
@@ -183,30 +230,29 @@ syd::main! {
         };
     }
 
-    #[allow(clippy::type_complexity)]
-    #[allow(clippy::disallowed_methods)]
-    let input: Option<(Box<dyn Read>, String)> = if let Some(path) = args.next() {
+    #[expect(clippy::disallowed_methods)]
+    let input: Option<(Input, String)> = if let Some(path) = args.next() {
         Some((
-            Box::new(File::open(&path)?),
+            Input::File(File::open(&path)?),
             XPathBuf::from(path).to_string(),
         ))
     } else if isatty(std::io::stdin()).unwrap_or(false) {
         None
     } else {
-        Some((Box::new(std::io::stdin()), "standard input".to_string()))
+        Some((Input::Stdin(std::io::stdin()), "standard input".to_string()))
     };
 
     if let Some((mut input_file, input_name)) = input {
         debug!("ctx": "sh",
             "msg": format!("copying from {input_name} to memory-file {}...",
                 file.as_raw_fd()));
-        let copylen = std::io::copy(&mut input_file, &mut file)?;
+        let copylen = syd::io::copy(&mut input_file, &mut file)?;
         debug!("ctx": "sh",
             "msg": format!("copied {} from {input_name} to memory-file {}",
                 human_size(copylen.try_into()?),
                 file.as_raw_fd()));
 
-        seal_memfd(&file)?;
+        seal_memfd_all(&file)?;
         debug!("ctx": "sh",
             "msg": format!("sealed memory-file {} against grows, shrinks and writes",
                 file.as_raw_fd()));
@@ -279,14 +325,16 @@ syd::main! {
         }
 
         // SAFETY: create a private, write-sealed copy of the memory-file.
-        let mut fdup = create_memfd(b"syd-sh\0", MFD_ALLOW_SEALING).map(File::from)?;
+        let mut fdup = safe_memfd_create(
+            c"syd-sh",
+            MFdFlags::MFD_ALLOW_SEALING | MFdFlags::MFD_CLOEXEC).map(File::from)?;
         debug!("ctx": "sh",
             "msg": format!("created memory-file {} with sealing allowed",
                 fdup.as_raw_fd()));
 
         // rewrite history!
         file.seek(SeekFrom::Start(0))?;
-        let copylen = std::io::copy(&mut file, &mut fdup)?;
+        let copylen = syd::io::copy(&mut file, &mut fdup)?;
         debug!("ctx": "sh",
             "msg": format!("copied {} from memory-file {} to {}",
                 human_size(copylen.try_into()?),
@@ -299,7 +347,7 @@ syd::main! {
                 human_size(input.len()),
                 fdup.as_raw_fd()));
 
-        seal_memfd(&fdup)?;
+        seal_memfd_all(&fdup)?;
         debug!("ctx": "sh",
             "msg": format!("sealed memory-file {} against grows, shrinks and writes",
                 fdup.as_raw_fd()));
