@@ -22,7 +22,7 @@ use std::{
     ptr::NonNull,
 };
 
-use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
+use libseccomp::{scmp_cmp, RawSyscall, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     fcntl::{fcntl, open, splice, FcntlArg, OFlag, SpliceFFlags},
@@ -31,38 +31,59 @@ use nix::{
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
         mman::{mmap_anonymous, MapFlags, ProtFlags},
-        prctl::set_dumpable,
-        resource::{getrlimit, setrlimit, Resource},
+        resource::{getrlimit, Resource},
         socket::{
-            accept4, listen, shutdown, AddressFamily, Backlog, Shutdown, SockFlag, SockType,
-            SockaddrLike, SockaddrStorage, UnixAddr,
+            listen, AddressFamily, Backlog, Shutdown, SockFlag, SockType, SockaddrLike,
+            SockaddrStorage, UnixAddr,
         },
         stat::Mode,
     },
-    unistd::{chdir, chroot, pipe2, sysconf, write, SysconfVar},
+    unistd::{chdir, chroot, sysconf, write, SysconfVar},
 };
 use sendfd::RecvWithFd;
 use syd::{
-    compat::epoll_ctl_safe,
-    config::PIPE_BUF,
-    confine::{confine_mdwe, seccomp_native_has_socketcall},
+    compat::{epoll_ctl_safe, set_dumpable},
+    config::{PIPE_BUF, VDSO_SYSCALLS},
+    confine::{confine_mdwe, confine_rlimit, confine_rlimit_zero, confine_scmp_madvise},
     cookie::{
-        safe_connect, safe_socket, CONNECT_COOKIE_ARG3, CONNECT_COOKIE_ARG4, CONNECT_COOKIE_ARG5,
-        SOCKET_COOKIE_ARG3, SOCKET_COOKIE_ARG4, SOCKET_COOKIE_ARG5,
+        safe_connect, safe_pipe2, safe_shutdown, safe_socket, safer_accept4, ACCEPT4_COOKIE_ARG4,
+        ACCEPT4_COOKIE_ARG5, CONNECT_COOKIE_ARG3, CONNECT_COOKIE_ARG4, CONNECT_COOKIE_ARG5,
+        PIPE2_COOKIE_ARG2, PIPE2_COOKIE_ARG3, PIPE2_COOKIE_ARG4, PIPE2_COOKIE_ARG5,
+        SHUTDOWN_COOKIE_ARG2, SHUTDOWN_COOKIE_ARG3, SHUTDOWN_COOKIE_ARG4, SHUTDOWN_COOKIE_ARG5,
+        SOCKET_COOKIE_ARG3, SOCKET_COOKIE_ARG4, SOCKET_COOKIE_ARG5, SYS_ACCEPT4, SYS_CONNECT,
+        SYS_SHUTDOWN, SYS_SOCKET,
     },
     err::{SydError, SydResult},
-    fs::{closeexcept, duprand, fillrandom, randint, retry_on_eintr},
+    fs::closeexcept,
     hash::SydHashMap,
     ignore_signals,
     landlock_policy::LandlockPolicy,
     parsers::sandbox::PortSet,
     path::XPathBuf,
+    retry::retry_on_eintr,
+    rng::{duprand, fillrandom, randint},
     sealbox::{
         check_madvise_guard_support, madvise_guard_install, mprotect_none, mprotect_readonly, mseal,
     },
     IgnoreSignalOpts,
 };
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 // Pong reply to Syd for debugging.
 const PONG: &[u8] =
     b"{\"ctx\":\"recv_proxy_fd\",\"tor\":\"pong\",\"msg\":\"Action brings good fortune.\"}\n\0";
@@ -71,7 +92,7 @@ syd::main! {
     // Parse options.
     let mut opts = parse_options()?;
 
-    // Ignore all signals except SIG{KILL,STOP,PIPE,CHLD,Core}.
+    // Ignore all signals except SIG{CHLD,KILL,STOP}.
     // This is used to ensure we can deny {rt_,}sigreturn(2) to mitigate SROP.
     ignore_signals(IgnoreSignalOpts::empty())?;
 
@@ -83,7 +104,7 @@ syd::main! {
     // Attempt to set file-max to hard limit overriding the soft limit.
     if let Ok((soft_limit, hard_limit)) = getrlimit(Resource::RLIMIT_NOFILE) {
         if soft_limit < hard_limit {
-            let _ = setrlimit(Resource::RLIMIT_NOFILE, hard_limit, hard_limit);
+            let _ = confine_rlimit(Resource::RLIMIT_NOFILE, Some(hard_limit));
         }
     }
 
@@ -138,8 +159,8 @@ syd::main! {
     // We allocate three pages and install guard-pages before
     // and after the page where we store the external network
     // address.
-    #[allow(clippy::cast_possible_truncation)]
-    #[allow(clippy::cast_sign_loss)]
+    #[expect(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_sign_loss)]
     let page_siz =
         NonZeroUsize::new(sysconf(SysconfVar::PAGE_SIZE)?.ok_or(Errno::EINVAL)? as usize)
             .ok_or(Errno::EINVAL)?;
@@ -180,8 +201,8 @@ syd::main! {
         fillrandom(addr_vec)?;
 
         // Find a random aligned offset within the usable range.
-        let addr_size = std::mem::size_of::<SockaddrStorage>();
-        let addr_align = std::mem::align_of::<SockaddrStorage>();
+        let addr_size = size_of::<SockaddrStorage>();
+        let addr_align = align_of::<SockaddrStorage>();
         let max_offset = page_siz.get().checked_sub(addr_size).ok_or(Errno::EINVAL)?;
         let rand_off = randint(0..=max_offset as u64)? as usize;
         let aligned_off = rand_off & !(addr_align - 1);
@@ -358,20 +379,18 @@ impl<'a> Proxy<'a> {
             // can {{dr}e,}allocate memory.
             // mmap{,2} and mprotect are further confined to disable PROT_EXEC.
             "brk",
-            "madvise",
+            //"madvise", advice are confined.
             "mremap",
             "munmap",
-            // can read random bytes (soon to be in VDSO).
-            "getrandom",
-            // can handle signals.
+            // can handle signals but cannot install new signal handlers.
             // can not return from signal handlers (mitigate SROP).
-            "sigaction",
+            //"sigaction",
             "sigaltstack",
             "sigpending",
             "sigprocmask",
             "sigsuspend",
             //"sigreturn",
-            "rt_sigaction",
+            //"rt_sigaction",
             "rt_sigpending",
             "rt_sigprocmask",
             "rt_sigqueueinfo",
@@ -379,18 +398,18 @@ impl<'a> Proxy<'a> {
             "rt_sigtimedwait",
             "rt_sigtimedwait_time64",
             // can set file flags.
-            "fcntl",
-            "fcntl64",
+            // see below for permitted fcntl(2) operations.
+            //"fcntl",
+            //"fcntl64",
             // can close files.
             "close",
             // can do I/O with pipes.
-            "pipe2",
+            //"pipe2", confined by syscall argument cookies.
             "splice",
+            //"prctl", see below for permitted prctl(2) operations.
             // can forward network.
-            // socket and connect are further confined as necessary.
-            "accept4",
+            // socket, connect, accept4, and shutdown are further confined as necessary.
             "setsockopt",
-            "shutdown",
             // can use EPoll API,
             // can not create new EPoll FDs.
             "epoll_ctl",
@@ -398,31 +417,92 @@ impl<'a> Proxy<'a> {
             "epoll_pwait",
             "epoll_pwait2",
         ];
+        let fcntl_ops: &[u64] = &[
+            libc::F_GETFD as u64,
+            libc::F_SETFD as u64,
+            libc::F_GETFL as u64,
+            libc::F_SETFL as u64,
+        ];
+        let prctl_ops: &[u64] = &[libc::PR_SET_VMA as u64];
 
         // Default allowlist.
-        for name in allow_call {
+        for name in allow_call.iter().chain(VDSO_SYSCALLS) {
             if let Ok(syscall) = ScmpSyscall::from_name(name) {
                 filter.add_rule(ScmpAction::Allow, syscall)?;
             }
         }
 
-        // Socket filtering only works if there's no multiplexing socketcall.
-        #[allow(clippy::useless_conversion)]
-        if seccomp_native_has_socketcall() {
-            for sysname in ["socket", "connect"] {
-                #[allow(clippy::disallowed_methods)]
-                filter.add_rule(ScmpAction::Allow, ScmpSyscall::from_name(sysname).unwrap())?;
+        // Allow safe madvise(2) advice.
+        confine_scmp_madvise(&mut filter)?;
+
+        // Allow safe fcntl(2) utility calls.
+        for sysname in ["fcntl", "fcntl64"] {
+            let syscall = match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => syscall,
+                Err(_) => continue,
+            };
+
+            for op in fcntl_ops {
+                filter.add_rule_conditional(
+                    ScmpAction::Allow,
+                    syscall,
+                    &[scmp_cmp!($arg1 == *op)],
+                )?;
             }
-        } else {
+        }
+
+        // Allow safe prctl(2) operations.
+        let sysname = "prctl";
+        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
+            for op in prctl_ops {
+                filter.add_rule_conditional(
+                    ScmpAction::Allow,
+                    syscall,
+                    &[scmp_cmp!($arg0 == *op)],
+                )?;
+            }
+        }
+
+        // Deny installing new signal handlers for {rt_,}sigaction(2).
+        for sysname in ["sigaction", "rt_sigaction"] {
+            let syscall = match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => syscall,
+                Err(_) => continue,
+            };
+
+            // Installs a signal handler if first argument is non-NULL.
+            // We deny this case, but allow returning the current handler.
+            filter.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == 0)])?;
+        }
+
+        // Restrict pipe2(2) flags.
+        // Restrict pipe2(2) using syscall argument cookies.
+        #[expect(clippy::disallowed_methods)]
+        #[expect(clippy::useless_conversion)]
+        filter.add_rule_conditional(
+            ScmpAction::Allow,
+            ScmpSyscall::from_name("pipe2").unwrap(),
+            &[
+                scmp_cmp!($arg1 == (OFlag::O_NONBLOCK | OFlag::O_CLOEXEC).bits() as u64),
+                scmp_cmp!($arg2 == (*PIPE2_COOKIE_ARG2).into()),
+                scmp_cmp!($arg3 == (*PIPE2_COOKIE_ARG3).into()),
+                scmp_cmp!($arg4 == (*PIPE2_COOKIE_ARG4).into()),
+                scmp_cmp!($arg5 == (*PIPE2_COOKIE_ARG5).into()),
+            ],
+        )?;
+
+        // Socket filtering only works if there's no multiplexing socketcall.
+        #[expect(clippy::disallowed_methods)]
+        #[expect(clippy::useless_conversion)]
+        if let Some(syscall) = SYS_SOCKET.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
             // Restrict socket(2) to the given domain, type and protocol.
             // Restrict socket(2) using syscall argument cookies.
             let sock_domain = domain as u64;
             let sock_type = (libc::SOCK_STREAM | libc::SOCK_NONBLOCK | libc::SOCK_CLOEXEC) as u64;
             let sock_protocol = 0u64;
-            #[allow(clippy::disallowed_methods)]
             filter.add_rule_conditional(
                 ScmpAction::Allow,
-                ScmpSyscall::from_name("socket").unwrap(),
+                syscall,
                 &[
                     scmp_cmp!($arg0 == sock_domain),
                     scmp_cmp!($arg1 == sock_type),
@@ -432,13 +512,18 @@ impl<'a> Proxy<'a> {
                     scmp_cmp!($arg5 == (*SOCKET_COOKIE_ARG5).into()),
                 ],
             )?;
+        } else {
+            filter.add_rule(ScmpAction::Allow, ScmpSyscall::from_name("socket").unwrap())?;
+        }
 
+        #[expect(clippy::disallowed_methods)]
+        #[expect(clippy::useless_conversion)]
+        if let Some(syscall) = SYS_CONNECT.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
             // Restrict connect(2) to a single safe pointer.
             // Restrict connect(2) using syscall argument cookies.
-            #[allow(clippy::disallowed_methods)]
             filter.add_rule_conditional(
                 ScmpAction::Allow,
-                ScmpSyscall::from_name("connect").unwrap(),
+                syscall,
                 &[
                     scmp_cmp!($arg1 == addr_ptr),
                     scmp_cmp!($arg2 == addr_len),
@@ -447,12 +532,65 @@ impl<'a> Proxy<'a> {
                     scmp_cmp!($arg5 == (*CONNECT_COOKIE_ARG5).into()),
                 ],
             )?;
+        } else {
+            filter.add_rule(
+                ScmpAction::Allow,
+                ScmpSyscall::from_name("connect").unwrap(),
+            )?;
+        }
+
+        #[expect(clippy::disallowed_methods)]
+        #[expect(clippy::useless_conversion)]
+        if let Some(syscall) = SYS_ACCEPT4.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
+            // Restrict accept4(2) to a single randomized fd.
+            // Restrict accept4(2) using syscall argument cookies.
+            filter.add_rule_conditional(
+                ScmpAction::Allow,
+                syscall,
+                &[
+                    scmp_cmp!($arg0 == self.proxy.as_raw_fd() as u64),
+                    scmp_cmp!($arg1 == 0u64), // struct sockaddr *_Nullable restrict addr
+                    scmp_cmp!($arg2 == 0u64), // socklen_t *_Nullable restrict addrlen
+                    scmp_cmp!($arg3 == (SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC).bits() as u64),
+                    scmp_cmp!($arg4 == (*ACCEPT4_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*ACCEPT4_COOKIE_ARG5).into()),
+                ],
+            )?;
+        } else {
+            filter.add_rule(
+                ScmpAction::Allow,
+                ScmpSyscall::from_name("accept4").unwrap(),
+            )?;
+        }
+
+        #[expect(clippy::disallowed_methods)]
+        #[expect(clippy::useless_conversion)]
+        if let Some(syscall) = SYS_SHUTDOWN.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall))
+        {
+            // Restrict shutdown(2) to Shutdown::Write.
+            // Restrict shutdown(2) using syscall argument cookies.
+            filter.add_rule_conditional(
+                ScmpAction::Allow,
+                syscall,
+                &[
+                    scmp_cmp!($arg1 == libc::SHUT_WR as u64),
+                    scmp_cmp!($arg2 == (*SHUTDOWN_COOKIE_ARG2).into()),
+                    scmp_cmp!($arg3 == (*SHUTDOWN_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*SHUTDOWN_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*SHUTDOWN_COOKIE_ARG5).into()),
+                ],
+            )?;
+        } else {
+            filter.add_rule(
+                ScmpAction::Allow,
+                ScmpSyscall::from_name("shutdown").unwrap(),
+            )?;
         }
 
         // Prevent executable memory.
         const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
         for name in ["mmap", "mmap2", "mprotect"] {
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let syscall = ScmpSyscall::from_name(name).unwrap();
             filter.add_rule_conditional(
                 ScmpAction::Allow,
@@ -527,10 +665,8 @@ impl<'a> Proxy<'a> {
             // Set the process dumpable attribute to not-dumpable.
             let _ = set_dumpable(false);
 
-            // Deny reading the timestamp counter (x86 only).
-            // SAFETY: Our nix version does not have a wrapper for SET_TSC yet.
-            #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-            let _ = unsafe { libc::prctl(libc::PR_SET_TSC, libc::PR_TSC_SIGSEGV) };
+            // Set nprocs and filesize rlimits to zero.
+            confine_rlimit_zero(&[Resource::RLIMIT_FSIZE, Resource::RLIMIT_NPROC])?;
         }
 
         if self.print {
@@ -548,14 +684,14 @@ impl<'a> Proxy<'a> {
     }
 
     fn handle_connection(&mut self, fd: RawFd, mut event_flags: EpollFlags) -> SydResult<()> {
-        // This query MUST never fail.
-        #[allow(clippy::disallowed_methods)]
-        let client = self.get_client_mut(fd).unwrap();
-
-        let fd_cli = client.fd_cli.as_raw_fd();
-        let fd_ext = client.fd_ext.as_raw_fd();
-        let is_ext = fd == fd_ext;
+        // This query may fail for stale epoll events.
+        let client = if let Some(client) = self.get_client_mut(fd) {
+            client
+        } else {
+            return Ok(());
+        };
 
+        let is_ext = fd == client.fd_ext.as_raw_fd();
         let is_inp = event_flags
             .intersects(EpollFlags::EPOLLIN | EpollFlags::EPOLLPRI)
             .then(|| event_flags.remove(EpollFlags::EPOLLIN | EpollFlags::EPOLLPRI))
@@ -578,19 +714,27 @@ impl<'a> Proxy<'a> {
             .is_some();
         assert!(event_flags.is_empty(), "BUG: {event_flags:?}");
 
+        let mut progressed = false;
         if is_inp || is_rd_hup || is_hup || is_err {
             // Handle readable events.
-            let _ = client.handle_input(is_ext);
+            progressed |= client.handle_input(is_ext).unwrap_or(false);
         }
 
         if is_out || is_hup || is_err {
             // Handle writable events.
-            let _ = client.handle_output(is_ext);
+            progressed |= client.handle_output(is_ext).unwrap_or(false);
         }
 
         if is_rd_hup {
             // Half-close the write side on RDHUP.
-            let _ = shutdown(if is_ext { fd_cli } else { fd_ext }, Shutdown::Write);
+            let _ = safe_shutdown(
+                if is_ext {
+                    &client.fd_cli
+                } else {
+                    &client.fd_ext
+                },
+                Shutdown::Write,
+            );
         }
 
         if is_hup || is_err {
@@ -599,52 +743,74 @@ impl<'a> Proxy<'a> {
 
             // Close on errors and HUP.
             self.close_connection(fd);
+
+            return Ok(());
+        }
+
+        // Re-drive read/write until no immediate progress is possible.
+        // This avoids ET starvation when the pipe (not the socket) was full.
+        while progressed {
+            progressed = false;
+
+            // Flush queues in both directions first, in case writes were blocking.
+            progressed |= client.handle_output(is_ext).unwrap_or(false);
+            progressed |= client.handle_output(!is_ext).unwrap_or(false);
+
+            // After flushing, try to pull more from both peers.
+            progressed |= client.handle_input(is_ext).unwrap_or(false);
+            progressed |= client.handle_input(!is_ext).unwrap_or(false);
         }
 
         Ok(())
     }
 
     fn add_connection(&mut self) -> SydResult<()> {
-        // Quoting accept(2):
-        // Linux accept() (and accept4()) passes already-pending network
-        // errors on the new socket as an error code from accept(). This
-        // behavior differs from other BSD socket implementations. For
-        // reliable operation the application should detect the network
-        // errors defined for the protocol after accept() and treat them
-        // like EAGAIN by retrying. In the case of TCP/IP, these are
-        // ENETDOWN, EPROTO, ENOPROTOOPT, EHOSTDOWN, ENONET, EHOSTUNREACH,
-        // EOPNOTSUPP, and ENETUNREACH.
-        #[allow(unreachable_patterns)]
-        let fd_cli = match accept4(
-            self.proxy.as_fd().as_raw_fd(),
-            SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC,
-        ) {
-            Ok(fd) => unsafe { OwnedFd::from_raw_fd(fd) },
-            Err(
-                Errno::EAGAIN
-                | Errno::EHOSTDOWN
-                | Errno::EHOSTUNREACH
-                | Errno::EINPROGRESS
-                | Errno::ENETDOWN
-                | Errno::ENETUNREACH
-                | Errno::ENONET
-                | Errno::ENOPROTOOPT
-                | Errno::EOPNOTSUPP
-                | Errno::EPROTO
-                | Errno::EWOULDBLOCK,
-            ) => return Ok(()), // No more connections to accept
-            Err(errno) => return Err(errno.into()),
-        };
+        loop {
+            // Quoting accept(2):
+            // Linux accept() (and accept4()) passes already-pending network
+            // errors on the new socket as an error code from accept(). This
+            // behavior differs from other BSD socket implementations. For
+            // reliable operation the application should detect the network
+            // errors defined for the protocol after accept() and treat them
+            // like EAGAIN by retrying. In the case of TCP/IP, these are
+            // ENETDOWN, EPROTO, ENOPROTOOPT, EHOSTDOWN, ENONET, EHOSTUNREACH,
+            // EOPNOTSUPP, and ENETUNREACH.
+            #[expect(unreachable_patterns)]
+            match retry_on_eintr(|| {
+                safer_accept4(
+                    &self.proxy,
+                    SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC,
+                )
+            }) {
+                Ok(fd_cli) => {
+                    let client = Client::new(fd_cli, self.extaddr)?;
+                    self.add_client(client)?;
+                }
+                Err(
+                    Errno::EAGAIN
+                    | Errno::EHOSTDOWN
+                    | Errno::EHOSTUNREACH
+                    | Errno::EINPROGRESS
+                    | Errno::ENETDOWN
+                    | Errno::ENETUNREACH
+                    | Errno::ENONET
+                    | Errno::ENOPROTOOPT
+                    | Errno::EOPNOTSUPP
+                    | Errno::EPROTO
+                    | Errno::EWOULDBLOCK,
+                ) => break, // No more connections to accept
+                Err(errno) => return Err(errno.into()),
+            }
+        }
 
-        let client = Client::new(fd_cli, self.extaddr)?;
-        self.add_client(client)
+        Ok(())
     }
 
     fn close_connection(&mut self, fd: RawFd) {
-        if let Some(client) = self.remove_client(fd) {
-            let _ = epoll_ctl_safe(&self.epoll.0, client.fd_cli.as_raw_fd(), None);
-            let _ = epoll_ctl_safe(&self.epoll.0, client.fd_ext.as_raw_fd(), None);
-        } // All client OwnedFds are closed end of this block.
+        // Dropping Client closes all associated OwnedFds.
+        // No need to issue EPOLL_CTL_DEL here, as there's
+        // only a single copy of each fd.
+        self.remove_client(fd);
     }
 
     fn add_client(&mut self, client: Client) -> SydResult<()> {
@@ -731,8 +897,8 @@ struct Client {
 
 impl Client {
     fn new(fd_cli: OwnedFd, extaddr: &SockaddrStorage) -> SydResult<Self> {
-        let (pipe_in_from_cli, pipe_out_to_cli) = pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
-        let (pipe_in_from_ext, pipe_out_to_ext) = pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
+        let (pipe_in_from_cli, pipe_out_to_cli) = safe_pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
+        let (pipe_in_from_ext, pipe_out_to_ext) = safe_pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
 
         // Set socket options on the client fd.
         set_socket_options(&fd_cli, Some(0));
@@ -769,7 +935,7 @@ impl Client {
         })
     }
 
-    fn handle_input(&mut self, ext: bool) -> Result<(), Errno> {
+    fn handle_input(&mut self, ext: bool) -> Result<bool, Errno> {
         let (src_fd, dst_fd, pipe_in_fd, pipe_out_fd, eof) = if ext {
             (
                 self.fd_ext.as_fd(),
@@ -789,12 +955,17 @@ impl Client {
         };
 
         // Edge-triggered epoll: read until EAGAIN.
+        let mut progressed = false;
         while !*eof {
             match Self::splice(&src_fd, &pipe_out_fd) {
-                Ok(0) => *eof = true,
+                Ok(0) => {
+                    *eof = true;
+                    break;
+                }
                 Ok(_) => {
                     // Data transferred - try to push through pipe immediately
                     // This prevents pipe buffer buildup.
+                    progressed = true;
                     loop {
                         match Self::splice(&pipe_in_fd, &dst_fd) {
                             Ok(0) | Err(Errno::EAGAIN) => break,
@@ -808,10 +979,10 @@ impl Client {
             };
         }
 
-        Ok(())
+        Ok(progressed)
     }
 
-    fn handle_output(&mut self, ext: bool) -> Result<(), Errno> {
+    fn handle_output(&mut self, ext: bool) -> Result<bool, Errno> {
         let (dst_fd, pipe_in_fd) = if ext {
             (self.fd_cli.as_fd(), self.pipe_ext.0.as_fd())
         } else {
@@ -819,15 +990,16 @@ impl Client {
         };
 
         // Edge-triggered epoll: write until EAGAIN.
+        let mut progressed = false;
         loop {
             match Self::splice(&pipe_in_fd, &dst_fd) {
                 Ok(0) | Err(Errno::EAGAIN) => break,
-                Ok(_) => {}
+                Ok(_) => progressed = true,
                 Err(errno) => return Err(errno),
             };
         }
 
-        Ok(())
+        Ok(progressed)
     }
 
     fn splice<F: AsFd>(src_fd: &F, dst_fd: &F) -> Result<usize, Errno> {
@@ -964,7 +1136,7 @@ fn parse_options() -> SydResult<TorOpts> {
     // 2. The PID fd and the socket fd passed by the Syd process.
     // Nothing can be done on closeexcept errors.
     // We do it early here so FD randomization doesn't effect performance.
-    #[allow(clippy::cast_sign_loss)]
+    #[expect(clippy::cast_sign_loss)]
     {
         let fd1 = fpid as libc::c_uint;
         let fd2 = sock as libc::c_uint;
@@ -990,7 +1162,7 @@ fn parse_options() -> SydResult<TorOpts> {
             // SAFETY: We are going to access this file descriptor using
             // `/proc/self/fd` indirection after isolating the mount-ns.
             // Note, confine chroots and/or chdirs to /proc/self/fd.
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let fd = open(&unix, OFlag::O_PATH | OFlag::O_CLOEXEC, Mode::empty())?;
 
             // SAFETY: Randomize fd to make it harder to locate for an attacker.
@@ -1063,7 +1235,7 @@ fn set_socket_options<F: AsFd>(fd: &F, backlog: Option<libc::c_int>) {
                 level,
                 optname,
                 &optval as *const _ as *const libc::c_void,
-                std::mem::size_of::<T>() as libc::socklen_t,
+                size_of::<T>() as libc::socklen_t,
             )
         };
         if ret == -1 {
