@@ -18,45 +18,50 @@ use std::{
         fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd},
         unix::ffi::OsStrExt,
     },
-    time::{Duration, Instant},
+    time::Instant,
 };
 
 use bitflags::bitflags;
 use data_encoding::HEXLOWER;
+use dur::Duration;
 use libseccomp::{ScmpAction, ScmpFilterContext, ScmpSyscall};
 use memchr::memchr3;
 use nix::{
     errno::Errno,
-    fcntl::OFlag,
+    fcntl::{open, OFlag},
     libc::{_exit, c_char, size_t, CLONE_FILES, ENOSYS, SIGCHLD, SIGKILL, SIGSYS},
-    mount::{mount, MsFlags},
+    mount::MsFlags,
     sched::{unshare, CloneFlags},
     sys::{
-        prctl::set_pdeathsig,
+        resource::Resource,
         signal::{sigprocmask, SigSet, SigmaskHow, Signal},
+        stat::Mode,
         wait::{Id, WaitPidFlag},
     },
-    unistd::{chdir, close, Gid, Pid, Uid},
+    unistd::{chdir, Gid, Pid, Uid},
 };
 
 use crate::{
-    compat::{pipe2_raw, waitid, WaitStatus},
+    compat::{pipe2_raw, set_pdeathsig, waitid, MFdFlags, WaitStatus},
     config::{MINI_STACK_SIZE, *},
-    confine::{confine_mdwe, MS_NOSYMFOLLOW},
+    confine::{confine_mdwe, confine_rlimit, CLONE_NEWTIME},
+    cookie::safe_memfd_create,
     debug,
     err::err2no,
-    fs::{
-        create_memfd, safe_clone, safe_copy_if_exists, seal_memfd, set_cloexec, set_nonblock,
-        MFD_ALLOW_SEALING,
-    },
+    fs::{close, pidfd_send_signal, safe_clone, seal_memfd_all, set_cloexec, set_nonblock},
     get_user_home, get_user_name,
     hash::SydHashSet,
     landlock::RulesetStatus,
     landlock_policy::LandlockPolicy,
     log::contains_ascii_unprintable,
+    lookup::safe_copy_if_exists,
+    mount::{
+        api::MountAttrFlags,
+        util::{mount_bind, mount_fs, set_root_mount_propagation},
+    },
     path::PATH_MAX,
     unshare::{GidMap, UidMap},
-    xpath, XPath, XPathBuf,
+    xpath, XPathBuf,
 };
 
 bitflags! {
@@ -103,7 +108,7 @@ pub enum WordExpError {
     /// Process was aborted unexpectedly with signal.
     ProcessError(i32),
     /// Timeout error
-    TimeoutError(u64),
+    TimeoutError(u128),
 }
 
 /// Out of memory.
@@ -145,7 +150,7 @@ impl From<i32> for WordExpError {
             // Used by pipe writer in the confined process.
             return Self::SystemError(Errno::from_raw(code));
         }
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         match code {
             WRDE_BADCHAR => Self::BadCharacter,
             WRDE_BADVAL => Self::BadValue,
@@ -161,7 +166,7 @@ impl From<i32> for WordExpError {
 
 impl From<WordExpError> for i32 {
     fn from(val: WordExpError) -> Self {
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         match val {
             WordExpError::BadCharacter => WRDE_BADCHAR,
             WordExpError::BadValue => WRDE_BADVAL,
@@ -292,7 +297,7 @@ impl WordExp<'_> {
     }
 
     /// Perform environment/tilde expansion and optionally command substitution.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub fn expand_full(input: &str, timeout: Duration) -> Result<Cow<'_, str>, WordExpError> {
         // Quick returns:
         // Empty string or no special characters present.
@@ -307,7 +312,11 @@ impl WordExp<'_> {
         // Create a memory fd to write input into,
         // and pass to the internal /bin/sh invoked
         // by wordexp(3).
-        let mut file = create_memfd(b"syd-exp\0", MFD_ALLOW_SEALING).map(File::from)?;
+        let mut file = safe_memfd_create(
+            c"syd-wordexp",
+            MFdFlags::MFD_ALLOW_SEALING | MFdFlags::MFD_CLOEXEC,
+        )
+        .map(File::from)?;
         debug!("ctx": "expand",
             "msg": format!("created memory-file {} with close-on-exec flag set",
                 file.as_raw_fd()));
@@ -317,23 +326,22 @@ impl WordExp<'_> {
         file.write_all(b"\n")?;
 
         // Handle system-wide configuration.
-        safe_copy_if_exists(&mut file, XPath::from_bytes(b"/etc/syd/init.sh"))?;
+        safe_copy_if_exists(&mut file, "/etc/syd/init.sh")?;
         file.write_all(b"\n")?;
 
         // Handle user-specific configuration.
-        let home = env::var_os("HOME")
-            .map(XPathBuf::from)
-            .unwrap_or(XPathBuf::from("/proc/self/fdinfo"));
-        safe_copy_if_exists(&mut file, &home.join(b".config/syd/init.sh"))?;
-        file.write_all(b"\n")?;
+        if let Some(home) = env::var_os("HOME").map(XPathBuf::from) {
+            safe_copy_if_exists(&mut file, &home.join(b".config/syd/init.sh"))?;
+            file.write_all(b"\n")?;
+        }
 
         // Write input.
-        file.write_all(b"echo -n ")?;
+        file.write_all(b"eval set -- x ")?;
         file.write_all(input.as_bytes())?;
-        file.write_all(b"\n")?;
+        file.write_all(b"\nshift\nprintf '%s ' \"$@\"\n")?;
 
         // Seal memfd for future writes.
-        seal_memfd(&file)?;
+        seal_memfd_all(&file)?;
         debug!("ctx": "expand",
             "msg": format!("sealed memory-file {} against grows, shrinks and writes",
                 file.as_raw_fd()));
@@ -344,7 +352,7 @@ impl WordExp<'_> {
             "msg": format!("set close-on-exec flag to off for memory-file {}",
                 file.as_raw_fd()));
 
-        let shell = format!("`. /proc/self/fd/{}`", file.as_raw_fd());
+        let shell = format!("`. /proc/thread-self/fd/{}`", file.as_raw_fd());
         debug!("ctx": "expand",
             "msg": format!("passing memory file {} to wordexp(3) with {} seconds timeout...",
                 file.as_raw_fd(), timeout.as_secs()));
@@ -405,7 +413,7 @@ impl WordExp<'_> {
                     if let Err(ref error) = pipe.write_all(word.as_bytes()) {
                         let err = err2no(error) as i32;
                         // SAFETY: In libc we trust.
-                        #[allow(clippy::arithmetic_side_effects)]
+                        #[expect(clippy::arithmetic_side_effects)]
                         unsafe {
                             _exit(128 + err)
                         };
@@ -413,7 +421,7 @@ impl WordExp<'_> {
                     if let Err(ref error) = pipe.write_all(b" ") {
                         let err = err2no(error) as i32;
                         // SAFETY: In libc we trust.
-                        #[allow(clippy::arithmetic_side_effects)]
+                        #[expect(clippy::arithmetic_side_effects)]
                         unsafe {
                             _exit(128 + err)
                         };
@@ -439,21 +447,12 @@ impl WordExp<'_> {
         let mut ret = Vec::new();
 
         loop {
-            if !sig && (err as i32 != 0 || epoch.elapsed() >= timeout) {
-                // a. out of memory condition
-                // b. timeout exceeded
-                // send SIGKILL once, and fall-through to wait.
+            if !sig && (err as i32 != 0 || epoch.elapsed() >= timeout.into()) {
+                // a. Out of memory condition
+                // b. Timeout exceeded
+                // Send SIGKILL once, and fall-through to wait.
                 sig = true;
-                // SAFETY: There's no libc wrapper for pidfd_send_signal yet.
-                let _ = unsafe {
-                    libc::syscall(
-                        libc::SYS_pidfd_send_signal,
-                        pid_fd.as_fd().as_raw_fd(),
-                        SIGKILL,
-                        0,
-                        0,
-                    )
-                };
+                let _ = pidfd_send_signal(&pid_fd, SIGKILL);
             } else if !eof {
                 // read one batch from pipe.
                 match pipe.read(&mut buf) {
@@ -474,7 +473,7 @@ impl WordExp<'_> {
                         continue;
                     }
                     Err(ref e) if matches!(err2no(e), Errno::EAGAIN | Errno::EINTR) => {
-                        std::thread::sleep(Duration::from_millis(100));
+                        std::thread::sleep(Duration::from_millis(100).into());
                         continue;
                     }
                     Err(ref e) => {
@@ -558,9 +557,13 @@ impl WordExp<'_> {
     ///
     /// Panics on all errors except Landlock and namespaces which are
     /// optional as they may not be available.
-    #[allow(clippy::cognitive_complexity)]
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::cognitive_complexity)]
+    #[expect(clippy::disallowed_methods)]
     pub fn confine() {
+        if env::var_os(ENV_SKIP_SCMP).is_some() {
+            return;
+        }
+
         // SAFETY: Determine user HOME directory.
         // This will be confined by Landlock.
         let uid = Uid::current();
@@ -585,12 +588,13 @@ impl WordExp<'_> {
                 | CloneFlags::CLONE_NEWNET
                 | CloneFlags::CLONE_NEWNS
                 | CloneFlags::CLONE_NEWPID
-                | CloneFlags::CLONE_NEWUTS,
+                | CloneFlags::CLONE_NEWUTS
+                | CLONE_NEWTIME,
         )
         .is_ok()
         {
             debug!("ctx": "expand",
-                "msg": "created/entered into new user, mount, pid, network, cgroup, ipc, and uts namespaces");
+                "msg": "created and entered into new user, mount, pid, network, cgroup, ipc, uts, and time namespaces");
             // SAFETY: map current user in new user namespace.
             let uid_buf = {
                 let uid_maps = vec![
@@ -648,30 +652,27 @@ impl WordExp<'_> {
             // SAFETY: remount rootfs as readonly,nosuid,nodev,nosymfollow
             // Careful here, unshare(2) may be available but mount(2) may not be,
             // so we must handle mount errors gracefully.
-            let mut flags: MsFlags = MsFlags::MS_BIND
-                | MsFlags::MS_REC
-                | MsFlags::MS_RDONLY
-                | MsFlags::MS_NOSUID
-                | MsFlags::MS_NODEV
-                | MS_NOSYMFOLLOW;
-            if mount(
-                Some("none"),
-                "/",
-                None::<&XPath>,
-                MsFlags::MS_PRIVATE | MsFlags::MS_REC,
-                None::<&XPath>,
-            )
-            .is_ok()
-            {
+            let mut flags = MountAttrFlags::MOUNT_ATTR_RDONLY
+                | MountAttrFlags::MOUNT_ATTR_NOSUID
+                | MountAttrFlags::MOUNT_ATTR_NODEV
+                | MountAttrFlags::MOUNT_ATTR_NOSYMFOLLOW;
+
+            if set_root_mount_propagation(MsFlags::MS_PRIVATE).is_ok() {
                 debug!("ctx": "expand",
-                    "msg": "set mount propagation to private in the new mount namespace");
-                mount(Some("/"), "/", Some("/"), flags, None::<&XPath>).expect("remount rootfs");
+                    "msg": "set mount propagation to private in new mount namespace");
+                open(
+                    "/",
+                    OFlag::O_CLOEXEC | OFlag::O_PATH | OFlag::O_DIRECTORY,
+                    Mode::empty(),
+                )
+                .and_then(|root| mount_bind(&root, &root, flags))
+                .expect("remount rootfs");
                 debug!("ctx": "expand",
-                    "msg": "remounted root with readonly, nosuid, nodev, and nosymfollow options in the new mount namespace");
+                    "msg": "remounted root with readonly, nosuid, nodev, and nosymfollow options in new mount namespace");
                 // SAFETY: mount private procfs
                 // pid=1 is required to exist before this.
-                flags.remove(MsFlags::MS_BIND | MsFlags::MS_REC | MS_NOSYMFOLLOW);
-                flags.insert(MsFlags::MS_NOEXEC);
+                flags.remove(MountAttrFlags::MOUNT_ATTR_NOSYMFOLLOW);
+                flags.insert(MountAttrFlags::MOUNT_ATTR_NOEXEC);
                 Self::mount_proc(flags);
             }
         }
@@ -835,12 +836,15 @@ impl WordExp<'_> {
             }
         }
 
-        // SAFETY: confine with seccomp.
-        // panics on errors.
+        // Set file size rlimits to zero, panic on errors.
+        #[expect(clippy::disallowed_methods)]
+        confine_rlimit(Resource::RLIMIT_FSIZE, None).expect("setrlimit(RLIMIT_FSIZE,0)");
+
+        // SAFETY: confine with seccomp, panics on errors.
         Self::confine_seccomp();
     }
 
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     fn confine_seccomp() {
         let mut filter = ScmpFilterContext::new(ScmpAction::Errno(ENOSYS)).expect("create filter");
 
@@ -856,7 +860,11 @@ impl WordExp<'_> {
         // Use a binary tree sorted by syscall number, if possible.
         let _ = filter.set_ctl_optimize(2);
 
-        for sysname in WORDEXP_SYSCALLS {
+        for sysname in WORDEXP_SYSCALLS
+            .iter()
+            .chain(GET_ID_SYSCALLS)
+            .chain(VDSO_SYSCALLS)
+        {
             if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
                 filter
                     .add_rule(ScmpAction::Allow, syscall)
@@ -869,15 +877,15 @@ impl WordExp<'_> {
             "msg": "loaded seccomp filter");
     }
 
-    #[allow(clippy::cognitive_complexity)]
-    #[allow(clippy::disallowed_methods)]
-    fn mount_proc(flags: MsFlags) {
+    #[expect(clippy::cognitive_complexity)]
+    #[expect(clippy::disallowed_methods)]
+    fn mount_proc(flags: MountAttrFlags) {
         let mut stack = [0u8; MINI_STACK_SIZE];
         safe_clone(
             Box::new(move || -> isize {
                 // pid=1 here.
                 debug!("ctx": "expand",
-                    "msg": "started init process in the new pid namespace");
+                    "msg": "started init process in new pid namespace");
 
                 // SAFETY: set parent-death signal to SIGKILL
                 if set_pdeathsig(Some(Signal::SIGKILL)).is_err() {
@@ -891,16 +899,22 @@ impl WordExp<'_> {
                     .expect("block signals");
 
                 // SAFETY: mount private procfs, continue on errors.
-                match mount(
-                    Some("proc"),
+                match open(
                     "/proc",
-                    Some("proc"),
-                    flags,
-                    Some("hidepid=2"),
-                ) {
+                    OFlag::O_CLOEXEC | OFlag::O_PATH | OFlag::O_DIRECTORY,
+                    Mode::empty(),
+                )
+                .and_then(|proc| {
+                    mount_fs(
+                        OsStr::new("proc"),
+                        proc,
+                        flags,
+                        Some("hidepid=4,subset=pid"),
+                    )
+                }) {
                     Ok(_) => {
                         debug!("ctx": "expand",
-                            "msg": "mounted proc with hidepid=2 in the new mount namespace");
+                            "msg": "mounted proc with hidepid=4,subset=pid in new mount namespace");
                     }
                     Err(errno) => {
                         debug!("ctx": "expand",
@@ -909,7 +923,7 @@ impl WordExp<'_> {
                 };
 
                 // SAFETY: block until the parent-death signal kills us.
-                std::thread::sleep(Duration::MAX);
+                std::thread::sleep(std::time::Duration::MAX);
 
                 unreachable!();
             }),
@@ -938,13 +952,8 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "breakpoint", // arm
     "brk",
     "cacheflush", // arm
-    "cachestat",
     "capget",
     "chdir",
-    "clock_getres",
-    "clock_getres_time64",
-    "clock_gettime",
-    "clock_gettime64",
     "clock_nanosleep",
     "clock_nanosleep_time64",
     "clone",
@@ -992,16 +1001,7 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "get_mempolicy",
     "get_robust_list",
     "get_thread_area",
-    "getcpu",
     "getcwd",
-    "getegid",
-    "getegid32",
-    "geteuid",
-    "geteuid32",
-    "getgid",
-    "getgid32",
-    "getgroups",
-    "getgroups32",
     "getitimer",
     "getpeername",
     "getpgid",
@@ -1010,17 +1010,11 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "getpmsg",
     "getppid",
     "getpriority",
-    "getrandom",
-    "getresgid",
-    "getresuid",
     "getrlimit",
     "getrusage",
     "getsid",
     "getsockopt",
     "gettid",
-    "gettimeofday",
-    "getuid",
-    "getuid32",
     "getxattr",
     "io_cancel",
     "io_destroy",
@@ -1041,7 +1035,7 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "llistxattr",
     "lseek",
     "lstat",
-    "madvise",
+    "madvise", // TODO: confine advice (no-op!).
     "membarrier",
     "mlock",
     "mlock2",
@@ -1162,7 +1156,6 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "sync_file_range",
     "tee",
     "tgkill",
-    "time",
     "timer_create",
     "timer_delete",
     "timer_getoverrun",
