@@ -20,7 +20,7 @@ use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     sched::{unshare, CloneFlags},
-    unistd::{getpid, write, Gid, Pid, Uid},
+    unistd::{getpid, lseek64, write, Gid, Pid, Uid, Whence},
 };
 use serde::{ser::SerializeMap, Serialize, Serializer};
 
@@ -28,13 +28,14 @@ use crate::{
     alert,
     cache::SysInterrupt,
     config::*,
-    confine::{confine_scmp_write, scmp_add_setid_rules, ExportMode},
+    confine::{confine_scmp_madvise, confine_scmp_write, scmp_add_setid_rules, ExportMode},
     cookie::{OPENAT2_COOKIE_ARG4, OPENAT2_COOKIE_ARG5},
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error,
-    fs::{closeexcept, retry_on_eintr, seccomp_notify_id_valid, tgkill},
+    fs::{closeexcept, seccomp_notify_id_valid, tgkill},
     info,
-    proc::proc_status,
+    proc::{proc_status_open, proc_status_read},
+    retry::retry_on_eintr,
     sandbox::Flags,
     sigset::SydSigSet,
     workers::WorkerCache,
@@ -42,6 +43,7 @@ use crate::{
 
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 enum InterruptSource {
+    Manual,
     Signal(SydSigSet),
     ProcessInvalid(Errno),
 }
@@ -53,6 +55,9 @@ impl Serialize for InterruptSource {
     {
         let mut map = serializer.serialize_map(Some(2))?;
         match self {
+            Self::Manual => {
+                map.serialize_entry("name", "manual")?;
+            }
             Self::Signal(set) => {
                 map.serialize_entry("name", "signal")?;
                 map.serialize_entry("set", set)?;
@@ -76,7 +81,7 @@ pub(crate) struct Interrupter {
     transit_gids: Vec<(Gid, Gid)>,
 
     should_exit: Arc<AtomicBool>,
-    cache: Arc<WorkerCache<'static>>,
+    cache: Arc<WorkerCache>,
 }
 
 impl Interrupter {
@@ -86,7 +91,7 @@ impl Interrupter {
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
         should_exit: Arc<AtomicBool>,
-        cache: Arc<WorkerCache<'static>>,
+        cache: Arc<WorkerCache>,
     ) -> Self {
         Self {
             flags,
@@ -98,7 +103,7 @@ impl Interrupter {
         }
     }
 
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn try_spawn(self, notif_pipe: (RawFd, RawFd)) -> Result<SydJoinHandle<()>, Errno> {
         thread::Builder::new()
             .name("syd_int".to_string())
@@ -122,7 +127,7 @@ impl Interrupter {
                 };
                 drop(pipe_rd);
                 let buf = [42u8; 1];
-                #[allow(clippy::disallowed_methods)]
+                #[expect(clippy::disallowed_methods)]
                 match retry_on_eintr(|| write(&pipe_wr, &buf)).unwrap() {
                     0 => return Err(Errno::EIO.into()), // Syd died before reading.
                     1 => {}
@@ -137,7 +142,7 @@ impl Interrupter {
                 // 2. Static FD of procfs(5).
                 // 3. Log FD.
                 // We have to sort the set as the FDs are randomized.
-                #[allow(clippy::cast_sign_loss)]
+                #[expect(clippy::cast_sign_loss)]
                 let mut set = vec![
                     self.seccomp_fd as libc::c_uint,
                     PROC_FD() as libc::c_uint,
@@ -174,7 +179,7 @@ impl Interrupter {
                         Ok(ctx) => ctx,
                         Err(error) => {
                             let errno = error.errno().unwrap_or(Errno::ENOSYS);
-                            alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                            alert!("ctx": "boot", "op": "confine_int_thread",
                                 "msg": format!("failed to confine: {error}"),
                                 "err": errno as i32);
                             std::process::exit(101);
@@ -186,7 +191,7 @@ impl Interrupter {
                     // because this confinement is a critical safety feature.
                     if let Err(error) = ctx.load() {
                         let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
-                        alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                        alert!("ctx": "boot", "op": "confine_int_thread",
                             "msg": format!("failed to confine: {error}"),
                             "err": errno as i32);
                         std::process::exit(101);
@@ -195,7 +200,7 @@ impl Interrupter {
                     let safe_setid = self
                         .flags
                         .intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
-                    info!("ctx": "confine", "op": "confine_interrupt_thread",
+                    info!("ctx": "confine", "op": "confine_int_thread",
                         "msg": format!("interrupt thread confined with{} SROP mitigation",
                             if safe_setid { "out" } else { "" }));
                 } else {
@@ -212,7 +217,7 @@ impl Interrupter {
     fn main(self, tgid: Pid) -> SydResult<()> {
         loop {
             // Wait for one cycle.
-            std::thread::sleep(INT_CYCLE_TIME);
+            std::thread::sleep(INT_CYCLE_TIME.into());
 
             // Unblock invalidated blocking system calls.
             {
@@ -226,7 +231,10 @@ impl Interrupter {
                     .unwrap_or_else(|err| err.into_inner());
 
                 // Handle interrupts as necessary.
-                map.retain(|_, interrupt| self.handle_interrupt(tgid, interrupt));
+                map.retain_mut(|interrupt| self.handle_interrupt(tgid, interrupt));
+
+                // Keep memory usage minimal.
+                map.shrink_to_fit();
             }
 
             // Check if it's the time to exit.
@@ -238,30 +246,70 @@ impl Interrupter {
         Ok(())
     }
 
-    fn handle_interrupt(&self, tgid: Pid, interrupt: &SysInterrupt) -> bool {
-        // Check pending signals for the thread.
+    // Handles syscall signal interrupts.
+    //
+    // Returns false if interrupt is handled, true otherwise.
+    fn handle_interrupt(&self, tgid: Pid, interrupt: &mut SysInterrupt) -> bool {
+        // Check if syd_emu is already done with the request.
+        if interrupt.delete {
+            return false;
+        }
+        // Check if syd_mon requested manual interruption.
+        if interrupt.signal {
+            Self::interrupt(tgid, interrupt, InterruptSource::Manual);
+            return false;
+        }
+
+        // Open proc_pid_status(5) if it's not open already.
         //
         // SAFETY:
         // 1. We want to wake the respective syd_emu thread in case the
         //    process is no longer valid otherwise we may end up with a
         //    deadlock: See miniupnpc tests, thx kepstin!
         // 2. To prevent PID reuse vectors we validate the request ID.
-        #[allow(clippy::cast_possible_wrap)]
-        let pid = Pid::from_raw(interrupt.request.pid as libc::pid_t);
-        let status = match proc_status(pid) {
-            Ok(status) if self.is_valid(interrupt.request.id) => {
-                // SAFETY: seccomp-id validated, proc_pid_status(5) is valid.
-                status
+        let status_fd = if let Some(fd) = interrupt.status.as_ref() {
+            if let Err(errno) = lseek64(fd, 0, Whence::SeekSet) {
+                if self.is_valid(interrupt.request.id) {
+                    let source = InterruptSource::ProcessInvalid(errno);
+                    Self::interrupt(tgid, interrupt, source);
+                } // no need to interrupt for invalid seccomp-id.
+                return false;
             }
+            fd
+        } else {
+            let fd = match proc_status_open(interrupt.request.pid()) {
+                Ok(fd) if self.is_valid(interrupt.request.id) => {
+                    // seccomp-id validated, proc_pid_status(5) is valid.
+                    fd
+                }
+                Err(errno) if self.is_valid(interrupt.request.id) => {
+                    let source = InterruptSource::ProcessInvalid(errno);
+                    Self::interrupt(tgid, interrupt, source);
+                    return false;
+                }
+                // seccomp-id invalid, no need to interrupt.
+                _ => return false,
+            };
+            interrupt.status = Some(fd);
+            #[expect(clippy::disallowed_methods)]
+            interrupt.status.as_ref().unwrap()
+        };
+
+        // Parse proc_pid_status(5).
+        let status = match proc_status_read(status_fd) {
+            Ok(status) => status,
             Err(errno) if self.is_valid(interrupt.request.id) => {
                 let source = InterruptSource::ProcessInvalid(errno);
                 Self::interrupt(tgid, interrupt, source);
                 return false;
             }
+            // seccomp-id invalid, no need to interrupt.
             _ => return false,
         };
 
-        // Check for per-{thread,process} pending signals:
+        // Check pending signals for the thread.
+        //
+        // 1. Check for per-{thread,process} pending signals:
         //
         // Aggregate pending signals from both the thread-local and
         // process-global queues. In Linux, each thread (LWP) has its
@@ -272,7 +320,7 @@ impl Interrupter {
         // interrupt this execution context.
         let mut sigset = status.sig_pending_thread | status.sig_pending_process;
 
-        // Exclude per-thread blocked signals:
+        // 2. Exclude per-thread blocked signals:
         //
         // Remove those signals currently masked by this thread's signal
         // mask. Signal masks are maintained per-thread via
@@ -287,12 +335,9 @@ impl Interrupter {
             return true;
         }
 
-        // Filter out restarting signals per-process,
-        // unless ignore_restart is set. This may be the
-        // case e.g. when the socket has a timeout for
-        // accept and connect.
-        // Note, `interrupt.ignore_restart` check
-        // was done before calling this function and
+        // 3. Filter out restarting signals per-process, unless ignore_restart is set.
+        // This may be the case e.g. when the socket has a timeout for accept and connect.
+        // Note, `interrupt.ignore_restart` check was done before calling this function and
         // sigset_restart is only Some if it is false.
         if !interrupt.ignore_restart {
             if let Some(sigset_restart) = self
@@ -301,7 +346,7 @@ impl Interrupter {
                 .sig_restart
                 .lock()
                 .unwrap_or_else(|err| err.into_inner())
-                .get(&Pid::from_raw(status.pid))
+                .get(&status.pid)
             {
                 sigset.del_set(*sigset_restart);
 
@@ -318,7 +363,7 @@ impl Interrupter {
     }
 
     // Interrupt the respective `syd_emu` thread.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     fn interrupt(tgid: Pid, interrupt: &SysInterrupt, source: InterruptSource) {
         match tgkill(tgid, interrupt.handler, libc::SIGALRM) {
             Ok(_) | Err(Errno::ESRCH) => {
@@ -344,7 +389,7 @@ impl Interrupter {
     }
 
     /// Confine Interrupter thread.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
         seccomp_fd: RawFd,
         tgid: Pid,
@@ -366,8 +411,8 @@ impl Interrupter {
         ctx.set_ctl_nnp(true)?;
 
         // Disable Speculative Store Bypass mitigations
-        // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
+        // with trace/allow_unsafe_exec_speculative:1
+        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
 
         // DO NOT synchronize filter to all threads.
         // Other threads will self-confine.
@@ -387,7 +432,7 @@ impl Interrupter {
         // Allow interrupt handler thread to send the
         // SIGALRM signal to threads in Syd's thread group.
         let sysname = "tgkill";
-        #[allow(clippy::cast_sign_loss)]
+        #[expect(clippy::cast_sign_loss)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 ctx.add_rule_conditional(
@@ -408,8 +453,8 @@ impl Interrupter {
         // Allow interrupt handler thread to
         // validate seccomp request IDs using ioctl(2).
         let sysname = "ioctl";
-        #[allow(clippy::cast_sign_loss)]
-        #[allow(clippy::unnecessary_cast)]
+        #[expect(clippy::cast_sign_loss)]
+        #[expect(clippy::unnecessary_cast)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 ctx.add_rule_conditional(
@@ -430,8 +475,8 @@ impl Interrupter {
         // Allow openat2 with the static procfs(5) fd only.
         // Apply system call argument cookies.
         let sysname = "openat2";
-        #[allow(clippy::cast_sign_loss)]
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::cast_sign_loss)]
+        #[expect(clippy::useless_conversion)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 ctx.add_rule_conditional(
@@ -454,10 +499,18 @@ impl Interrupter {
         // We need this because std::thread::spawn has unwanted
         // side-effects such as opening /sys/devices/system/cpu/online
         // on some architectures.
+        //
+        // Note, we avoid this when profiling is enabled,
+        // as gperf requires it to write profiling data.
         for sysname in ["open", "openat", "stat", "lstat", "statx", "newfstatat"] {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Errno(Errno::ENOSYS as i32), syscall)?;
+                    let action = if !cfg!(feature = "prof") {
+                        ScmpAction::Errno(Errno::ENOSYS as i32)
+                    } else {
+                        ScmpAction::Allow
+                    };
+                    ctx.add_rule(action, syscall)?;
                 }
                 Err(_) => {
                     info!("ctx": "confine", "op": "allow_int_syscall",
@@ -482,6 +535,17 @@ impl Interrupter {
             }
         }
 
+        // Allow safe prctl(2) operations.
+        let sysname = "prctl";
+        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
+            for (_, op) in INT_PRCTL_OPS {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
+            }
+        } else {
+            info!("ctx": "confine", "op": "allow_int_syscall",
+                "msg": format!("invalid or unsupported syscall {sysname}"));
+        }
+
         // Prevent executable memory.
         const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
         for sysname in ["mmap", "mmap2", "mprotect"] {
@@ -504,8 +568,11 @@ impl Interrupter {
         // No proc_pid_mem(5) access required here.
         confine_scmp_write(&mut ctx, None, false)?;
 
+        // Allow safe madvise(2) advice.
+        confine_scmp_madvise(&mut ctx)?;
+
         // Allow safe system calls.
-        for sysname in INT_SYSCALLS {
+        for sysname in INT_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
